/*
	File: CASYNCDB.CPP
	Description:
		A virtual class to perform Winsock asynchronous database
		lookup.
	Copyright 1995, David G. Roberts
*/

#include "casyncdb.h"
#include "myassert.h"

/*
	Function: CAsyncDB::~CAsyncDB
	Description:
		Destructor for CAsyncDB class.  Ensures that any pending
		lookup is cancelled before the object is destroyed.
*/
CAsyncDB::~CAsyncDB()
{
	// Make sure we stop anything that's currently in progress.
	StopLookup();
	delete[] m_pBuffer;
}

/*
	Function: CAsyncDB::StartLookup
	Description:
		Starts a lookup process using a CAsyncDB object.  The object
		must not have a lookup pending.  This means that object may
		have been created using the default constructor, a previous
		lookup is complete, or it was cancelled using StopLookup.
	Inputs:
		pCallback -- pointer to the object to receive the callback.
		iToken -- a user-specified value that is passed to the
			callback.
	Outputs:
		Returns an error code if a problem occured.  The error can be
		a Winsock error or a private ASYNCDBERR_* code.
		ASYNCDBERR_NOERROR will be returned if no error occurs.
*/
int CAsyncDB::StartLookup(CAsyncDBCallback * pCallback, int iToken)
{
	// If something is already going, don't interrupt it.  The caller
	// must invoke StopLookup first.
	if (m_stState == INPROGRESS)
		return m_iLastError = ASYNCDBERR_INPROGRESS;
	// Okay, now we're going.
	m_stState	= INPROGRESS;

	// Kick off the lookup.
	m_hWSATask = StartWinsockLookup();

    // Did something go wrong?
	if (m_hWSATask == NULL) {
		// Save error for future use.
		m_stState		= INVALID;
		return m_iLastError = WSAGetLastError();
	}

	// Set up the callback.
	m_pCallback = pCallback;
    m_iToken = iToken;
	Register((WPARAM)m_hWSATask, WPARAM_TYPE_HANDLE);

    return ASYNCDBERR_NOERROR;
}

/*
	Function: CAsyncDB::StopLookup
	Description:
		Stops an in progress lookup.
	Inputs: None.
	Outputs:
		Returns an error code if a problem occured.  The error can be
		a Winsock error or a private ASYNCDBERR_* code.
		ASYNCDBERR_NOERROR will be returned if no error occurs.
*/
int CAsyncDB::StopLookup(void)
{
	// See if we're really in progress.
	if (m_stState != INPROGRESS)
		return m_iLastError = ASYNCDBERR_NOTINPROGRESS;

	// Okay, now the address state is invalid.
	m_stState = INVALID;

	// Remove task handle from event callback list.
	Deregister((WPARAM)m_hWSATask, WPARAM_TYPE_HANDLE);
	m_pCallback = NULL;
    m_iToken = 0;

	// Cancel the lookup with Winsock.
	int iCancelResult = WSACancelAsyncRequest(m_hWSATask);
	if (iCancelResult == SOCKET_ERROR)
		m_iLastError = WSAGetLastError();
	// Map WSAEINVAL and WSAEALREADY to no error.  We'll assume that
	// the task handle really was valid but that Winsock has finished
	// the lookup.  Our application hasn't dispatched the Winsock
	// event message that is probably sitting in the message queue
	// right now, which is why we think the operation is still active
	// while Winsock doesn't.
	if (iCancelResult == SOCKET_ERROR && (m_iLastError == WSAEINVAL ||
		m_iLastError == WSAEALREADY)) {
		// Make function return 0, but leave m_iLastError set to real
        // error code, just in case somebody wants to check this.
		iCancelResult = 0;
	}

    // Return an error code.
	if (iCancelResult == 0)
		return ASYNCDBERR_NOERROR;
	else
		return m_iLastError;
}

/*
	Function: CAsyncDB::WinsockEventCallback
	Description:
		Virtual function to handle the Winsock event message.
		Basically, this just deregisters the callback, gets the
		address from the SERVENT structure, stores it away so the
		app can retrieve it later, and finally calls the completion
		callback function.
	Inputs:
		Standard wndproc parameters.
	Outputs: None.
*/
void CAsyncDB::WinsockEventCallback(HWND hwnd, UINT msg,
	WPARAM wParam, LPARAM lParam)
{
	UNUSED_PARAMETER(hwnd);
	UNUSED_PARAMETER(msg);

	// Check our parameters.
	ASSERT((HANDLE) wParam == m_hWSATask);

	// Deregister the event callback.
	Deregister((WPARAM)m_hWSATask, WPARAM_TYPE_HANDLE);

	// Break out the result code.
	int iError = WSAGETASYNCERROR(lParam);

	if (iError == 0) {
		// No error, so now we're valid.
		m_stState	= VALID;
		m_iLastError = ASYNCDBERR_NOERROR;
	}
	else {
		// Some error occurred.
		m_stState = INVALID;
		m_iLastError = iError;
	}

	// Invoke the completion callback.
	ASSERT(m_pCallback != NULL);
	CAsyncDBCallback * pCallback = m_pCallback;
    m_pCallback = NULL;
	pCallback->AsyncDBCallback(m_iToken, iError);
}
