/*
	File: CASYNCDB.H
	Description:
		A virtual class to perform Winsock asynchronous database
		lookup.
	Copyright 1995, David G. Roberts
*/

#ifndef _CASYNCDB_H
#define _CASYNCDB_H

#include <windows.h>
#include "winsock.h"
#ifndef _CWINSOCK_H
#include "cwinsock.h"
#endif

/* CONSTANTS */
/* Error codes */
#define ASYNCDBERR_NOERROR			0
#define ASYNCDBERR_INPROGRESS		1
#define ASYNCDBERR_NOTINPROGRESS	2

/*
	Class: CAsyncDBCallback
	Description:
		A class describing a callback interface used by CAsyncDB to
		inform another object that an asynchronous lookup has
		completed. Classes using a CAsyncDB derived class for
		asynchronous lookup should inherit this callback class
		interface and override the virtual function appropriately.
*/
class CAsyncDBCallback {
public:
	// iToken will contain the token value passed to the lookup class
	//   when it is constructed.
	// iError will contain a Winsock error number.
	virtual void AsyncDBCallback(int iToken, int iError) = 0;
};

/*
	Class: CAsyncDB
	Description:
		CAsyncDB is used to look up service port numbers.
*/
class CAsyncDB : protected CWinsockEventCallback {
public:
	enum STATE { INVALID, INPROGRESS, VALID };
protected:
	char *				m_pBuffer;		// Buffer for lookup info.
	CAsyncDBCallback *	m_pCallback;	// Pointer to callback object.
	int					m_iToken;		// Callback token.
	STATE				m_stState;		// Current lookup state.
	HANDLE				m_hWSATask;		// Winsock task handle.
	int					m_iLastError;	// Last error for this object.

	// Make the event callback private so that only the event dispatch
	// window can "see" it.  It won't be visible to users of the
	// class.
	virtual void WinsockEventCallback(HWND hwnd, UINT msg,
		WPARAM wParam, LPARAM lParam);

	// This function should be customized in all classes that inherit
	// from this base class.  It's what kicks off the actual lookup
	// in Winsock and is specific to the information that is being
	// looked up.
    virtual HANDLE StartWinsockLookup(void) = 0;
private:
	// Hide this stuff so the compiler doesn't generate default
	// versions of them and nobody can use them by mistake.
	CAsyncDB(CAsyncDB&);	// Copy constructor
	operator=(CAsyncDB&);	// Assignment operator.
public:
	CAsyncDB(void)
		: m_pBuffer(new char[MAXGETHOSTSTRUCT]), m_pCallback(NULL),
		m_iToken(0), m_stState(INVALID) {};
	virtual ~CAsyncDB();	// Virtual destructor because it's a base
							// class.
	int StartLookup(CAsyncDBCallback * pCallback, int iToken);
	int StopLookup(void);
	STATE State(void) { return m_stState; };
	int LastError(void) { return m_iLastError; };
};

#endif
