/*
	File: CFILEMAP.CPP
	Description:
		An object used to open and pseudo-memory-map a file.  The
		memory-mapping is simulated using C++ operator overloading.
	Copyright 1995, David G. Roberts
*/

#include "cfilemap.h"

/*
	Function: CFileMap::CFileMap
	Description:
		Constructor for CFileMap class.  The function allocates a
		buffer cSize in length.
*/
CFileMap::CFileMap(const unsigned nSize) :
	m_pData(NULL), m_nBufferSize(nSize),
	m_hFile(NULL), m_iError(CMM_NO_ERROR), m_bIsOpen(FALSE)
{
}

/*
	Function: CFileMap::~CFileMap
	Description:
		Simply make sure the file is closed before we're destroyed.
*/
CFileMap::~CFileMap()
{
	// Make sure file is closed.
	Close();
}

/*
	Function: CFileMap::Open
	Description:
		Open the file, allocate the buffer, and read the first bytes
		of the file into it.
*/
int CFileMap::Open(LPCSTR lpszFilename)
{
	// Open file using Windows' _lopen.
	m_hFile = _lopen(lpszFilename, OF_READ | OF_SHARE_DENY_WRITE);
	if (m_hFile == HFILE_ERROR) {
		m_hFile = NULL;
		return m_iError = CMM_OPEN_ERROR;
	}

	// Allocate the buffer.
    m_pData = new char[m_nBufferSize];

	// Read as much of file into buffer as possible, starting from
	// beginning of file.
	FillBuffer(0L);

	m_bIsOpen = TRUE;

	return m_iError = CMM_NO_ERROR;
}

/*
	Function: CFileMap::Close
	Description:
		Delete the data buffer and close the file.
*/
int CFileMap::Close(void)
{
	// Get rid of the buffer.
	delete[] m_pData;
	m_pData = NULL;

    // Close the file.
	if (m_hFile == NULL)
		return m_iError = CMM_NO_ERROR;
	if (_lclose(m_hFile) == HFILE_ERROR)
		m_iError = CMM_CLOSE_ERROR;
	else
		m_iError = CMM_NO_ERROR;
	m_hFile = NULL;

    m_bIsOpen = FALSE;

	return m_iError;
}

/*
	Function: CFileMap::FileLength
	Description:
		Return the length of the file.
*/
long CFileMap::FileLength(void)
{
	LONG lSeekPos = _llseek(m_hFile, 0, 2); // Seek to end of file
	if (lSeekPos == HFILE_ERROR) {
		return m_iError = CMM_SEEK_ERROR;
	}
	else {
    	m_iError = CMM_NO_ERROR;
		return lSeekPos;
	}
}

/*
	Function: CFileMap::FillBuffer
	Description:
		Fill the buffer with file data starting at lBaseIndex
		within the file.
*/
int CFileMap::FillBuffer(const long lBaseIndex)
{
	// Seek to correct position in file.
	LONG lSeekPos = _llseek(m_hFile, lBaseIndex, 0);
	if (lSeekPos == HFILE_ERROR || lSeekPos != lBaseIndex)
		return m_iError = CMM_SEEK_ERROR;

	// Read data in.
	UINT uResult = _lread(m_hFile, m_pData, m_nBufferSize);
	if (uResult == (UINT)HFILE_ERROR) {
		return m_iError = CMM_READ_ERROR;
	}
	if (uResult != m_nBufferSize) {
		// Zero fill remainder of buffer.
		for (unsigned i = uResult; i < m_nBufferSize; i++)
			m_pData[i] = 0;
	}

	// Update buffer variables.
	m_lBaseIndex	= lBaseIndex;
	m_uLen			= uResult;

	return m_iError = CMM_NO_ERROR;
}

/*
	Function: CFileMap::ReadAndGet
	Description:
		All access to the file data is through the Get member
		function.  The Get function is inlined for speed.  If the
		data is in the buffer, then the inlined code should execute
		quickly.  If the Get function determines that the desired
		data is not in the buffer, then it calls ReadAndGet to read
		the data into the buffer and return the appropriate data.
		Since reading the file is a slow operation, this code is
		put into its own function to save space in the inlined
		Get function.
*/
int CFileMap::ReadAndGet(const long lIndex)
{
	// We didn't have the right data, so get the right data
	long lBaseIndex = (lIndex / m_nBufferSize) * m_nBufferSize;
	int iResult = FillBuffer(lBaseIndex);
	if (iResult != CMM_NO_ERROR) {
		return m_iError = iResult;
	}

	// Try again...
	unsigned uOffset = (unsigned) (lIndex - m_lBaseIndex);
	if (uOffset < m_uLen)
		return m_pData[uOffset];
	else
		return EOF;
}

