/*
	File: CFINGOBJ.CPP
	Description:
		An class to implement the finger protocol.
	Copyright 1995, David G. Roberts
*/

#include "cfingobj.h"
#include "myassert.h"

/*
	Function: CFingerObj::CFingerObj
	Description:
		Constructor for the CFingerObj class. Stores the constructor
		parameters in class member variables and kicks off the
		state machine at the Start state.
*/
CFingerObj::CFingerObj(LPCSTR lpszHost, u_short ushPort,
	LPCSTR lpszUser, HWND hwnd, int iMessage, HINSTANCE hInstance)
	: m_pszHost(new char[lstrlen(lpszHost) + 1]), m_ushPort(ushPort),
	m_pszUser(new char[lstrlen(lpszUser) + 1]), m_hwnd(hwnd),
	m_iMessage(iMessage), m_hInstance(hInstance),
	m_stState(START), m_evEvent(NULLEVENT), m_stream(),
    m_plnTail(&m_lnHead),
	m_pDataBuffer(new char[BUFFER_QUANTUM]), m_uDataSize(0),
	m_uBufferSize(BUFFER_QUANTUM)
{
	m_stream.SetCallback(this, 0);
	lstrcpy(m_pszHost, lpszHost);
    lstrcpy(m_pszUser, lpszUser);
	Event(NULLEVENT);
}

/*
	Function: CFingerObj::~CFingerObj
	Description:
		Destructor for the CFingerObj class.  Simply deletes the
		memory which was allocated in the constructor.
*/
CFingerObj::~CFingerObj()
{
	delete[] m_pszHost;
	delete[] m_pszUser;
	delete[] m_pDataBuffer;
}

/*
	Function: CFingerObj::AsyncDBCallback
	Description:
		The function which handles the completion callback from
		the CHost object.  The error code from the host lookup
		is saved and a LOOKUPDONE event generated to the state
		machine.
*/
void CFingerObj::AsyncDBCallback(int iToken, int iError)
{
	UNUSED_PARAMETER(iToken);	// Eliminate the compiler warning.

	m_iError = iError;
	Event(LOOKUPDONE);	
}

/*
	Function: CFingerObj::StreamCallback
	Description:
		This function receives callback from the CStreamSocket
		object.  The function turns the various CStreamSocket events
		into events for the CFingerObj state machine.
*/
void CFingerObj::StreamCallback(int iToken, STREAMEVENT iEvent,
	int iError, char FAR *lpBuffer)
{
	UNUSED_PARAMETER(iToken);	// Eliminate the compiler warning.

	m_iError = iError;
	m_lpSentBuffer = lpBuffer;
	switch (iEvent) {
		case CStreamCallback::SE_ACCEPT:
			// Error!  We are connecting to someone else, not
			// accepting connections from others.
			WSCFatalError(
			"Recieved SE_ACCEPT event in CFingerObj::StreamCallback");
		case CStreamCallback::SE_CLOSE:
			Event(CLOSE);
			break;
		case CStreamCallback::SE_CONNECT:
			Event(CONNECT);
			break;
		case CStreamCallback::SE_DATAREADY:
			Event(DATAREADY);
			break;
		case CStreamCallback::SE_DATASENT:
			Event(DATASENT);
			break;
		default:
        	break;
	}
}

/*
	Function: CFingerObj::Event
	Description:
		The main function for the state machine.  It dispatches
		the input event to the appropriate state function.  If
		the state function returns TRUE, the machine continues to
		dispatch the event to the next function.  Presumably, the
		previous state function has set m_evEvent to NULLEVENT
		in order to "swallow" the prior event.
*/
void CFingerObj::Event(EVENT ev)
{
	m_evEvent = ev;

	BOOL bContinue = FALSE;

	do {
		switch (m_stState) {
			case START:
				bContinue = Start(m_evEvent);
				break;
			case AWAITLOOKUP:
				bContinue = AwaitLookup(m_evEvent);
				break;
			case MAKECONNECTION:
				bContinue = MakeConnection(m_evEvent);
				break;
			case AWAITCONNECTION:
				bContinue = AwaitConnection(m_evEvent);
				break;
			case CONNECTED:
				bContinue = Connected(m_evEvent);
				break;
			case AWAITDATA:
				bContinue = AwaitData(m_evEvent);
				break;
			case CLOSED:
				bContinue = Closed(m_evEvent);
                break;
			case ERROREXIT:
				bContinue = ErrorExit(m_evEvent);
				break;
			case FINISH:
				bContinue = FALSE;
				break;
		}
	} while (bContinue);
}

/*
	Function: CFingerObj::Start
	Description:
		The first function of the state machine.  Starts a host
        lookup for the host name given to the CFingerObj constructor.
*/
BOOL CFingerObj::Start(EVENT ev)
{
	int iError;
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

	// Note: this little test won't work when IPv6 is deployed
	// and IP addresses move to 128 bits.  We're assuming IP addresses
	// are 32 bits wide and can fit in a u_long.
	u_long addr = inet_addr(m_pszHost);
	// Note: the INADDR_NONE constant on the following line may
    // generate a warning.  Ignore it.  It's fine.
	if (addr == INADDR_NONE) {
		// Addr must be a host name. Kick off a lookup.
		m_bDotted = FALSE;
		iError = m_host.StartLookup(m_pszHost, this, 0);
		if (iError != WSANOERROR) {
			ReportWSAError(
				"Couldn't start a name lookup in CFingerObj::Start",
				iError, m_hwnd, m_hInstance);
			m_stState = ERROREXIT;
            m_evEvent = NULLEVENT;
			return TRUE;
		}
		m_stState = AWAITLOOKUP;
		return FALSE;
	}
	else {
		// Addr was in dotted decimal notation.
        m_bDotted = TRUE;
		addr = ntohl(addr);
		m_saAddr.Addr(addr);
		m_stState = MAKECONNECTION;
		m_evEvent = NULLEVENT;
        return TRUE;
    }
}

BOOL CFingerObj::AwaitLookup(EVENT ev)
{
	// We're only expecting a LOOKUPDONE event.
	if (ev != LOOKUPDONE)
		return FALSE;
	if (m_iError != WSANOERROR) {
		// Something didn't go right.
		if (m_iError == WSAHOST_NOT_FOUND || m_iError == WSANO_DATA ||
			m_iError == WSATRY_AGAIN) {
			// Okay, we couldn't find the name string in the database.
			// Use a nice dialog box to inform the user.
			ErrorWithHostName("Host Not Found",
				"Couldn't find an address for host '%s'.");
		}
		else {
			// Some other sort of error occurred.  Alert the user with
            // the more ugly dialog box.
			const char szFormat[] =
				"Unexpected error while looking up address "
				"for host '%s'.";
			char * pszMsg = new char[sizeof(szFormat) +
				lstrlen(m_pszHost) + 1];
			wsprintf(pszMsg, szFormat, (LPSTR)m_pszHost);
			ReportWSAError(pszMsg, m_iError, m_hwnd, m_hInstance);
			delete[] pszMsg;
		}
		m_stState = ERROREXIT;
		m_evEvent = NULLEVENT;
        return TRUE;
	}
	// Set our address to the first address from the host lookup.
	BOOL bResult;
	m_saAddr = m_host.FirstAddr(&bResult);
	if (bResult != TRUE) {
		MessageBox(m_hwnd, "Error accessing address after lookup.",
			"Error", MB_OK | MB_ICONEXCLAMATION);
		m_stState = ERROREXIT;
		m_evEvent = NULLEVENT;
        return TRUE;
	}
	m_stState = MAKECONNECTION;
    m_evEvent = NULLEVENT;
    return TRUE;
}

BOOL CFingerObj::MakeConnection(EVENT ev)
{
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

	// Okay, we've got our address.  Fill in the port and try to make
	// a connection.
	m_saAddr.Port(m_ushPort);
	int iError = m_stream.Connect(m_saAddr);
	if (iError != WSANOERROR && iError != WSAEWOULDBLOCK) {
		// Something went wrong.  We don't expect anything right here,
		// so don't bother making things too pretty.
		const char szFormat[] =
			"Error while connecting to foreign host '%s'.";
		char * pszMsg = new char[sizeof(szFormat) +
			lstrlen(m_pszHost) + 1];
		wsprintf(pszMsg, szFormat, (LPSTR)m_pszHost);
		ReportWSAError(pszMsg, iError, m_hwnd, m_hInstance);
		delete[] pszMsg;
		m_stState = ERROREXIT;
		m_evEvent = NULLEVENT;
		return TRUE;
	}
	m_stState = AWAITCONNECTION;
	m_evEvent = NULLEVENT;
    return FALSE;
}

BOOL CFingerObj::AwaitConnection(EVENT ev)
{
	if (ev != CONNECT)
		return FALSE;
	// See if we connected or had an error.
	if (m_iError == 0) {
		// Everything was okay.  We're now connected.  Move on.
		m_stState = CONNECTED;
		m_evEvent = NULLEVENT;
		return TRUE;
	}
	// Okay, something went wrong.  See what it was.
	if (m_iError == WSAENETUNREACH || m_iError == WSAETIMEDOUT) {
		// Perhaps we can't reach the foreign host on the primary
		// address.  If the foreign host has multiple addresses, try
		// one of the others until we're out of addresses.
		BOOL bResult;
		m_saAddr = m_host.NextAddr(&bResult);
		if (!m_bDotted && bResult == TRUE) {
			// We didn't have a dotted address to begin with and
			// the host has more addresses left, so try again.
            m_stream.Close();
			m_stState = MAKECONNECTION;
			m_evEvent = NULLEVENT;
			return TRUE;
		}
		// Okay, there are no more addresses left to try.  Generate
		// a user friendly error.
		if (m_iError == WSAENETUNREACH) {
			ErrorWithHostName("Network Unreachable",
				"The network for host '%s' is unreachable.");
		}
		else if (m_iError == WSAETIMEDOUT) {
			ErrorWithHostName("Connection Timed Out",
				"Connection attempt timed out while trying "
				"to reach host '%s'.");
		}
		m_stState = ERROREXIT;
		m_evEvent = NULLEVENT;
		return TRUE;
	}
	// Okay, we got an error.  See if it's something we might expect
	// and try to make it sound a bit nicer.
	if (m_iError == WSAECONNREFUSED) {
		// We reached the other host but they refused us.
		ErrorWithHostName("Connection Refused",
			"Host '%s' refused the connection on the specified port. "
			"Either you used the wrong port number or the desired "
			"service is not running on the host.");
		m_stState = ERROREXIT;
		m_evEvent = NULLEVENT;
		return TRUE;
	}
	// It was something unexpected (like we're out of socket
	// descriptors).  Signal with an ugly error message.
	const char szFormat[] =
		"Error while connecting to host '%s'.";
	char * pszMsg = new char[sizeof(szFormat) +
		lstrlen(m_pszHost) + 1];
	wsprintf(pszMsg, szFormat, (LPSTR)m_pszHost);
	ReportWSAError(pszMsg, m_iError, m_hwnd, m_hInstance);
	delete[] pszMsg;
	m_stState = ERROREXIT;
	m_evEvent = NULLEVENT;
	return TRUE;
}

BOOL CFingerObj::Connected(EVENT ev)
{
	UNUSED_PARAMETER(ev);

	// Send the name of the user followed by a CRLF pair.
	int iError = m_stream.Send(m_pszUser, lstrlen(m_pszUser));
	if (iError != WSANOERROR && iError != WSAEWOULDBLOCK) {
		// We expect to either have the data sent with no errors or
		// to block waiting for Winsock (in reality, with this little
		// data it should never block).  If something else happened,
		// flag the error.
		ReportWSAError("Error trying to send data.", iError, m_hwnd,
			m_hInstance);
		m_stState = ERROREXIT;
		m_evEvent = NULLEVENT;
		return TRUE;
	}
	iError = m_stream.Send("\015\012", 2);	// CRLF = 15,12 (octal)
	if (iError != WSANOERROR && iError != WSAEWOULDBLOCK) {
		ReportWSAError("Error trying to send data.", iError, m_hwnd,
			m_hInstance);
		m_stState = ERROREXIT;
		m_evEvent = NULLEVENT;
		return TRUE;
	}
	m_stState = AWAITDATA;
	m_evEvent = NULLEVENT;
	return FALSE;
}

BOOL CFingerObj::AwaitData(EVENT ev)
{
	int iLength;
	int iError;
	char * pNewBuffer;
	unsigned i;
	unsigned j;
	BOOL bFound;
	LineNode * plnNewNode;

	switch (ev) {
		case DATAREADY:
			// Figure out how much space is left in the buffer.
			// Read that number of bytes, max.
			iLength = m_uBufferSize - m_uDataSize;
			iError = m_stream.Receive(m_pDataBuffer + m_uDataSize,
				&iLength, 0);
			if (iError == WSANOERROR && iLength > 0) {
				m_uDataSize += iLength;
				do {
					// Scan for a CRLF.
					bFound = FALSE;
					for (i = 0; i < m_uDataSize; i++) {
						if (m_pDataBuffer[i] == '\015' &&
							(i + 1) < m_uDataSize &&
							m_pDataBuffer[i + 1] == '\012') {
								bFound = TRUE;
								break;
						}
					}
					if (bFound) {
						// Allocate a new line node and add it to the
						// end of the linked list.
						plnNewNode = new LineNode;
						m_plnTail->m_plnNext = plnNewNode;
						m_plnTail = plnNewNode;
						// Allocate space for the line data.  At this
						// point, i equals the number of characters
						// in the line minus the CRLF.  Allocate
						// enough for the characters before the CRLF
						// and the null terminator.
						plnNewNode->m_pszLine = new char[i + 1];
						// Copy the characters to the line buffer.
						for (j = 0; j < i; j++) {
							plnNewNode->m_pszLine[j] =
								m_pDataBuffer[j];
						}
						plnNewNode->m_pszLine[i] = '\0';
						// Copy the characters in the data buffer
						// following the CRLF down to the start of
						// the buffer.
						i += 2; // Skip over the CRLF pair.
						for (j = 0; j < (m_uDataSize - i); j++) {
							m_pDataBuffer[j] = m_pDataBuffer[i + j];
						}
						m_uDataSize -= i;
					}
					else {
						// See if our buffer is totally full.  If so,
						// then enlarge it by another quantum.
						if (m_uDataSize == m_uBufferSize) {
							pNewBuffer = new char[m_uBufferSize +
								BUFFER_QUANTUM];
							for (i = 0; i < m_uBufferSize; i++) {
								pNewBuffer[i] = m_pDataBuffer[i];
							}
							delete[] m_pDataBuffer;
							m_pDataBuffer = pNewBuffer;
							m_uBufferSize += BUFFER_QUANTUM;
						}
					}
				} while (bFound && m_uDataSize > 0);
				return FALSE;
			}
			// Either an error or end-of-file.
			m_stream.Close();
			if (iError != WSANOERROR) {
				// An error.  Report it and error exit.
				ReportWSAError("Error receiving data.", iError,
					m_hwnd, m_hInstance);
				m_stState = ERROREXIT;
				m_evEvent = NULLEVENT;
				return TRUE;
			}
			// End-of-file.  The socket was closed gracefully.
			m_stState = CLOSED;
			m_evEvent = NULLEVENT;
            return TRUE;
		case CLOSE:
			m_stream.Close();
			m_stState = CLOSED;
			m_evEvent = NULLEVENT;
			return TRUE;
		default:
			return FALSE;
	}
}

BOOL CFingerObj::Closed(EVENT ev)
{
	UNUSED_PARAMETER(ev);

	// Count the number of lines in the text.
	int iNumLines = 0;
	LineNode * plnNode = m_lnHead.m_plnNext;
	while (plnNode != NULL) {
		iNumLines++;
		plnNode = plnNode->m_plnNext;
	}

	// Allocate a block of pointers to each of the lines and
	// convert the list of LineNodes to an array of line pointers.
	if (iNumLines > 0) {
		char **ppszLinePointers = new char*[iNumLines];
		plnNode = m_lnHead.m_plnNext;
		int i = 0;
		while (plnNode != NULL) {
			ppszLinePointers[i] = plnNode->m_pszLine;
			LineNode * plnNext = plnNode->m_plnNext;
			delete plnNode;
			plnNode = plnNext;
			i++;
		}
		ASSERT(i == iNumLines);
		PostMessage(m_hwnd, m_iMessage, (WPARAM) iNumLines,
			(LPARAM) ppszLinePointers);
	}
	else {
		PostMessage(m_hwnd, m_iMessage, 0, 0);
	}

	m_stState = FINISH;
	m_evEvent = NULLEVENT;
	return TRUE;
}

BOOL CFingerObj::ErrorExit(EVENT ev)
{
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

	// If we've got some lines allocated already, free them up.
	LineNode * plnNode = m_lnHead.m_plnNext;
	while (plnNode != NULL) {
		delete[] plnNode->m_pszLine;
		LineNode * plnNext = plnNode->m_plnNext;
		delete plnNode;
        plnNode = plnNext;
	}

	// Close up the socket and inform the window that we're done but
    // have no text for it.
    m_stream.Close();
	PostMessage(m_hwnd, m_iMessage, 0, 0);
	m_stState = FINISH;
	m_evEvent = NULLEVENT;
	return TRUE;
}

void CFingerObj::ErrorWithHostName(LPCSTR lpszTitle,
	LPCSTR lpszErrorMsg)
{
	char * pszFinalMsg = new char[lstrlen(lpszErrorMsg) +
		lstrlen(m_pszHost) + 1];
	wsprintf(pszFinalMsg, lpszErrorMsg, (LPSTR)m_pszHost);
	MessageBox(m_hwnd, pszFinalMsg, lpszTitle,
		MB_OK | MB_ICONEXCLAMATION);
	delete[] pszFinalMsg;
}
