/*
	File: CMIME.H
	Description:
		A bunch of functions for parsing and manipulating MIME
		messages and body parts.
	Copyright 1995, David G. Roberts
*/

#ifndef _CMIME_H
#define _CMIME_H

#include <windows.h>
#include "clist.h"
#include "cfilemap.h"

/*
	Class: CMIMEParamTuple
	Description:
		A simple class to store a 2-tuple of MIME parameter name and
		value strings.
*/
class CMIMEParamTuple {
private:
	char *	m_pszParam;
	char *	m_pszValue;
public:
	CMIMEParamTuple() : m_pszParam(NULL), m_pszValue(NULL) {};
	CMIMEParamTuple(const char * pszParam, const char * pszValue)
		{ Param(pszParam); Value(pszValue); };
	CMIMEParamTuple(const CMIMEParamTuple&);
	~CMIMEParamTuple();
	void Param(const char * pszParam);
	char * Param(void) { return m_pszParam; };
	void Value(const char * pszValue);
	char * Value(void) { return m_pszValue; };
	CMIMEParamTuple& operator=(const CMIMEParamTuple&);
};


/*
	Class: CMIMEParamList
	Description:
		A simple list of parameter tuples.  Uses CList as a base
		class and fills in the functionality appropriate for a list
		of CMIMEParamTuples.
*/
class CMIMEParamList : public CList {
private:
	virtual int Compare(void * pvThing1, void * pvThing2);
    virtual void DeleteData(void * pvThing);
public:
	// Constructor and destructor.
	CMIMEParamList() {};
	virtual ~CMIMEParamList();

	// Manipulation functions.
	void Add(CMIMEParamTuple * pTuple);
	void Insert(CMIMEParamTuple * pTuple);
	CMIMEParamTuple * First(void)
		{ return (CMIMEParamTuple *) CList::First(); };
	CMIMEParamTuple * Next(void)
		{ return (CMIMEParamTuple *) CList::Next(); };
};

/*
	Class: CMIMEBodyPart
	Description:
		Parses and stores data about individual MIME body parts.
		Handles multipart messages using a recursive parsing
		strategy.  When the parse is complete, a tree of
		CMIMEBodyPart nodes will have been constructed which
		represents all the body parts in the message.
*/
class CMIMEBodyPart {
private:
	char *				m_pszFilename;
    CFileMap			m_filemap;
	char *				m_pszType;
	char *  			m_pszSubType;
	CMIMEParamList *	m_pParamList;
	char *				m_pszEncoding;
	long				m_lStart; // First char in the part
	long				m_lEnd;	// Last char in the part + 1
	long				m_lCurrent;
	// Relationships to other body parts in the message.
	// Note that we don't use a CList-derived list of children
	// because we need to move more easily through the heirarchy
    // as we display the various body parts.
	CMIMEBodyPart *		m_pParent;
	CMIMEBodyPart *		m_pFirstChild;
	CMIMEBodyPart *		m_pSibling;

	// Private functions
	BOOL ParseHeaders(void);
	BOOL ParseMultipart(void);
	char * GetNextHeader(void);
	long FindCRLF(void);
    long StringSearch(char * pszString);
	BOOL HeaderCompare(char * pszHeader, char * pszHeaderName);
	BOOL ParseMIMEVersion(char * pszHeader);
	BOOL ParseContentType(char * pszHeader);
	BOOL ParseContentTransferEncoding(char * pszHeader);

public:
	CMIMEBodyPart(CMIMEBodyPart * pParent, char * pszFilename,
		long lStart, long lEnd);
	~CMIMEBodyPart();

	BOOL Open(void);
	void Close(void);
	BOOL Parse(void);

	char * Type(void) { return m_pszType; };
	char * SubType(void) { return m_pszSubType; };
	char * Encoding(void) { return m_pszEncoding; };
	CMIMEParamTuple * FirstParam(void)
		{ return m_pParamList->First(); };
	CMIMEParamTuple * NextParam(void)
    	{ return m_pParamList->Next(); };

	int FirstLine(char * pszLine, int nLen);
    int NextLine(char * pszLine, int nLen);

	CMIMEBodyPart * Parent(void) { return m_pParent; };
	CMIMEBodyPart * FirstChild(void) { return m_pFirstChild; };
	CMIMEBodyPart * Sibling(void) { return m_pSibling; };
};

/*
	Function: CMIMEMsg
	Description:
		An object which holds the top of the CMIMEBodyPart tree
		and allows a user to navigate the tree.  MIME viewers should
		create one of these objects and access individual body
		parts in the tree using the navigation functions.
*/
class CMIMEMsg {
private:
	CMIMEBodyPart *	m_pBodyParts;
	CMIMEBodyPart *	m_pCurrentPart;
public:
	CMIMEMsg() : m_pBodyParts(NULL), m_pCurrentPart(NULL) {};
	~CMIMEMsg() { delete m_pBodyParts; };

	BOOL Parse(char *);

	CMIMEBodyPart *	FirstPart(void);
	CMIMEBodyPart *	NextPart(void);
    CMIMEBodyPart * CurrentPart(void) { return m_pCurrentPart; };
};

#endif
