/*
	File: CSMTP.CPP
	Description:
		A class for sending email messages via SMTP (RFC 821).
	Copyright 1995, David G. Roberts

	11/17/95: Fixed bug in CSMTP::StreamCallback that caused a response
	          line to be ignored and the CSMTP state machine to hang.
*/

#include "csmtp.h"
#include "myassert.h"

const char CR = '\015';
const char LF = '\012';

// Used to include at the end of text.  The compiler will splice
// strings.
// Use like m_sock.Send("Foo" CRLF,...).
#define CRLF "\015\012"

/*
	Function: CSMTP::Event
	Description:
		The event dispatch loop for the state machine.
*/
void CSMTP::Event(EVENT ev)
{
	while ((this->*m_pState)(ev)) {
		ev = EV_NULLEVENT;
	};
}

/*
	Function: CSMTP::StreamCallback
	Description:
		Receives callbacks from the CStreamSocket object and
		conditions them into appropriate events for the SMTP
		state machine.  In particular, when a line of response
		text comes back from the server, this function reads it,
		and if a full line of text is present, interprets the
		text to generate an EV_CODE_X event.
*/
void CSMTP::StreamCallback(int iToken, STREAMEVENT iEvent, int iError,
	char FAR *lpBuffer)
{
	UNUSED_PARAMETER(iToken);	// Eliminate the compiler warning.

	m_iError = iError;
	switch (iEvent) {
		case CStreamCallback::SE_CLOSE:
			Event(EV_CLOSE);
			break;
		case CStreamCallback::SE_DATAREADY:
			// Read the data.
			if (ReadReplyData()) {
				// If no errors, see if we have a complete reply line.
				// 11/17/95: Previously the receipt of a continuation
				// line followed by a standard response line in the
				// same call to ReadReplyData would cause the
				// continuation line to be discarded and the standard
				// response	line to be ignored. This caused the state
				// machine to hang waiting for the standard response
				// line.
				// Fix: changed "if (LineComplete()) {" to
				// to "while (LineComplete()) {".
				while (LineComplete()) {
					// Yes, it's a complete line.  See if the line is
					// a continuation reply (of the form "250-Foo").
					if (m_szReplyBuffer[3] != '-') {
						// If not, save the first character (should
						// be the leading digit of a numeric reply
						// code.
						char ch = m_szReplyBuffer[0];
						// Eliminate the line from the receive buffer.
						ShrinkReceiveBuffer();
						// Turn the character code into a state
						// machine event.
						switch (ch) {
							case '1': Event(EV_CODE_1); break;
							case '2': Event(EV_CODE_2); break;
							case '3': Event(EV_CODE_3); break;
							case '4': Event(EV_CODE_4); break;
							case '5': Event(EV_CODE_5); break;
							default:
								WSCFatalError(
									"Invalid SMTP reply received "
									"in CSMTP.");
						}
					}
					else {
						// Yes, it was a continuation line.  Just
						// throw it in the bit bucket.
						ShrinkReceiveBuffer();
					}
				}
			}
			else {
            	// Something went wrong.
				WSCFatalError("Couldn't read from socket in CSMTP.");
			}
			break;
		case CStreamCallback::SE_DATASENT:
			if (lpBuffer == (char FAR *)m_szLineBuffer) {
				m_bBufferQueued = FALSE;
				Event(EV_BUFFERFREE);
			}
			break;
		default:
			// Error!  Something we weren't expecting.
			WSCFatalError(
				"Unexpected socket event in CSMTP::StreamCallback");
			break;
	}
}

/*
	Function: CSMTP::ReadReplyData
	Description:
		When data arrives at the socket to be read, this function
		is called to read it into a local buffer.  The function
		ensures that data already in the buffer is preserved and
		that we don't read more than we can handle.  The function
		normally returns TRUE.  If an error occurs during the read,
		it returns FALSE.
*/
BOOL CSMTP::ReadReplyData(void)
{
	// Figure out how much room is left in the buffer.
	int iBufferLen = MAX_REPLY_SIZE - m_iReplyLen;
	ASSERT(iBufferLen > 0);

	// Receive the data.
	int iResult = m_sock.Receive(&m_szReplyBuffer[m_iReplyLen],
		&iBufferLen);

	if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
		// If no errors, update the length of the data in the buffer.
    	m_iReplyLen += iBufferLen;
		return TRUE;
	}
	else {
    	// Save the error and signal it.
		m_iError = iResult;
		return FALSE;
	}
}

/*
	Function: CSMTP::LineComplete
	Description:
		This function is called when data arrives on the socket
		and has been read into a local buffer.  The function returns
		true if a full line of text is present in the buffer
		(i.e., that a CRLF has been found).  It returns FALSE if
		it cannot find a CRLF in the current data.
*/
BOOL CSMTP::LineComplete(void)
{
	ASSERT(m_iReplyLen <= MAX_REPLY_SIZE);

	for (int i = 0; i < m_iReplyLen; i++) {
		if ((m_szReplyBuffer[i] == CR) && ((i + 1) < m_iReplyLen) &&
			(m_szReplyBuffer[i + 1] == LF)) {
			return TRUE;
		}
	}
    return FALSE;
}

/*
	Function: CSMTP::ShrinkReceiveBuffer
	Description:
		This function is called to remove a full line of response
		text from the buffer when it has been processed.  Any
		text that remains in the buffer following the processed
		line is moved to the front of the buffer.
*/
void CSMTP::ShrinkReceiveBuffer(void)
{
	ASSERT(m_iReplyLen <= MAX_REPLY_SIZE);

	// Find the CRLF at the end of the line.
	for (int i = 0; i < m_iReplyLen; i++) {
		if ((m_szReplyBuffer[i] == CR) && ((i + 1) < m_iReplyLen) &&
			(m_szReplyBuffer[i + 1] == LF)) {
			// Set i equal to the character following the CRLF.
			i += 2;
			// Copy the characters following the CRLF to the beginning
			// of the line.
			for (int j = 0; j < m_iReplyLen - i; j++) {
				m_szReplyBuffer[j] = m_szReplyBuffer[i + j];
			}
			// Subtract number of characters removed from the current
			// number of characters in the buffer.
			m_iReplyLen -= i;
            // Break out of the for(i) loop.
			break;
		}
	}
}

/*
	Function: CSMTP::ConnMsgCallback
	Description:
		Receives callbacks from the connection manager object
		and turns them into events for the SMTP state machine.
*/
void CSMTP::ConnMgrCallback(int iToken,
	CConnMgrCallback::RESULT result)
{
	UNUSED_PARAMETER(iToken);

	ASSERT(result != CConnMgrCallback::INVALID_RESULT);

    // Turn connection callback result into events.
	switch (result) {
		case CConnMgrCallback::OK_CONNECTED:
			Event(EV_CONNECT);
			break;
		case CConnMgrCallback::ERR_NOTCONNECTED:
			m_iError = m_connmgr.LastError();
			Event(EV_ERROR);
			break;
	}
}

/*
	Function: CSMTP::Abort
	Description:
		Public function called when the program wants to abort
		the SMTP transfer.  The function generates an EV_ABORT
        event to the state machine.
*/
void CSMTP::Abort(BOOL bDoCallback)
{
	m_bDoCallback = bDoCallback;
	Event(EV_ABORT);
}

/*
	Function: CSMTP::SendMessage
	Description:
		The function that kicks off SMTP processing.  The parameters
		specify the host and port to connect to, a stream from
		which to draw the message, the from address, and a list
		of recipients.  The function stores away the necessary
		information and starts the SMTP state machine in the
        START state.
*/
BOOL CSMTP::SendMessage(LPCSTR lpszHost, u_short ushPort,
	istream * pisMsgStream, LPCSTR lpszFrom, CAddrList * palAddrList)
{
	ASSERT(lpszHost != NULL);
	ASSERT(lpszFrom != NULL);
	ASSERT(palAddrList != NULL);

	// Make sure we're currently stopped.
	// Store parameters away.
	m_pszHost = new char[lstrlen(lpszHost) + 1];
	lstrcpy(m_pszHost, lpszHost);
	m_ushPort = ushPort;
	m_pisMsgStream = pisMsgStream;
	m_pszFrom = new char[lstrlen(lpszFrom) + 1];
	lstrcpy(m_pszFrom, lpszFrom);
	m_palAddrList = palAddrList;

	// Initialize our state correctly.
	m_iReplyLen = 0;
	m_bBufferQueued = FALSE;
    m_bDoCallback = TRUE;

	// Start us off.
	NextState(&CSMTP::Start);
	Event(EV_NULLEVENT);

    return TRUE;
}

/*
	Function: CSMTP::Start
	Description:
		Start the state machine.  Fire off the connection manager
		to try to connect and then go to the AWAITCONNECTION state.
*/
BOOL CSMTP::Start(EVENT ev)
{
	UNUSED_PARAMETER(ev);

	// Make the connection to the remote host.  Let the connection
    // manager do all the work.
	m_connmgr.Connect(m_pszHost, m_ushPort, &m_sock, this, 0);
	NextState(&CSMTP::AwaitConnection);
    return TRUE;
}

/*
	Function: CSMTP::AwaitConnection
	Description:
		Wait for the connection event from the connection
		manager or an error.  If the connection succeeded, wait
		for the greeting message from the server.
*/
BOOL CSMTP::AwaitConnection(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_connmgr.Abort();
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
            break;
		case EV_CONNECT:
			NextState(&CSMTP::AwaitGreeting);
			break;
		case EV_ERROR:
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		default:
			return FALSE;
	}
    return TRUE;
}

/*
	Function: CSMTP::AwaitGreeting
	Description:
		Waits for the remote SMTP server to send the initial
		greeting.
*/
BOOL CSMTP::AwaitGreeting(EVENT ev)
{
	int iResult;
    char szMyHostName[80];

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		case EV_CODE_2:
			// Okay, we got the greeting.  Send the HELO message in
			// response.
			if (gethostname(szMyHostName, sizeof(szMyHostName)) ==
				SOCKET_ERROR) {
				m_iError = WSAGetLastError();
				m_result = RESULT_FAILURE;
				NextState(&CSMTP::Exit);
			}
			else {
				wsprintf(m_szLineBuffer, "HELO %s" CRLF,
					(LPSTR)szMyHostName);
				iResult = m_sock.Send(m_szLineBuffer,
					lstrlen(m_szLineBuffer));
				if (iResult == WSANOERROR ||
					iResult == WSAEWOULDBLOCK) {
					NextState(&CSMTP::AwaitHelloResponse);
				}
				else {
					m_iError = iResult;
					m_result = RESULT_FAILURE;
					NextState(&CSMTP::Exit);
				}
			}
			break;
		case EV_CODE_1:
		case EV_CODE_3:
		case EV_CODE_4:
		case EV_CODE_5:
			// All other codes are unexpected and generate a failure.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CSMTP::AwaitHelloResponse
	Description:
		Waits for the response to the "HELO" command.
*/
BOOL CSMTP::AwaitHelloResponse(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		case EV_CODE_2:
			wsprintf(m_szLineBuffer, "MAIL FROM:<%s>" CRLF,
				(LPSTR)m_pszFrom);
			iResult = m_sock.Send(m_szLineBuffer,
				lstrlen(m_szLineBuffer));
			if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
				NextState(&CSMTP::AwaitMailFromResponse);
			}
			else {
				m_iError = iResult;
				m_result = RESULT_FAILURE;
				NextState(&CSMTP::Exit);
			}
			break;
		case EV_CODE_1:
		case EV_CODE_3:
		case EV_CODE_4:
		case EV_CODE_5:
			// All other codes are unexpected and generate a failure.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Quit);	// Now go send QUIT.
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CSMTP::AwaitMailFromResponse
	Description:
		Awaits for the response from the "MAIL FROM" command.
*/
BOOL CSMTP::AwaitMailFromResponse(EVENT ev)
{
	int iResult;
	char * pszRecipient;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		case EV_CODE_2:
			pszRecipient = m_palAddrList->First();
            ASSERT(pszRecipient != NULL);
			wsprintf(m_szLineBuffer, "RCPT TO:<%s>" CRLF,
				(LPSTR)pszRecipient);
			iResult = m_sock.Send(m_szLineBuffer,
				lstrlen(m_szLineBuffer));
			if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
				NextState(&CSMTP::AwaitRcptToResponse);
			}
			else {
				m_iError = iResult;
				m_result = RESULT_FAILURE;
				NextState(&CSMTP::Exit);
			}
			break;
		case EV_CODE_1:
		case EV_CODE_3:
		case EV_CODE_4:
		case EV_CODE_5:
			// All other codes are unexpected and generate a failure.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Quit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CSMTP::AwaitRcptToResponse
	Description:
		Waits for the response from the "RCPT TO" command.
		If there are more recipients remaining, it sends another
		"RCPT TO" command and waits for the response again.
*/
BOOL CSMTP::AwaitRcptToResponse(EVENT ev)
{
	int iResult;
	char * pszRecipient;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		case EV_CODE_2:
			pszRecipient = m_palAddrList->Next();
			if (pszRecipient == NULL) {
				lstrcpy(m_szLineBuffer, "DATA" CRLF);
				iResult = m_sock.Send(m_szLineBuffer,
					lstrlen(m_szLineBuffer));
				if (iResult == WSANOERROR || iResult ==
					WSAEWOULDBLOCK) {
					NextState(&CSMTP::AwaitDataResponse);
				}
				else {
					m_iError = iResult;
					m_result = RESULT_FAILURE;
					NextState(&CSMTP::Exit);
				}
			}
			else {
				wsprintf(m_szLineBuffer, "RCPT TO:<%s>" CRLF,
					(LPSTR)pszRecipient);
				iResult = m_sock.Send(m_szLineBuffer,
					lstrlen(m_szLineBuffer));
				if (iResult == WSANOERROR || iResult ==
					WSAEWOULDBLOCK) {
					NextState(&CSMTP::AwaitRcptToResponse);
				}
				else {
					m_iError = iResult;
					m_result = RESULT_FAILURE;
					NextState(&CSMTP::Exit);
				}
			}
			break;
		case EV_CODE_1:
		case EV_CODE_3:
		case EV_CODE_4:
		case EV_CODE_5:
			// All other codes are unexpected and generate a failure.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Quit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CSMTP::AwaitDataResponse
	Description:
		Waits for the response from the "DATA" command.
*/
BOOL CSMTP::AwaitDataResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		case EV_CODE_3:
			// Okay, here we go.
			NextState(&CSMTP::SendData);
			break;
		case EV_CODE_1:
		case EV_CODE_2:
		case EV_CODE_4:
		case EV_CODE_5:
			// All other codes are unexpected and generate a failure.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Quit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CSMTP::SendData
	Description:
		Sends the lines of text making up the message.  If the
		lines start with ".", the transparency rules are invoked
		to escape the leading "." character.  If the socket
		transmit buffer becomes full and a write command blocks,
		the function quits sending data until the buffer becomes
		free again.
*/
BOOL CSMTP::SendData(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		case EV_CODE_1:
		case EV_CODE_2:
		case EV_CODE_3:
		case EV_CODE_4:
		case EV_CODE_5:
			// All other codes are unexpected and generate a failure.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Quit);
			break;
		case EV_NULLEVENT:
			// This stops the machine from progressing when we fill up
			// the socket's buffer.  The machine will progress when
			// the socket sends a SE_DATASENT, which is converted to
			// EV_BUFFERFREE by the StreamCallback function.
			if (m_bBufferQueued) {
            	return FALSE;
			}
			// else fall through.
		case EV_BUFFERFREE:
			if (m_pisMsgStream->peek() == EOF) {
				// Send ".CRLF"
				lstrcpy(m_szLineBuffer, "." CRLF);
				iResult = m_sock.Send(m_szLineBuffer,
					lstrlen(m_szLineBuffer));
				if (iResult == WSANOERROR || iResult ==
					WSAEWOULDBLOCK) {
					NextState(&CSMTP::AwaitDataTransferResponse);
				}
				else {
					m_iError = iResult;
					m_result = RESULT_FAILURE;
					NextState(&CSMTP::Exit);
				}
			}
			else {
				// Get a line of data from the stream.
				m_pisMsgStream->getline(m_szLineBuffer,
					MAX_LINE_SIZE);
				int iLineLength = lstrlen(m_szLineBuffer);
				if (iLineLength > (MAX_LINE_SIZE - 3)) {
					// Whoa!  The line we read was too long.
					m_result = RESULT_FAILURE;
					NextState(&CSMTP::Exit);
				}
				else {
					// Strip all the high bits to make sure we don't
					// violate SMTP's 7-bit US-ASCII coding
					// requirement.
					for (int i = 0; i < iLineLength; i++) {
						m_szLineBuffer[i] =
							(char) (m_szLineBuffer[i] & 0x7F);
					}
					lstrcat(m_szLineBuffer, CRLF);	// Add CRLF.
					if (m_szLineBuffer[0] == '.') {
						// It's got a leading period, so add one, as
						// per SMTP transparency rules
						// (RCF 821, 4.5.2).
						iResult = m_sock.Send(".", 1);
						if (iResult != WSANOERROR && iResult !=
							WSAEWOULDBLOCK) {
							m_iError = iResult;
							m_result = RESULT_FAILURE;
							NextState(&CSMTP::Exit);
                            return TRUE;
						}
					}
					// Send the line.
					iResult = m_sock.Send(m_szLineBuffer,
						lstrlen(m_szLineBuffer));
					if (iResult == WSANOERROR) {
						// If we didn't block, then keep going.
						return TRUE;
					}
					else if (iResult == WSAEWOULDBLOCK) {
						// We blocked, so signal that buffer is queued
						// and keep going.  Note that the EV_NULLEVENT
						// handling above will stop the machine.
						m_bBufferQueued = TRUE;
						return TRUE;
					}
					else {
						m_iError = iResult;
						m_result = RESULT_FAILURE;
						NextState(&CSMTP::Exit);
					}
				}
			}
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CSMTP::AwaitDataTransferResponse
	Description:
		Waits for the response after the data has been sent.
*/
BOOL CSMTP::AwaitDataTransferResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Exit);
			break;
		case EV_CODE_2:
			// It worked!!
			m_result = RESULT_SUCCESS;
			NextState(&CSMTP::Quit);
			break;
		case EV_CODE_1:
		case EV_CODE_3:
		case EV_CODE_4:
		case EV_CODE_5:
			// All other codes are unexpected and generate a failure.
			m_result = RESULT_FAILURE;
			NextState(&CSMTP::Quit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CSMTP::Quit
	Description:
		We're all done.  The data has been sent.  Send the "QUIT"
		command to close the connection.
*/
BOOL CSMTP::Quit(EVENT ev)
{
	UNUSED_PARAMETER(ev);

    int iResult;

	lstrcpy(m_szLineBuffer, "QUIT" CRLF);
	iResult = m_sock.Send(m_szLineBuffer, lstrlen(m_szLineBuffer));
	if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
		NextState(&CSMTP::AwaitQuitResponse);
	}
	else {
		m_iError = iResult;
		m_result = RESULT_FAILURE;
		NextState(&CSMTP::Exit);
	}
	return TRUE;
}

/*
	Function: CSMTP::AwaitQuitResponse
	Description:
		Waits for the response to the "QUIT" command.
*/
BOOL CSMTP::AwaitQuitResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CSMTP::Exit);
			break;
		case EV_CLOSE:	// Just about anything is okay.  We
		case EV_CODE_1:	// transferred the message, so just move on.
		case EV_CODE_2:	// Note that success or failure is indicated
		case EV_CODE_3:	// before this state is entered, so don't
		case EV_CODE_4:	// mess with it here.
		case EV_CODE_5:
			NextState(&CSMTP::Exit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CSMTP::Exit
	Description:
		Deletes all the allocated memory, performs the completion
		callback (which is either a window message or a function
		callback), and transitions to the Stop state.
*/
BOOL CSMTP::Exit(EVENT ev)
{
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

	// Close the socket.
	m_sock.Close();

	// Free up the stuff we allocated when we started.
	delete[] m_pszHost;
	delete[] m_pszFrom;

	// Now, invoke the callback with the connection attempt result.
	if (m_bDoCallback) {
    	DoCallback();
	}

	NextState(&CSMTP::Stop);
	return TRUE;
}

/*
	Function: CSMTP::Stop
	Description:
		The final state of the state machine.  Ignores all events.
*/
BOOL CSMTP::Stop(EVENT ev)
{
	UNUSED_PARAMETER(ev);
	// Simply ignore all events.
	return FALSE;
}

/*********************************************************************
	Callbacks for specialized CSMTP derivatives.
*********************************************************************/
/*
	Function: CSMTPWithWinMsg::DoCallback
	Description:
		Turn the state machine completion into a window message.
*/
void CSMTPWithWinMsg::DoCallback(void)
{
	PostMessage(m_hwnd, m_msg, m_iToken, m_result);
}

/*
	Function: CSMTPWithCallback::DoCallback
	Description:
		Turn the state machine completion into a function callback.
*/
void CSMTPWithCallback::DoCallback(void)
{
	m_pCallback->SMTPCallback(m_iToken, m_result);
}

