/*
	File: CSMTP.H
	Description:
		A class for sending email messages via SMTP (RFC 821).
	Copyright 1995, David G. Roberts
*/

#ifndef _CSMTP_H
#define _CSMTP_H

#ifndef _CSOCKET_H
#include "csocket.h"
#endif
#ifndef _CCONNMGR_H
#include "cconnmgr.h"
#endif
#ifndef _CADDRLST_H
#include "caddrlst.h"
#endif
#include <iostream.h>

/*
	Class: CSMTP
	Description:
		Implementes the SMTP protocol.
*/
class CSMTP : private CConnMgrCallback, private CStreamCallback {
public:
	enum RESULT { RESULT_ABORT, RESULT_SUCCESS, RESULT_FAILURE };
protected:
	RESULT			m_result;
private:
	enum CONSTANTS { MAX_REPLY_SIZE = 512, MAX_LINE_SIZE = 1001 };
	enum EVENT { EV_NULLEVENT, EV_CONNECT, EV_CLOSE, EV_CODE_1,
		EV_CODE_2, EV_CODE_3, EV_CODE_4, EV_CODE_5, EV_BUFFERFREE,
		EV_ERROR, EV_ABORT};
	typedef BOOL (CSMTP::*PSTATEFUNC)(EVENT ev);

	CConnMgr		m_connmgr;
	CStreamSocket	m_sock;
	char			m_szReplyBuffer[MAX_REPLY_SIZE];
    int				m_iReplyLen;
	char			m_szLineBuffer[MAX_LINE_SIZE];
	BOOL			m_bBufferQueued;
	PSTATEFUNC		m_pState;
	int				m_iError;
	char *			m_pszHost;
	u_short			m_ushPort;
	istream *		m_pisMsgStream;
    char *			m_pszFrom;
	CAddrList *		m_palAddrList;
	BOOL			m_bDoCallback;

	// State machine interface.
	void Event(EVENT ev);
	void NextState(PSTATEFUNC pState) { m_pState = pState; };

	// State functions.
	BOOL Start(EVENT ev);
	BOOL AwaitConnection(EVENT ev);
	BOOL AwaitGreeting(EVENT ev);
    BOOL AwaitHelloResponse(EVENT ev);
	BOOL AwaitMailFromResponse(EVENT ev);
	BOOL AwaitRcptToResponse(EVENT ev);
	BOOL AwaitDataResponse(EVENT ev);
	BOOL SendData(EVENT ev);
    BOOL AwaitDataTransferResponse(EVENT ev);
	BOOL Quit(EVENT ev);
    BOOL AwaitQuitResponse(EVENT ev);
	BOOL Exit(EVENT ev);
	BOOL Stop(EVENT ev);

	// Virtual function called when the CSMTP state machine has
	// finished.
	virtual void DoCallback(void) = 0;

	// Aux functions.
	BOOL ReadReplyData(void);
	BOOL LineComplete(void);
    void ShrinkReceiveBuffer(void);

	virtual void ConnMgrCallback(int iToken,
		CConnMgrCallback::RESULT result);
	virtual void StreamCallback(int iToken, STREAMEVENT iEvent,
		int iError, char FAR * lpBuffer);

	// Eliminate the use of the copy constructor and operator=.
	CSMTP(const CSMTP&);
	operator=(const CSMTP&);
public:
	CSMTP() { m_connmgr.SetCallback(this, 0); };
	virtual ~CSMTP() { Abort(FALSE); };

	BOOL SendMessage(LPCSTR lpszHost, u_short ushPort,
		istream * pisMsgStream, LPCSTR lpszFrom,
		CAddrList * palAddrList);
	void Abort(BOOL bDoCallback);
};

/*
	Class: CSMTPWithWinMsg
	Description:
		Descendant of the CSMTP class.  Sends a window message
		back to the application when the message has been sent.
*/
class CSMTPWithWinMsg : public CSMTP {
private:
	HWND m_hwnd;
	UINT m_msg;
	int  m_iToken;

	virtual void DoCallback(void);
public:
	CSMTPWithWinMsg(HWND hwnd, UINT msg, int iToken) :
		m_hwnd(hwnd), m_msg(msg), m_iToken(iToken) {};
	~CSMTPWithWinMsg() {};
};

/*
	Class: CSMTPCallback
	Description:
		Used with CSMTPWithCallback.  An object wishing to receive
		a callback from a CSMTP object should inherit this class.
*/
class CSMTPCallback {
public:
	virtual void SMTPCallback(int iToken, CSMTP::RESULT result) = 0;
};

/*
	Class: CSMTPWithCallback
	Description:
		Descendant of the CSMTP class.  Provides a callback when
		the message has been sent.  An object wishing to receive
		the callback should inherit CSMTPCallback.
*/
class CSMTPWithCallback : public CSMTP {
private:
	CSMTPCallback * m_pCallback;
	int  m_iToken;

	virtual void DoCallback(void);
public:
	CSMTPWithCallback(CSMTPCallback * pCallback, int iToken) :
		m_pCallback(pCallback), m_iToken(iToken) {};
	~CSMTPWithCallback() {};
};
#endif
