/*
	File: CWINSOCK.CPP
	Description:
		A few classes and miscellaneous functions for dealing with
		Winsock in a more friendly manner.  Note that classes
		encapsulating sockets and addresses can be found in separate
		files.
	Copyright 1995, David G. Roberts
*/

#include <stdlib.h>
#include "cwinsock.h"
#include "myassert.h"

/* MODULE PRIVATE TYPES */

/*
	Class: CwParamList
	Description:
		A small class used to keep a linked list of pointers to
		CWinsockEventCallback objects, keyed with a WPARAM for lookup.
		CwParamList objects are used to keep lists of outstanding
		Winsock asynchronous callbacks so the private Winsock window
		can dispatch them to the appropriate objects.  Since both
		sockets and Winsock asynchronous handles don't have to be
		unique between each other, a second type value is used to
		distinguish between the two for lookup purposes.

		Note: This class could easily be implemented with a number of
		off-the-shelf container class libraries, but it was
		implemented here for cross-compiler compatibility.
*/
class CwParamList {
private:
	struct SListNode {
		SListNode *				lplnNext;
		SListNode *				lplnPrev;
		WPARAM					wParam;
		int						iType;
		CWinsockEventCallback *	lpecbCallback;
	};
	SListNode	m_lnHead;
	SListNode	m_lnTail;
	SListNode *	FindNode(WPARAM wParam, int iType);
public:
	BOOL Add(WPARAM wParam, int iType,
		CWinsockEventCallback * lpecbCallback);
	BOOL Remove(WPARAM wParam, int iType);
	BOOL Find(WPARAM wParam, int iType,
		CWinsockEventCallback * & lpecbCallback);
	CwParamList();
	~CwParamList();
};

/* MODULE VARIABLES */
static BOOL		sg_bWinsockStarted = FALSE;
static HWND		sg_hwndWSCEventWindow = NULL;
static char		sg_szWSCWindowClass[] = "WSAEventWindow";
CwParamList			sg_clWparamList;

/* MODULE PRIVATE DECLARATIONS */
static LRESULT CALLBACK WSCEventWndProc(HWND hwnd, UINT msg,
	WPARAM wParam, LPARAM lParam);

/*********************************************************************
	PUBLIC CLASS MEMBER FUNCTIONS
*********************************************************************/

/*
	Function: CWinsock::CWinsock
	Description:
		Constructor for the CWinsock class.  Initializes Winsock by
		calling WSAStartup and checking to make sure the Winsock DLL
		version is compatible with the application.
	Inputs:
		wVersionRequested -- The maximum Winsock version the app
			supports.
		wMinVersion -- The minimum Winsock version the app supports.
	Outputs:
		No direct outputs.
		The result of this function can be retrieved by calling
		CWinsock::LastError.
*/
CWinsock::CWinsock(WORD wVersionRequested, WORD wMinVersion)
{
	// Make sure we haven't started Winsock already.
	ASSERT(sg_bWinsockStarted == FALSE);

	// Start up Winsock.
	m_iLastError = WSAStartup(wVersionRequested,
		(LPWSADATA)&m_wsaData);

	// If we got an error, get out of here and indicate that we don't
	// need any cleanup in the destructor.
	if (m_iLastError != WSANOERROR) {
		// Indicate that we don't need to be cleaned up
		m_bNeedCleanup = FALSE;
		return;
	}

	// Indicate that we're up and running.
	sg_bWinsockStarted	= TRUE;
	m_bNeedCleanup		= TRUE;

	// See if the DLL and the app each support a common version.
	// Check to make sure that the version the DLL returns in wVersion
	// is at least enough to satisfy the applications needs.
	if (HIBYTE(m_wsaData.wVersion) < HIBYTE(wMinVersion) ||
		(HIBYTE(m_wsaData.wVersion) == HIBYTE(wMinVersion) &&
		LOBYTE(m_wsaData.wVersion) < LOBYTE(wMinVersion)) ) {
		// The DLL version is incompatible, so fake a
		// WSAVERNOTSUPPORTED error.
		m_iLastError = WSAVERNOTSUPPORTED;
	}
}

/*
	Function: CWinsock::~CWinsock
	Description:
		Destructor for CWinsock class.  The function calls WSACleanup
		to terminate use of a Winsock DLL.  If WSACleanup returns an
		error condition, the function reports it to the user using
		MessageBox.  Note that the application can't report
		the error itself since it's occuring in the destructor and
		there is no way to pass it to the app.  Further, if the
		CWinsock class instance is a global variable, then the
		destructor will be called as the app exits, so there is no
		chance for the app to handle the error.
	Inputs:
		None.
	Outputs:
		None.
*/
CWinsock::~CWinsock()
{
	char szErrorMsg[80];

	// If WSAStartup wasn't successful (because the required DLL
	// version wasn't compatible), then don't call WSACleanup.
	if (!m_bNeedCleanup)
		return;
	// Clean up Winsock and report any error that might occur.
	if (WSACleanup() == SOCKET_ERROR) {
		m_iLastError = WSAGetLastError();
		wsprintf(szErrorMsg,
			"Winsock reported error [%d] when shutting down.",
			m_iLastError);
		MessageBox(NULL, szErrorMsg, "Winsock Error",
			MB_OK | MB_ICONEXCLAMATION);
	}

	// Indicate that we're shut down.
	sg_bWinsockStarted = FALSE;
}

/*
	Function: CWinsockEventCallback::Register
	Description:
		Registers a CWinsockEventCallback object with the
		WSCEventWndProc.  This allows the WndProc to callback the
		object when a window message is received.
	Inputs:
		wParam -- wParam value to register a callback for.
		iType -- Equals WPARAM_TYPE_HANDLE if wParam is a Winsock task
			handle, or WPARAM_TYPE_SOCKET if wParam is a socket
			number.
	Outputs:
		None.
*/
void CWinsockEventCallback::Register(WPARAM wParam, int iType)
{
	if (!sg_clWparamList.Add(wParam, iType, this)) {
		// Some sort of error happened.
		WSCFatalError("Can't register CWinsockEventCallback object.");
	}
}

/*
	Function: CWinsockEventCallback::Deregister
	Description:
		Deregisters a CWinsockEventCallback object that had previously
		been registered with CWinsockEventCallback::Register.
	Inputs:
		wParam -- wParam value to register a callback for.
		iType -- Equals WPARAM_TYPE_HANDLE if wParam is a Winsock task
			handle, or WPARAM_TYPE_SOCKET if wParam is a socket
			number.
	Outputs:
		None.
*/
void CWinsockEventCallback::Deregister(WPARAM wParam, int iType)
{
	if (!sg_clWparamList.Remove(wParam, iType)) {
		// Some sort of error happened.
		WSCFatalError(
			"Can't deregister CWinsockEventCallback object.");
	}
}

/*********************************************************************
	PRIVATE CLASS MEMBER FUNCTIONS
*********************************************************************/

/*
	Function: CwParamList::CwParamList
	Description:
		Constructor of the CwParamList class.  Initializes internal
		list structure.
	Inputs:
		None.
	Outputs:
		None.
*/
CwParamList::CwParamList()
{
	m_lnHead.lplnNext = &m_lnTail;
	m_lnHead.lplnPrev = NULL;
	m_lnTail.lplnNext = NULL;
	m_lnTail.lplnPrev = &m_lnHead;
}

/*
	Function: CwParamList::~CwParamList
	Description:
		Destructor for the CwParamList class.  Takes care of deleting
		any orphan nodes remaining on in the list.
*/
CwParamList::~CwParamList()
{
	SListNode * lplnNext;
	SListNode * lplnCurrent;

	// Delete all orphaned nodes in list.
	lplnNext = m_lnHead.lplnNext;
	while (lplnNext != &m_lnTail) {
		lplnCurrent	= lplnNext;
		lplnNext	= lplnCurrent->lplnNext;
		delete lplnCurrent;
	};
}

/*
	Function: CwParamList::Add
	Description:
		Adds a node to a CwParamList.
	Inputs:
		wParam -- Key for the list management routines.  Typically
			either a Winsock asyncronous task handle or a socket
			descriptor.
		iType -- Equals WPARAM_TYPE_HANDLE if wParam is a Winsock task
			handle, or WPARAM_TYPE_SOCKET if wParam is a socket
			number.
		lpecbCallback -- A pointer to a CWinsockEventCallback that is
			stored in the list node and accessed according to the key.
	Outputs:
		The function returns TRUE if the information can be added to
		the list.  If there is no memory to allocate a new list node
		or a node with the same wParam key is already in the list,
		the function returns FALSE.
*/
BOOL CwParamList::Add(WPARAM wParam, int iType,
	CWinsockEventCallback * lpecbCallback)
{
	// Check some stuff
	ASSERT(lpecbCallback != NULL);

	// See if a node with that wParam/iType key already exists.
	SListNode * lplnNewNode = FindNode(wParam, iType);
    // If so, we have an error condition.
	if (lplnNewNode != NULL)
		return FALSE;

	// Create a new node and link it to the front of the list.
	lplnNewNode = new SListNode;
	if (lplnNewNode == NULL)
		return FALSE;
	// Fill in info.
	lplnNewNode->wParam			= wParam;
	lplnNewNode->iType			= iType;
	lplnNewNode->lpecbCallback	= lpecbCallback;
	// Set links in new node.
	lplnNewNode->lplnNext		= m_lnHead.lplnNext;
	lplnNewNode->lplnPrev		= &m_lnHead;
	// Set previous pointer in following node.
	m_lnHead.lplnNext->lplnPrev	= lplnNewNode;
	// Set next pointer in head node.
	m_lnHead.lplnNext			= lplnNewNode;

	return TRUE;
}

/*
	Function: CwParamList::Remove
	Description:
		Removes a node from the CwParamList.
	Inputs:
		wParam -- Key indicating which node to remove.
		iType -- Equals WPARAM_TYPE_HANDLE if wParam is a Winsock task
			handle, or WPARAM_TYPE_SOCKET if wParam is a socket
			number.
	Outputs:
		The function returns TRUE if the node can be removed.
		The function returns FALSE if a node with the specified key
		cannot be located in the list.
*/
BOOL CwParamList::Remove(WPARAM wParam, int iType)
{
	// See if we can find the node.
	SListNode *	lplnNode = FindNode(wParam, iType);
	if (lplnNode == NULL)
		return FALSE;
	// Set pointers to bypass the indicated node.
	lplnNode->lplnPrev->lplnNext	= lplnNode->lplnNext;
	lplnNode->lplnNext->lplnPrev	= lplnNode->lplnPrev;
	// Delete the node.
	delete lplnNode;
	return TRUE;
}

/*
	Function: CwParamList::Find
	Description:
		Find a node in the list and return its lpecbCallback info.
		This function is little more than a wrapper for
		CwParamList::FindNode but it keeps the internal structure of
		the list data hidden.
	Inputs:
		wParam -- Key for finding the info.
		iType -- Equals WPARAM_TYPE_HANDLE if wParam is a Winsock task
			handle, or WPARAM_TYPE_SOCKET if wParam is a socket
			number.
		lplpecbCallback -- Reference to a pointer to a callback
			object.  The pointer will be filled in with the data from
			the found node.
	Outputs:
		The function returns TRUE if it is successful.
		The function returns FALSE if the wParam key isn't valid and a
		corresponding node can't be found.
*/
BOOL CwParamList::Find(WPARAM wParam, int iType,
	CWinsockEventCallback * & lpecbCallback)
{
	// Find the node.
	SListNode * lplnNode = FindNode(wParam, iType);
	if (lplnNode == NULL)
		return FALSE;
	// Copy the pointer to the callback object into the output
	// parameter.
	lpecbCallback = lplnNode->lpecbCallback;
	return TRUE;
}

/*
	Function: CwParamList::FindNode
	Description:
		Finds a list node in the linked list and returns a pointer
		to it.
	Input:
		wParam -- Search key of the node to find.
		iType -- Equals WPARAM_TYPE_HANDLE if wParam is a Winsock task
			handle, or WPARAM_TYPE_SOCKET if wParam is a socket
			number.
	Output:
		Returns a FAR pointer to a CwParamList::SListNode if
		successful.  Returns NULL if the given wParam key can't be
		located in the list.
*/
CwParamList::SListNode * CwParamList::FindNode(WPARAM wParam,
	int iType)
{
	// Start at the first node after the head.
	SListNode * lplnNode = m_lnHead.lplnNext;

	// Walk the list until we get to the tail.
	while (lplnNode != &m_lnTail) {
		// If we have a match, return a pointer to the node.
		if (lplnNode->wParam == wParam && lplnNode->iType == iType)
			return lplnNode;
		// Move on...
		lplnNode = lplnNode->lplnNext;
	}
	return NULL;
}

/*********************************************************************
	MISC FUNCTIONS
*********************************************************************/

/*
	Function: WSCCreateEventWindow
	Description:
		Creates a window to receive Winsock events.
	Inputs:
		hInstance -- The instance handle of the application.
	Outputs:
		BOOL -- Returns TRUE if the function was successful, and FALSE
			if an error occurred.
*/
BOOL WSCCreateEventWindow(HINSTANCE hInstance,
	HINSTANCE hPrevInstance)
{
	// Check arguments.
    ASSERT(hInstance != NULL);

	// We can't do this twice.
	if (sg_hwndWSCEventWindow != NULL)
		return FALSE;

	if (!hPrevInstance) {
		// Create the window class.
		WNDCLASS wc;
		wc.style			= 0;
		wc.lpfnWndProc		= WSCEventWndProc;
		wc.cbClsExtra		= 0;
		wc.cbWndExtra		= 0;
		wc.hInstance		= hInstance;
		wc.hIcon			= NULL;
		wc.hCursor			= NULL;
		wc.hbrBackground	= NULL;
		wc.lpszMenuName		= NULL;
		wc.lpszClassName	= sg_szWSCWindowClass;

		BOOL bResult = RegisterClass(&wc);
		if (!bResult)
			return FALSE;
	}

	// Create the window itself.  Note that the window is by default
	// invisible and we simply never show it with ShowWindow.
	sg_hwndWSCEventWindow = CreateWindow(
		sg_szWSCWindowClass,			// window class name
		"",								// window title text
		0,								// window style
		CW_USEDEFAULT, CW_USEDEFAULT,	// position = default
		CW_USEDEFAULT, CW_USEDEFAULT,	// size = default
		NULL,							// no parent window
		NULL,							// menu = window class default
		hInstance,						// window owner
		NULL);							// no additional creation data

	if (sg_hwndWSCEventWindow == NULL)
		return FALSE;
	else
		return TRUE;
}

/*
	Function: WSCGetEventWindow
	Description:
		Returns the handle of the window created by
		WSCCreateEventWindow to receive Winsock events.  This is
		typically passed to a WSAAsyncGetXByY or WSAAsyncSelect call.
*/
HWND WSCGetEventWindow(void)
{
	return sg_hwndWSCEventWindow;
}

/*
	Function: ReportWSAError
	Description:
		Generates a message box in response to a Winsock error.  The
		message box contains a bit of text passed to the function as
		well as a textual description of the error corresponding to
		the iError argument.  HWnd is used by MessageBox and the
		parent window and hInstance is used by LoadString to get the
		Winsock error text string resource.
	Inputs:
		lpszMsg -- A pointer to a text message to be printed before
			the error message.
		iError -- The Winsock error number.
		hwnd -- Window to use as the message box parent.
		hInstance -- The application instance handle.  This is used to
			read the error message string from the application's
			string resource table using LoadString.  This cannot be
			NULL.
	Outputs:
		None.
*/
void ReportWSAError(LPCSTR lpszMsg, int iError, HWND hwnd,
	HINSTANCE hInstance)
{
	// Make sure arguments are cool.
	ASSERT(lpszMsg != NULL);
	ASSERT(hInstance != NULL);

	// Convert iError == 0 to WSABASEERR.
	if (iError == 0)
		iError = WSABASEERR;

	// Retrieve the error message text.  This may fail if Windows is
	// out of memory (which could also be the cause of the Winsock
    // error).
	char szErrorMsg[81];
	int cErrMsgLen = LoadString(hInstance, iError, szErrorMsg,
		sizeof(szErrorMsg)) - 1; // -1 for the '\0' terminator.
	// If we couldn't find the error message text, fill in something.
	if (cErrMsgLen == 0)
		wsprintf(szErrorMsg, "[%d] Unknown Error", iError);

	// Create the final error message string.
	int cMsgLen = lstrlen(lpszMsg);
	// Need 80 bytes for message, 80 bytes for error text, 2 bytes for
    // CRLF, 1 byte for period, 1 byte for '\0'.
	const int FINAL_MSG_SIZE = (80 + 80 + 2 + 1 + 1);
	char szFinalMsg[FINAL_MSG_SIZE];
	// Copy message to final string, truncating to make room for
	// all the other text, if necessary.
	if ((cMsgLen + cErrMsgLen + 4) > FINAL_MSG_SIZE) {
		lstrcpyn(szFinalMsg, lpszMsg, FINAL_MSG_SIZE - 4 -
			cErrMsgLen);
	}
	else {
		lstrcpy(szFinalMsg, lpszMsg);
	}
	lstrcat(szFinalMsg, "\r\n");
	lstrcat(szFinalMsg, szErrorMsg);
	lstrcat(szFinalMsg, ".");

	// Display it using MessageBox.
	MessageBox(hwnd, szFinalMsg, "Winsock Error",
		MB_OK | MB_ICONEXCLAMATION);
}

void WSCFatalError(LPCSTR lpszMsg)
{
	MessageBox(NULL, lpszMsg, "Winsock Classes Fatal Error",
		MB_OK | MB_ICONEXCLAMATION);
	exit(1);
}

/*********************************************************************
	MODULE PRIVATE FUNCTIONS
*********************************************************************/

/*
	Function: WSCEventWndProc
	Description:
		This function is used to receive all Winsock related event
		messages.  The function corresponds to a hidden, private
		window that CWinsock allocates.  When a message arrives,
		WSCEventWndProc finds a callback routine corresponding to the
		message and dispatches the event to that routine to be
		handled.
	Inputs:
		The standard set of WndProc messages: hwnd, msg, wParam,
		lParam.
	Outputs:
		Standard WndProc LRESULT.
*/
static LRESULT CALLBACK WSCEventWndProc(HWND hwnd, UINT msg,
	WPARAM wParam, LPARAM lParam)
{
	CWinsockEventCallback * lpecCallback;

	switch (msg) {
		// Since all these messages all use task handles, we can unify
		// their processing.
		case WM_WSAGETHOSTBYADDR:
		case WM_WSAGETHOSTBYNAME:
		case WM_WSAGETPROTOBYNAME:
		case WM_WSAGETPROTOBYNUMBER:
		case WM_WSAGETSERVBYNAME:
		case WM_WSAGETSERVBYPORT:
			// Look up task handle.
			if (sg_clWparamList.Find(wParam, WPARAM_TYPE_HANDLE,
				lpecCallback)) {
				// Make sure we aren't off in the weeds.
				ASSERT(lpecCallback != NULL);
				// Dispatch control to appropriate callback function.
				lpecCallback->WinsockEventCallback(hwnd, msg,
					wParam, lParam);
				return 0;
			}
			else {
				// We didn't find an entry for the task handle.
				return 0;
			}
		case WM_WSAASYNCSELECT:
			// Look up socket.
			if (sg_clWparamList.Find(wParam, WPARAM_TYPE_SOCKET,
				lpecCallback)) {
                // Make sure we aren't off in the weeds.
				ASSERT(lpecCallback != NULL);
				// Dispatch control to appropriate callback function.
				lpecCallback->WinsockEventCallback(hwnd, msg,
					wParam, lParam);
				return 0;
			}
			else {
				// We didn't find an entry for the socket.
				return 0;
			}
		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
	}
}

