/*
	File: CWINSOCK.H
	Description:
		Header file for CWINSOCK.CPP.

		A few classes and miscellaneous functions for dealing with
		Winsock in a more friendly manner.  Note that classes
		encapsulating sockets and addresses can be found in separate
		files.
	Copyright 1995, David G. Roberts
*/

#ifndef _CWINSOCK_H
#define _CWINSOCK_H

#include <windows.h>
#include "winsock.h"

/* CONSTANTS */
#define WPARAM_TYPE_HANDLE		0
#define WPARAM_TYPE_SOCKET		1

// I can't believe that Winsock doesn't define this on its own.
#define WSANOERROR	0

// Private window messages.  Note that we don't have to keep these
// unique, even within the application, since they'll just be sent to
// our own window proc.
#define WM_WSAGETHOSTBYADDR		(WM_USER + 0)
#define WM_WSAGETHOSTBYNAME		(WM_USER + 1)
#define WM_WSAGETPROTOBYNAME	(WM_USER + 2)
#define WM_WSAGETPROTOBYNUMBER	(WM_USER + 3)
#define WM_WSAGETSERVBYNAME		(WM_USER + 4)
#define WM_WSAGETSERVBYPORT		(WM_USER + 5)
#define WM_WSAASYNCSELECT		(WM_USER + 6)

/* MACROS */
// A quick macro to denote an unused argument parameter.  This simply
// references the parameter in its own expression, which the optimizer
// will eliminate.
#define UNUSED_PARAMETER(x)	x

/* MISC FUNCTIONS */
BOOL WSCCreateEventWindow(HINSTANCE hInstance,
	HINSTANCE hPrevInstance);
HWND WSCGetEventWindow(void);
void ReportWSAError(LPCSTR lpszMsg, int iError, HWND hwnd,
	HINSTANCE hInstance);
void WSCFatalError(LPCSTR lpszMsg);

/*
	Class: CWinsock
	Description:
		This class is used to initialize Winsock in an application
		and retrieve information about the particular Winsock
		implementation running under it.  A single instance of this
		class should be created as a global or as a local in WinMain.
		The constructor of this class calls WSAStartup and the
		destructor calls WSACleanup.  This ensures that the calls
		are balanced without the programmer having to remember to call
		WSACleanup at program exit.
*/
class CWinsock {
private:
	WSADATA	m_wsaData;			// Holds results of WSAStartup call
	int		m_iLastError;		// Last error from Winsock call
	BOOL	m_bNeedCleanup;		// Indicates WSAStartup was successful
								//   and the destructor should call
								//   WSACleanup
public:
	CWinsock(WORD wVersionRequested, WORD wMinVersion);
	~CWinsock();
	int LastError(void) { return m_iLastError; };
	LPWSADATA GetWSAData(void) { return (LPWSADATA)&m_wsaData; };
};

/*
	Class: CWinsockEventCallback
	Description:
		This class is used to receive a callback whenever a Winsock
		window message is received by the Winsock event window.
		Objects needing to get a callback should inherit this
		class and override the WinsockEventCallback member function.
*/
class CWinsockEventCallback {
protected:
	void Register(WPARAM wParam, int iType);
	void Deregister(WPARAM wParam, int iType);
	HWND GetEventWindow(void) { return WSCGetEventWindow(); };
public:
	// Override this with the desired functionality.
	virtual void WinsockEventCallback(HWND hwnd, UINT msg,
		WPARAM wParam, LPARAM lParam) = 0;
};

#endif // #ifndef _CWINSOCK_H
