/*
	File: CCONNMGR.CPP
	Description:
		A stream socket connection manager class.
	Copyright 1995, David G. Roberts
*/

#include "cconnmgr.h"
#include "tracer.h"

void CConnMgr::AsyncDBCallback(int iToken, int iError)
{
	TracerEntryExit tracer("CConnMgr::AsyncDBCallback");
	UNUSED_PARAMETER(iToken);	// Eliminate the compiler warning.

	TRACEOUT(Tracer::TR_MSG, "iToken = " << iToken << ", iError = " << iError);

	m_iError = iError;
	Event(LOOKUPDONE);
}

void CConnMgr::StreamCallback(int iToken, STREAMEVENT iEvent, int iError,
	char FAR *lpBuffer)
{
	TracerEntryExit tracer("CConnMgr::StreamCallback");
	UNUSED_PARAMETER(iToken);	// Eliminate the compiler warnings.
    UNUSED_PARAMETER(lpBuffer);

	m_iError = iError;
	if (iEvent == CStreamCallback::SE_CONNECT) {
		Event(CONNECT);
	}
	else {
		// Error!  An unexpected event.
		m_pSocket->Close();
		WSCFatalError(
			"Unexpected socket event in CConnMgr::StreamCallback");
	}

}

void CConnMgr::Event(EVENT ev)
{
	TracerEntryExit tracer("CConnMgr::Event");
	while ((this->*m_pState)(ev)) {
		ev = NULLEVENT;
	};
}

BOOL CConnMgr::Connect(LPCSTR lpszHost, u_short ushPort,
	CStreamSocket * pSocket, CStreamCallback * pCallback, int iToken)
{
	TracerEntryExit tracer("CConnMgr::Connect");
	if (m_pState != &CConnMgr::Stop) {
		// We've already got something else going on.  You'll have to
		// abort first and then retry.
		return FALSE;
	}

	// Initialize a whole bunch of variables.
	m_result	= CConnMgrCallback::INVALID_RESULT;
	m_pszHost	= new char[lstrlen(lpszHost) + 1];
	lstrcpy(m_pszHost, lpszHost);
	m_ushPort			= ushPort;
	m_pSocket			= pSocket;
	m_pStreamCallback	= pCallback;
	m_iStreamToken		= iToken;
	m_bDoCallback		= TRUE;

	// Hook all events on the stream.
	m_pSocket->SetCallback(this, 0);
	// Make sure the stream is closed.
    m_pSocket->Close();

	// Start the machine.
	NextState(&CConnMgr::Start);
	Event(NULLEVENT);

    return TRUE;
}

void CConnMgr::Abort(BOOL bDoCallback)
{
	TracerEntryExit tracer("CConnMgr::Abort");
	m_bDoCallback = bDoCallback;

	Event(ABORT);
}

BOOL CConnMgr::Start(EVENT ev)
{
	TracerEntryExit tracer("CConnMgr::Start");
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

	// Note: this little test here won't work when IPv6 is deployed and
	// IP addresses move to 128 bits.  We're assuming IP addresses are
    // 32 bits wide and can fit in a u_long.
	u_long addr = inet_addr(m_pszHost);
	if (addr == INADDR_NONE) {
		// Addr must be a host name. Kick off a lookup.
		m_bDotted = FALSE;
		m_iError = m_host.StartLookup(m_pszHost, this, 0);
		if (m_iError != WSANOERROR) {
			m_result = CConnMgrCallback::ERR_NOTCONNECTED;
			NextState(&CConnMgr::Exit);
		}
		else {
			NextState(&CConnMgr::AwaitLookup);
		}
	}
	else {
		// Addr was in dotted decimal notation.
        m_bDotted = TRUE;
		addr = ntohl(addr);
		m_saAddr.Addr(addr);
		NextState(&CConnMgr::MakeConnection);
	}

	return TRUE;
}

BOOL CConnMgr::AwaitLookup(EVENT ev)
{
	TracerEntryExit tracer("CConnMgr::AwaitLookup");
	// We're only expecting a LOOKUPDONE or an ABORT event.
	if (ev != LOOKUPDONE && ev != ABORT)
		return FALSE;

	// If we're being asked to abort, then stop the lookup and exit.
	if (ev == ABORT) {
    	m_host.StopLookup();
		m_result = CConnMgrCallback::ABORTED;
		NextState(&CConnMgr::Exit);
		return TRUE;
	}

	// See if the lookup produced an error.
	if (m_iError != WSANOERROR) {
		// Something didn't go right.
		m_result = CConnMgrCallback::ERR_NOTCONNECTED;
		NextState(&CConnMgr::Exit);
	}
	else {
		// Set our address to the first address from the host lookup.
		BOOL bResult;
		m_saAddr = m_host.FirstAddr(&bResult);
		if (bResult != TRUE) {
			// Can't access the address for some strange reason.
			WSCFatalError(
		"Can't access address following lookup in CConnMgr::AwaitLookup");
		}
		else {
			NextState(&CConnMgr::MakeConnection);
		}
	}

	return TRUE;
}

BOOL CConnMgr::MakeConnection(EVENT ev)
{
	TracerEntryExit tracer("CConnMgr::MakeConnection");
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

    // The state transitions on all events.

	// Okay, we've got our address.  Fill in the port and try to make
	// a connection.
	m_saAddr.Port(m_ushPort);
	m_iError = m_pSocket->Connect(m_saAddr);
	if (m_iError != WSANOERROR && m_iError != WSAEWOULDBLOCK) {
		// Something went wrong.
		m_pSocket->Close();
		m_result = CConnMgrCallback::ERR_NOTCONNECTED;
		NextState(&CConnMgr::Exit);
	}
	else {
		NextState(&CConnMgr::AwaitConnection);
	}

    return TRUE;
}

BOOL CConnMgr::AwaitConnection(EVENT ev)
{
	TracerEntryExit tracer("CConnMgr::AwaitConnection");
	// Accept only connection events and abort events.
	if (ev != CONNECT && ev != ABORT)
		return FALSE;

	// Handle an abort request.
	if (ev == ABORT) {
		m_pSocket->Close();
		m_result = CConnMgrCallback::ABORTED;
		NextState(&CConnMgr::Exit);
		return TRUE;
	}

	// See if we connected or had an error.
	if (m_iError == 0) {
		// Everything was okay.  We're now connected.  Move on.
		m_result = CConnMgrCallback::OK_CONNECTED;
		NextState(&CConnMgr::Exit);
		return TRUE;
	}
	// Okay, something went wrong.  See what it was.
	if (m_iError == WSAENETUNREACH || m_iError == WSAETIMEDOUT) {
		// Perhaps we can't reach the foreign host on the primary address.
		// If the foreign host has multiple addresses, try one of the
		// others until we're out of addresses.
		BOOL bResult;
		m_saAddr = m_host.NextAddr(&bResult);
		if (!m_bDotted && bResult == TRUE) {
			// We didn't have a dotted address to begin with and
			// the host has more addresses left, so try again.
            m_pSocket->Close();
			NextState(&CConnMgr::MakeConnection);
			return TRUE;
		}
	}
	// No more addresses left to try, so signal the error to the user.
	m_result = CConnMgrCallback::ERR_NOTCONNECTED;
	m_pSocket->Close();
	NextState(&CConnMgr::Exit);
	return TRUE;
}

BOOL CConnMgr::Exit(EVENT ev)
{
	TracerEntryExit tracer("CConnMgr::Exit");
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

	// Free up the stuff we allocated when we started.
	delete[] m_pszHost;

	// Make all events on the stream now go to the object specified.
	m_pSocket->SetCallback(m_pStreamCallback, m_iStreamToken);
	// Now, invoke the callback with the connection attempt result.
	if (m_bDoCallback) {
		m_pManagerCallback->ConnMgrCallback(m_iManagerToken, m_result);
	}

	NextState(&CConnMgr::Stop);
	return TRUE;
}

BOOL CConnMgr::Stop(EVENT ev)
{
	TracerEntryExit tracer("CConnMgr::Stop");
	UNUSED_PARAMETER(ev);
	// Simply ignore all events.
	return FALSE;
}

