/*
	File: CMIME.CPP
	Description:
		A bunch of functions for parsing and manipulating MIME
		messages and body parts.
	Copyright 1995, David G. Roberts

    4/18/96: Modified CMIMEBodyPart::ParseMultipart to deal with
	         the case where the first multipart boundary occurs
			 on the first line of body text with only a single
			 blank line separating the message headers from the
			 boundary line. In particular, Eudora generates
			 such messages.
*/

#include <ctype.h>		// For tolower
#include "cmime.h"
#include "rfc822.h"
#include "utility.h"
#include "myassert.h"

#define UNUSED_PARAMETER(x) x

/*********************************************************************
	Class: CMIMEParamTuple
*********************************************************************/

/*
	Function: CMIMEParamTuple::CMIMEParamTuple
	Description:
		Copy constructor.  Make sure that we do more than a simple
		member-wise copy.  We need to actually duplicate the
		strings in the tuple.
*/
CMIMEParamTuple::CMIMEParamTuple(const CMIMEParamTuple& tuple) :
	m_pszParam(NULL), m_pszValue(NULL)
{
	Param(tuple.m_pszParam);
	Value(tuple.m_pszValue);
}

/*
	Function: CMIMEParamTuple::~CMIMEParamTuple
	Description:
		Delete the param and value strings before the CMIMEParamTuple
		object is destroyed itself.
*/
CMIMEParamTuple::~CMIMEParamTuple()
{
	delete[] m_pszParam;
	delete[] m_pszValue;
}

/*
	Function: CMIMEParamTuple::Param
	Description:
		Makes a local copy of the parameter string and stores it
		in the CMIMEParamTuple.
*/
void CMIMEParamTuple::Param(const char * pszParam)
{
	delete[] m_pszParam;
	if (pszParam != NULL) {
		m_pszParam = new char[lstrlen(pszParam) + 1];
		lstrcpy(m_pszParam, pszParam);
	}
	else
		m_pszParam = NULL;
}

/*
	Function: CMIMEParamTuple::Value
	Description:
		Makes a local copy of the value string and stores it
		in the CMIMEParamTuple.
*/
void CMIMEParamTuple::Value(const char * pszValue)
{
	delete[] m_pszValue;
	if (pszValue != NULL) {
		m_pszValue = new char[lstrlen(pszValue) + 1];
		lstrcpy(m_pszValue, pszValue);
	}
	else
    	m_pszValue = NULL;
}

/*
	Function: CMIMEParamTuple::operator=
	Description:
		The overloaded assignment operator for the CMIMEParamTuple
		class.
*/
CMIMEParamTuple& CMIMEParamTuple::operator=
	(const CMIMEParamTuple& tuple)
{
	if (this != &tuple) {
		delete[] m_pszParam;
		delete[] m_pszValue;
		Param(tuple.m_pszParam);
        Value(tuple.m_pszValue);
	}
	return *this;
}

/*********************************************************************
	Class: CMIMEParamList
*********************************************************************/

/*
	Function: CMIMEParamList::~CMIMEParamList
	Description:
		Deletes the pararm tuples before the list is destroyed.
*/
CMIMEParamList::~CMIMEParamList()
{
	CMIMEParamTuple * pTuple = First();
	while (pTuple != NULL) {
		delete pTuple;
		pTuple = Next();
	}
}

/*
	Function: CMIMEParamList::Compare
	Description:
		Simply returns 0.  A CMIMEParamList is unordered.  Returning
		0 causes an Insert operation to stop at the first element,
		which results in the element being stored at the beginning
		of the list.
*/
int CMIMEParamList::Compare(void * pvThing1, void * pvThing2)
{
	UNUSED_PARAMETER(pvThing1);
	UNUSED_PARAMETER(pvThing2);

	return 0;
}

/*
	Function: CMIMEParamList::DeleteData
	Description:
		Deletes an element of the list, in this case, a
		CMIMEParamTuple object.
*/
void CMIMEParamList::DeleteData(void * pvThing)
{
	delete ((CMIMEParamTuple *) pvThing);
}

/*
	Function: CMIMEParamList::Add
	Description:
		Add a tuple to the end of the list.  A copy of the tuple
		is made before it is added.
*/
void CMIMEParamList::Add(CMIMEParamTuple * pTuple)
{
	CMIMEParamTuple * pNewTuple = new CMIMEParamTuple(*pTuple);
	CList::Add(pNewTuple);
}

/*
	Function: CMIMEParamList::Insert
	Description:
		Insert a tuple into the list.  Because a CMIMEParamList is
		unordered, the tuple is inserted at the front of the list.
		A copy of the tuple is made before it is added.
*/
void CMIMEParamList::Insert(CMIMEParamTuple * pTuple)
{
	CMIMEParamTuple * pNewTuple = new CMIMEParamTuple(*pTuple);
	CList::Insert(pNewTuple);
}

/*********************************************************************
	Class: CMIMEBodyPart
*********************************************************************/

/*
	Function: CMIMEBodyPart::CMIMEBodyPart
	Description:
		Constructor for CMIMEBodyPart.  Just a huge initialization
		list for this one.
*/
CMIMEBodyPart::CMIMEBodyPart(CMIMEBodyPart * pParent,
	char * pszFilename, long lStart, long lEnd) :
	m_pszFilename(DuplicateString(pszFilename)),
	m_pszType(NULL), m_pszSubType(NULL),
	m_pParamList(new CMIMEParamList), m_pszEncoding(NULL),
	m_lStart(lStart), m_lEnd(lEnd),
    m_pParent(pParent), m_pFirstChild(NULL), m_pSibling(NULL)
{
}

/*
	Function: CMIMEBodyPart::~CMIMEBodyPart
	Description:
		Delete all the memory we allocated during this process.
*/
CMIMEBodyPart::~CMIMEBodyPart()
{
	Close();

	delete[]	m_pszFilename;
	delete[]	m_pszType;
	delete[]	m_pszSubType;
	delete		m_pParamList;
	delete[]	m_pszEncoding;

	// Delete all the children.
	CMIMEBodyPart * pChild, *pNextChild;
	pChild = m_pFirstChild;
	while (pChild != NULL) {
		pNextChild = pChild->m_pSibling;
		delete pChild;
		pChild = pNextChild;
	}
}

/*
	Function: CMIMEBodyPart::Open
	Description:
		Open the body part for parsing and reading.
*/
BOOL CMIMEBodyPart::Open(void)
{
	// Simply open the filemap object.
	return (m_filemap.Open(m_pszFilename) == CMM_NO_ERROR) ?
		TRUE : FALSE;
}

/*
	Function: CMIMEBodyPart::Close
	Description:
		Close the body part.  No more parsing or reading should
		be performed after this point.
*/
void CMIMEBodyPart::Close(void)
{
	// Simply close the filemap object.
    m_filemap.Close();
}

/*
	Function: CMIMEBodyPart::Parse
	Description:
		Parse the open body part.  The function first parses the
		headers.  If the headers indicate that the body part is
		multipart, then find the delimiters between the subparts and
		parse the subparts recursively.
*/
int CMIMEBodyPart::Parse(void)
{
	if (!m_filemap.IsOpen()) {
		return FALSE;
	}

	// Correct the values of start and end if they're set to
	// the defaults.
	if (m_lStart < 0)
		m_lStart = 0;
	if (m_lEnd < 0) {
		// Set lEnd to number of bytes in the file.
		m_lEnd = m_filemap.FileLength();
	}

	// Get all the headers of interest.
	ParseHeaders();

	// If this is a multipart message, parse it recursively.
	if (lstrcmpi(m_pszType, "multipart") == 0)
		if (ParseMultipart() != TRUE)
        	return FALSE;

	return TRUE;
}

/*
	Function: CMIMEBodyPart::ParseMultipart
	Description:
		Use the specified boundary to locate the various body
		parts of the multipart message.  Create a CMIMEBodyPart
		for each body part and recursively call CMIMEBodyPart::Parse
		to parse the parts.  Returns TRUE if no errors occurred,
		or FALSE otherwise.
*/
BOOL CMIMEBodyPart::ParseMultipart(void)
{
	// Find the "boundary" parameter tuple.
	CMIMEParamTuple * pTuple = m_pParamList->First();
	while (pTuple != NULL) {
		if (lstrcmpi(pTuple->Param(), "boundary") == 0)
			break;
		pTuple = m_pParamList->Next();
	}
	if (pTuple == NULL)
		return FALSE;

	// Create the real boundary string.
	int nBoundaryLen = lstrlen(pTuple->Value());
	char * pszBoundary = new char[nBoundaryLen + 4 + 1];
	lstrcpy(pszBoundary, "\r\n--");
	lstrcat(pszBoundary, pTuple->Value());

	// 4/18/96: Added decrement of m_lCurrent by 2 to
	// account for situation where the boundary is the first
	// line of the body. The decrement allows the leading
	// "\r\n" (CRLF) of the search string to match the single
	// blank line separating the headers from the body. If this
	// was not done, the first boundary would not be found when
	// it occurred as the first line.
	m_lCurrent -= 2;
	// Search for the first boundary.
	long lBoundaryStart = StringSearch(pszBoundary);
	BOOL bFoundStartBoundary = FALSE;

	long lBodyPartStart, lBodyPartEnd;
	while (lBoundaryStart != -1) {
		// Okay, we found a boundary.  Is it a terminating boundary.
		BOOL bEndingBoundary = FALSE;
		if (m_filemap[m_lCurrent] == '-' &&
			m_filemap[m_lCurrent + 1] == '-') {
			bEndingBoundary = TRUE;
		}
		// Find the end of the line.
		long lBoundaryEnd = FindCRLF();
		if (lBoundaryEnd < 0) {
			delete[] pszBoundary;
			return FALSE;
		}
		lBoundaryEnd += 2;  // Jump over CRLF.
		if (lBoundaryEnd >= m_lEnd) {
			delete[] pszBoundary;
			return FALSE;
		}

		if (!bFoundStartBoundary) {
        	// We haven't seen a starting boundary before.
        	// See if we got an end before we got a start.
			if (bEndingBoundary) {
				delete[] pszBoundary;
				return FALSE;
			}
			lBodyPartStart = lBoundaryEnd;
			bFoundStartBoundary = TRUE;
		}
		else {
			// If we've already found a boundary before, then
			// this was a separating boundary or the terminating
			// boundary.
			lBodyPartEnd = lBoundaryStart;
			// Create a child body part and add it to the heirarchy.
			CMIMEBodyPart * pBodyPart = new CMIMEBodyPart(this,
				m_pszFilename, lBodyPartStart, lBodyPartEnd);
			if (m_pFirstChild == NULL) {
				m_pFirstChild = pBodyPart;
			}
			else {
				// Add it to the end of the sibling chain.
				CMIMEBodyPart * pChild = m_pFirstChild;
				while (pChild->m_pSibling != NULL)
					pChild = pChild->m_pSibling;
				pChild->m_pSibling = pBodyPart;
			}

			// The next body part starts right after this boundary.
			lBodyPartStart = lBoundaryEnd;

			// If this was the terminating boundary, then we're
			// done.
			if (bEndingBoundary)
				break;
		}

		// Find the next boundary.
		lBoundaryStart = StringSearch(pszBoundary);
	}

	delete[] pszBoundary;
	if (lBoundaryStart == -1)
		return FALSE;

	// Now we've found all the parts.  Traverse the child list
	// and have each child body part parse itself.
	CMIMEBodyPart * pChild = m_pFirstChild;
	while (pChild != NULL) {
		if (!(pChild->Open())) return FALSE;
		if (!(pChild->Parse())) return FALSE;
		pChild->Close();
        pChild = pChild->m_pSibling;
	}

	return TRUE;
}

/*
	Function: CMIMEBodyPart::ParseHeaders
	Description:
		Parse the body part headers to determine the content type
		and encoding of the message.
*/
BOOL CMIMEBodyPart::ParseHeaders(void)
{
	char * pszHeader;
    BOOL bResult;
	BOOL bFoundContentType = FALSE;
    BOOL bFoundEncoding = FALSE;

	// Start looking for headers at the start of the current
    // body part range.
	m_lCurrent = m_lStart;

	while ((pszHeader = GetNextHeader()) != NULL) {
		if (lstrcmp(pszHeader, "") == 0) {
			delete[] pszHeader;
			break;
		}
		else if (HeaderCompare(pszHeader, "MIME-Version")) {
			bResult = ParseMIMEVersion(pszHeader);
			if (bResult != TRUE) {
				delete[] pszHeader;
				return FALSE;
			}
		}
		else if (HeaderCompare(pszHeader, "Content-Type")) {
			bResult = ParseContentType(pszHeader);
			if (bResult != TRUE) {
				delete[] pszHeader;
				return FALSE;
			}
			bFoundContentType = TRUE;
		}
		else if
			(HeaderCompare(pszHeader, "Content-Transfer-Encoding")) {
			bResult = ParseContentTransferEncoding(pszHeader);
			if (bResult != TRUE) {
				delete[] pszHeader;
				return FALSE;
			}
			bFoundEncoding = TRUE;
		}
		delete[] pszHeader;
	}

	// Fill in the defaults if the appropriate headers were missing.
	if (!bFoundContentType) {
		m_pszType = DuplicateString("text");
		m_pszSubType = DuplicateString("plain");
	}
	if (!bFoundEncoding) {
		m_pszEncoding = DuplicateString("7bit");
	}

	return TRUE;
}

/*
	Function: CMIMEBodyPart::GetNextHeader
	Description:
		Find the next header in the file starting at m_lCurrent.
		Returns a newly created string with the header in it.
		The caller is responsible for deleteing the string when
		it is finished with the header.  The returned header has
		been unfolded and any trailing whitespace has been removed.
		All internal CRLFs have been converted to a single space.
		The final CRLF has been removed.
*/
char * CMIMEBodyPart::GetNextHeader(void)
{
	long lEndPos;
	long lStartPos = m_lCurrent;

	// Find the first CRLF pair that doesn't have linear whitespace
	// following it.
	do {
		lEndPos = FindCRLF();
		// Move the current position up.
		if (lEndPos >= 0) {
			m_lCurrent = lEndPos + 2;
		}
	} while (lEndPos >= 0 && IsLWSP(m_filemap[lEndPos + 2]));

    // If we couldn't find CRLF or an error occurred.
	if (lEndPos < 0)
		return NULL;

	// Push lEndPos to indicate the character following the CRLF.
	lEndPos += 2;

	// Calculate the header length.
	int nLen = (int)(lEndPos - lStartPos);
	ASSERT(nLen > 0);

	// Allocate a character string for it.
	char * pszHeader = new char[nLen + 1];
	// Copy the header to the buffer.  Replace CRLF pairs with a
	// single space character.
	BOOL bFoundCR = FALSE;
	int iCopyIndex = 0;
	for (int i = 0; i < nLen; i++) {
		char c = (char) m_filemap[lStartPos + i];
		// If we've just seen a CR, see if the next character is
		// an LF.
		if (bFoundCR) {
			if (c == LF) {
            	// If so, simply add a space to the output string.
				pszHeader[iCopyIndex++] = ' ';
				bFoundCR = FALSE;
			}
			else {
				// Else the previous CR was isolated by itself and
				// should be in the output string.
				pszHeader[iCopyIndex++] = CR;
				// Is the current character a CR
				if (c == CR) {
					// If so, make sure we check the next character
					// for an LF.
					bFoundCR = TRUE;
				}
				else {
					// Nope.  Just copy it to the output.
					pszHeader[iCopyIndex++] = c;
					bFoundCR = FALSE;
				}
			}
		}
		else {
			// Is the character a CR.
			if (c == CR) {
				// Yes, so check the next character for LF.
				bFoundCR = TRUE;
			}
			else {
				// Nope. Copy it to the output.
				pszHeader[iCopyIndex++] = c;
				bFoundCR = FALSE;
			}
		}
	}

	// Remove any trailing whitespace and terminate the string.
	iCopyIndex--;
	while (IsLWSP(pszHeader[iCopyIndex]))
		iCopyIndex--;
	iCopyIndex++;
	pszHeader[iCopyIndex] = '\0';

	return pszHeader;
}

/*
	Function: CMIMEBodyPart::FindCRLF
	Description:
		Finds the next CRLF pair in the file after the position
		specified by m_lCurrent.  Returns the index in the file
		of the CR character.  Returns -1 if no CRLF pair can be found.

		The function uses a hard-coded Boyer-Moore string search
		algorithm.  It looks for LF every two characters.  When it
		finds one, it checks if the preceeding character was CR.
		If so, it's found one.  If not, it increments two locations
		and checks again.  If the current character was a CR rather
		than LF, it increments only one character in order to check
        the next character for LF.
*/
long CMIMEBodyPart::FindCRLF(void)
{
	long lSearch = m_lCurrent + 1;

	while (1) {
		int c = m_filemap[lSearch];
		if (c < 0)  // Error or EOF
			return -1;
		else if (c == LF) {
			if (m_filemap[lSearch - 1] == CR)
				return lSearch - 1;  // Found it.
			else
				lSearch += 2;
		}
		else if (c == CR)
			lSearch++; // Boyer-Moore step.
		else
			lSearch += 2;
	}
}

/*
	Function: CMIMEBodyPart::StringSearch
	Description:
		Searches for a string in the message text starting at
		m_lCurrent and ending at m_lEnd.  Returns the position of
		the first character in the string and advances m_lCurrent
		to the character following the string.  The function returns
		-1 if the string cannot be found before m_lEnd is reached.

		The function uses a Boyer-Moore string search algorithm in
		order to reduce the number of calls to CFileMap::Get
		and move through the string quickly.  Note, this is a
		pretty confusing search algorithm if you aren't used to it.
		Look it up in a textbook if you want more background
		(Sedgewick, "Algorithms in C," or Abrash, "Zen of Code
		Optimization," for instance).
*/
long CMIMEBodyPart::StringSearch(char * pszString)
{
	int	iAdvanceTable[256];
	int i;

	ASSERT(pszString != NULL);
	int nLen = lstrlen(pszString);

	// If we've been asked to match a zero-length string, match it
	// immediately.
	if (nLen == 0)
    	return m_lCurrent;

	// Fill in the the advance table with the default length.
	for (i = 0; i < 256; i++)
		iAdvanceTable[i] = nLen;
	// Fill in the corresponding positions in the advance table
	// with the positions of the characters in the search string.
	for (i = 0; i < nLen - 1; i++)
		iAdvanceTable[pszString[i]] = nLen - i - 1;

	int iAdvance;
	long lSearchPos = m_lCurrent + nLen - 1;
	while (lSearchPos < m_lEnd) {
		// Search from end of string back to front
		for (i = 0; i < nLen; i++) {
			int c = m_filemap[lSearchPos - i];
			if (c < 0) return -1;
			if (c != pszString[nLen - 1 - i]) {
				// We found a mismatch.  See how far we can
				// advance based on the mismatched character.
				iAdvance = iAdvanceTable[c] - i;
                // We always move forward, never back.
				if (iAdvance <= 0)
					iAdvance = 1;
				break;
			}
		}
		if (i == nLen) {
			// We found it.
            m_lCurrent = lSearchPos + 1;
			return lSearchPos - (nLen - 1);
		}
		else {
			// Didn't find it, so advance the tail a bit more.
			lSearchPos += iAdvance;
		}
	}

	// Didn't find it before we ran out of room.
	return -1;
}

/*
	Function: CMIMEBodyPart::HeaderCompare
	Description:
		Compare a header with a header name.  If the header name
		equals the part of the header preceeding the first colon
		(and after whitespace has been stripped), then return TRUE,
		else return FALSE.  The comparison is case-insensitive.
*/
BOOL CMIMEBodyPart::HeaderCompare(char * pszHeader,
	char *pszHeaderName)
{
	// Find the first colon in the header.
	int i = 0;
	while (pszHeader[i] != ':')
		i++;
	// Now back up to skip any whitespace between the header
	// name and the colon.
	i--; // Skip back past colon.
	while (IsLWSP(pszHeader[i]))
		i--;
	i++; // i = length of header name.

	// Compare the lengths.
	if (i != lstrlen(pszHeaderName))
		return FALSE;

	// Compare the characters.
	for (int j = 0; j < i; j++)
		if (tolower(pszHeader[j]) != tolower(pszHeaderName[j]))
			return FALSE;

	return TRUE;
}

/*
	Function: CMIMEBodyPart::ParseMIMEVersion
	Description:
		Parse the MIME version header.  The function looks for
		version 1.0.  If version 1.0 is found and the header is
		syntactically correct, the function returns TRUE else it
		returns FALSE.
*/
BOOL CMIMEBodyPart::ParseMIMEVersion(char * pszHeader)
{
	ASSERT(pszHeader != NULL);

	// Allocate a token buffer.  The maximum token length could
	// be the size of the header minus the header name.  Allocate
	// one the size of the whole header, which is sure to be enough.
	int nTokenLen = lstrlen(pszHeader);
	char * pszToken = new char[nTokenLen];
	int iToken;

	// Move up to the colon separating the header name from
	// the header body.
	while (*pszHeader != ':' && *pszHeader != '\0')
		pszHeader++;
	if (*pszHeader == '\0') goto error;
	// Move past the colon.
	pszHeader++;

	// Look for the '1'.
	do {
		pszHeader = GetToken(FALSE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	if (iToken != TOKEN_ATOM || lstrcmp(pszToken, "1") != 0)
		goto error;

	// Look for the '.'.
	do {
		pszHeader = GetToken(FALSE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	if (iToken != TOKEN_SPECIAL || lstrcmp(pszToken, ".") != 0)
		goto error;

	// Look for the '0'.
	do {
		pszHeader = GetToken(FALSE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	if (iToken != TOKEN_ATOM || lstrcmp(pszToken, "0") != 0)
		goto error;

	delete[] pszToken;
	return TRUE;

error:
	delete[] pszToken;
    return FALSE;
}

/*
	Function: CMIMEBodyPart::ParseContentType
	Description:
		Parses a Content-Type header.  The header name must have
		been checked previously.  The type and subtype fields are
		parsed from the header and stored in the node.  If
		parameters are present, they are parsed and stored in the
		parameter list.  If everything goes alright, the function
		returns TRUE.  If an error occurs, it returns FALSE.
*/
BOOL CMIMEBodyPart::ParseContentType(char * pszHeader)
{
	ASSERT(pszHeader != NULL);

	// Allocate a token buffer.  The maximum token length could
	// be the size of the header minus the header name.  Allocate
	// one the size of the whole header, which is sure to be enough.
	int nTokenLen = lstrlen(pszHeader);
	char * pszToken = new char[nTokenLen];
	int iToken;
	CMIMEParamTuple tuple;

	// Move up to the colon separating the header name from
	// the header body.
	while (*pszHeader != ':' && *pszHeader != '\0')
		pszHeader++;
	if (*pszHeader == '\0') goto error;
	// Move past the colon.
	pszHeader++;

	// Get the type.
	do {
		pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	if (iToken != TOKEN_MIMETOKEN)
		goto error;
	m_pszType = DuplicateString(pszToken);

	// Get the "/" separator between the type and subtype.
	do {
		pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	if (iToken != TOKEN_TSPECIAL || *pszToken != '/')
		goto error;

	// Get the subtype.
	do {
		pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	if (iToken != TOKEN_MIMETOKEN)
		goto error;
	m_pszSubType = DuplicateString(pszToken);

	// See if we have any parameters.
	do {
		pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	while (iToken == TOKEN_TSPECIAL && *pszToken == ';') {
		// Yes, we have a parameter.
		// Parse a token for the parameter name.
		do {
			pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
				&iToken);
		} while (iToken == TOKEN_COMMENT);
		if (iToken != TOKEN_MIMETOKEN)
			goto error;
		tuple.Param(pszToken);
		// Get the equals sign.
		do {
			pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
				&iToken);
		} while (iToken == TOKEN_COMMENT);
		if (iToken != TOKEN_TSPECIAL || *pszToken != '=')
			goto error;
		// Get the parameter value.
		do {
			pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
				&iToken);
		} while (iToken == TOKEN_COMMENT);
		if (iToken != TOKEN_MIMETOKEN && iToken != TOKEN_QSTRING)
			goto error;
		tuple.Value(pszToken);
		// Add the tuple to the list of parameters.
		m_pParamList->Add(&tuple);
		// See if have another following ';'.
		do {
			pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
				&iToken);
		} while (iToken == TOKEN_COMMENT);
	}

	// Okay, we should be at the end.
	delete[] pszToken;
	if (iToken != TOKEN_END)
		return FALSE;
	else
		return TRUE;

error:
	delete[] pszToken;
	return FALSE;
}

/*
	Function: CMIMEBodyPart::ParseContentTransferEncoding
	Description:
		Parse a Content-Transfer-Encoding header.  Sets the
		m_pszEncoding member to a string describing the encoding.
*/
int CMIMEBodyPart::ParseContentTransferEncoding(char * pszHeader)
{
	ASSERT(pszHeader != NULL);

	// Allocate a token buffer.  The maximum token length could
	// be the size of the header minus the header name.  Allocate
	// one the size of the whole header, which is sure to be enough.
	int nTokenLen = lstrlen(pszHeader);
	char * pszToken = new char[nTokenLen];
	int iToken;

	// Move up to the colon separating the header name from
	// the header body.
	while (*pszHeader != ':' && *pszHeader != '\0')
		pszHeader++;
	if (*pszHeader == '\0') goto error;
	// Move past the colon.
	pszHeader++;

	// Get the encoding.
	do {
		pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	if (iToken != TOKEN_MIMETOKEN)
		goto error;
	m_pszEncoding = DuplicateString(pszToken);

	// Check if we're at the end.
	do {
		pszHeader = GetToken(TRUE, pszHeader, pszToken, nTokenLen,
			&iToken);
	} while (iToken == TOKEN_COMMENT);
	if (iToken != TOKEN_END)
		goto error;
	delete[] pszToken;
	return TRUE;

error:
	delete[] pszToken;
	return FALSE;
}

/*
	Function: CMIMEBodyPart::FirstLine
	Description:
		Returns the first line of text in the body part.  The CRLF
		is left in the text.  Simply resets the current pointer
        to the first byte of the body part and then calls NextLine.
*/
int CMIMEBodyPart::FirstLine(char * pszLine, int nLen)
{
	m_lCurrent = m_lStart;
	return NextLine(pszLine, nLen);
}

/*
	Function: CMIMEBodyPart::NextLine
	Description:
		Returns the next line of text in the body part.  The CRLF
		is left in the text.  Returns the number of characters
		transferred to the buffer, or -1 if there are no more lines
		of text.
*/
int CMIMEBodyPart::NextLine(char * pszLine, int nLen)
{
	if (m_lCurrent >= m_lEnd)
		return -1;

	BOOL bWasCR = FALSE;
	int iLineIndex = 0;
	while (m_lCurrent < m_lEnd && iLineIndex < (nLen - 1)) {
		int c = m_filemap[m_lCurrent++];
		if (c < 0) {
			return -1;
		}
        pszLine[iLineIndex++] = (char) c;
		if (c == LF && bWasCR)
			break;
		if (c == CR)
			bWasCR = TRUE;
		else
			bWasCR = FALSE;
	}
	pszLine[iLineIndex] = '\0';
    return iLineIndex;
}

/*********************************************************************
	Class: CMIMEMsg
*********************************************************************/

/*
	Function: CMIMEMsg::Parse
	Description:
		Parse the given filename.  Returns TRUE if the parse was
		successful or FALSE if an error occurred.
*/
BOOL CMIMEMsg::Parse(char * pszFilename)
{
	ASSERT(pszFilename != NULL);

	if (m_pBodyParts != NULL) return FALSE;

	m_pBodyParts = new CMIMEBodyPart(NULL, pszFilename, 0, -1);
	if (!(m_pBodyParts->Open())) return FALSE;
	if (!(m_pBodyParts->Parse())) return FALSE;
	m_pBodyParts->Close();

    return TRUE;
}

/*
	Function: CMIMEMsg::FirstPart
	Description:
		Moves to the first "viewable" body part in the tree,
		skipping over multipart nodes in favor of their
		children, as necessary.
*/
CMIMEBodyPart *	CMIMEMsg::FirstPart(void)
{
	m_pCurrentPart = m_pBodyParts;
	while (m_pCurrentPart->FirstChild() != NULL)
		m_pCurrentPart = m_pCurrentPart->FirstChild();
	return m_pCurrentPart;
}

/*
	Function: CMIMEMsg::NextPart
	Description:
		Move to the next node in the MIME body part tree.  Returns
		the address of the next CMIMEBodyPart object in the tree
		or NULL if there are no more nodes in the tree.
*/
CMIMEBodyPart *	CMIMEMsg::NextPart(void)
{
	// See if we can move sideways in the tree.
	CMIMEBodyPart * pTemp = m_pCurrentPart;
	while (pTemp != NULL && pTemp->Sibling() == NULL) {
		// If not, then move up until we can move sideways.
		pTemp = pTemp->Parent();
	}
	// See if we've gone through all the nodes.
	if (pTemp == NULL)
		return NULL;
	else
		pTemp = pTemp->Sibling();
	// If the node we found is a multipart node, move down as
	// far as possible.
	while (pTemp->FirstChild() != NULL)
		pTemp = pTemp->FirstChild();

	// Okay, we're there.
	return m_pCurrentPart = pTemp;
}

