/*
	File: CNNTP.CPP
	Description:
		A class for interacting with an NNTP server.
	Copyright 1995, David G. Roberts

	4/20/96: Fixed GetTempFileName bug for Win32 systems
			 in CNNTP::OpenFile.
*/

#include <ctype.h>  // need toupper()
#include <stdlib.h> // need srand()/rand()
#include <time.h>	// need time()
#include "cnntp.h"
#include "utility.h"
#include "myassert.h"

const char CR = '\015';
const char LF = '\012';

// Used to include at the end of text.  The compiler will
// splice strings.
// Use like m_sock.Send("Foo" CRLF,...).  Note that we have to use a
// macro rather than a string constant because we need the
// preprocessor to perform a textual substitution so the string
// splicing will work.
#define CRLF "\015\012"

void CNNTP::Event(EVENT ev)
{
	while ((this->*m_pState)(ev)) {
		ev = EV_NULLEVENT;
	};
}

void CNNTP::StreamCallback(int iToken, STREAMEVENT iEvent, int iError,
	char FAR *lpBuffer)
{
	UNUSED_PARAMETER(iToken);	// Eliminate the compiler warning.
    UNUSED_PARAMETER(lpBuffer);

	m_iError = iError;
	switch (iEvent) {
		case CStreamCallback::SE_CLOSE:
			Event(EV_CLOSE);
			break;
		case CStreamCallback::SE_DATAREADY:
			// Read the data.
			if (ReadReplyData()) {
				while (LineComplete()) {
					CopyLine();
					ShrinkReceiveBuffer();
					if (m_bDataMode) {
						if (MessageTerminator())
							Event(EV_ENDDATA);
						else
							Event(EV_DATA);
					}
					else {
						switch (m_szReplyLineBuffer[0]) {
							case '1': Event(EV_CODE1); break;
							case '2': Event(EV_CODE2); break;
							case '3': Event(EV_CODE3); break;
							case '4': Event(EV_CODE4); break;
							case '5': Event(EV_CODE5); break;
							default:
								// Something went wrong.
								Event(EV_ERROR); break;
						}
					}
				}
			}
			else {
				// Something went wrong.
				WSCFatalError("Couldn't read from socket in CNNTP.");
			}
			break;
		case CStreamCallback::SE_DATASENT:
			// All commands to the server are one line.
			// Even if something blocks, we'll wait for a response
			// before sending another one, which means that the buffer
			// will be free again before we use it.
			break;
		default:
			// Error!  Something we weren't expecting.
			WSCFatalError(
				"Unexpected socket event in CNNTP::StreamCallback");
			break;
	}
}

BOOL CNNTP::ReadReplyData(void)
{
	// Figure out how much room is left in the buffer.
	int iBufferLen = MAX_REPLY_SIZE - m_iReceiveBufferLen;
	ASSERT(iBufferLen > 0);

	// Receive the data.
	int iResult = m_sock.Receive(
		&m_szReceiveBuffer[m_iReceiveBufferLen], &iBufferLen);

	if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
		// If no errors, update the length of the data in the buffer.
    	m_iReceiveBufferLen += iBufferLen;
		return TRUE;
	}
	else {
    	// Save the error and signal it.
		m_iError = iResult;
		return FALSE;
	}
}

BOOL CNNTP::LineComplete(void)
{
	ASSERT(m_iReceiveBufferLen <= MAX_REPLY_SIZE);

	for (int i = 0; i < m_iReceiveBufferLen; i++) {
		if ((m_szReceiveBuffer[i] == CR) &&
			((i + 1) < m_iReceiveBufferLen) &&
			(m_szReceiveBuffer[i + 1] == LF)) {
			return TRUE;
		}
	}
	return FALSE;
}

void CNNTP::CopyLine(void)
{
	ASSERT(m_iReceiveBufferLen <= MAX_REPLY_SIZE);

	for (int i = 0; i < m_iReceiveBufferLen; i++) {
		if ((m_szReceiveBuffer[i] == CR) &&
			((i + 1) < m_iReceiveBufferLen) &&
			(m_szReceiveBuffer[i + 1] == LF)) {
            m_szReplyLineBuffer[i] = '\0';
			break;
		}
		else {
			m_szReplyLineBuffer[i] = m_szReceiveBuffer[i];
		}
	}
}

void CNNTP::ShrinkReceiveBuffer(void)
{
	ASSERT(m_iReceiveBufferLen <= MAX_REPLY_SIZE);

	// Find the CRLF at the end of the line.
	for (int i = 0; i < m_iReceiveBufferLen; i++) {
		if ((m_szReceiveBuffer[i] == CR) &&
			((i + 1) < m_iReceiveBufferLen) &&
			(m_szReceiveBuffer[i + 1] == LF)) {
			// Set i equal to the character following the CRLF.
			i += 2;
			// Copy the characters following the CRLF to the beginning
			// of the line.
			for (int j = 0; j < m_iReceiveBufferLen - i; j++) {
				m_szReceiveBuffer[j] = m_szReceiveBuffer[i + j];
			}
			// Subtract number of characters removed from the current
			// number of characters in the buffer.
			m_iReceiveBufferLen -= i;
            // Break out of the for(i) loop.
			break;
		}
	}
}

BOOL CNNTP::MessageTerminator(void)
{
	if (m_szReplyLineBuffer[0] == '.' &&
		m_szReplyLineBuffer[1] == '\0')
		return TRUE;
	else
		return FALSE;
}

void CNNTP::ConnMgrCallback(int iToken,
	CConnMgrCallback::RESULT result)
{
	UNUSED_PARAMETER(iToken);

	ASSERT(result != CConnMgrCallback::INVALID_RESULT);

    // Turn connection callback result into events.
	switch (result) {
		case CConnMgrCallback::OK_CONNECTED:
			Event(EV_CONNECT);
			break;
		case CConnMgrCallback::ERR_NOTCONNECTED:
			m_iError = m_connmgr.LastError();
			Event(EV_ERROR);
			break;
	}
}

void CNNTP::Abort(HWND hwnd, UINT msg, BOOL bDoCallback)
{
	m_hwnd = hwnd;
	m_msg = msg;
	m_bDoCallback = bDoCallback;
	Event(EV_ABORT);
}

void CNNTP::Connect(HWND hwnd, UINT msg,
	char * lpszHost, u_short ushPort)
{
	ASSERT(lpszHost != NULL);

	// Store parameters away.
	m_hwnd = hwnd;
	m_msg = msg;
	m_pszHost = new char[lstrlen(lpszHost) + 1];
	lstrcpy(m_pszHost, lpszHost);
	m_ushPort = ushPort;

	// Initialize our state correctly.
    m_bDataMode = FALSE;
	m_iReceiveBufferLen = 0;
    m_bDoCallback = TRUE;

	// Start us off.
	NextState(&CNNTP::Start);
	Event(EV_NULLEVENT);
}

void CNNTP::Group(HWND hwnd, UINT msg, char * pszGroup)
{
	m_hwnd = hwnd;
	m_msg = msg;
	m_pszGroup = DuplicateString(pszGroup);
    m_bDoCallback = TRUE;
	Event(EV_GROUP);
}

void CNNTP::NextHeader(HWND hwnd, UINT msg)
{
	m_hwnd = hwnd;
	m_msg = msg;
	m_bDoCallback = TRUE;
	Event(EV_NEXTHEADER);
}

void CNNTP::Header(HWND hwnd, UINT msg)
{
	m_hwnd = hwnd;
	m_msg = msg;
    m_bDoCallback = TRUE;
	Event(EV_HEADER);
}

void CNNTP::Article(HWND hwnd, UINT msg, long lArticle)
{
	m_hwnd = hwnd;
	m_msg = msg;
	m_lArticle = lArticle;
    m_bDoCallback = TRUE;
	Event(EV_ARTICLE);
}

void CNNTP::Quit(HWND hwnd, UINT msg)
{
	m_hwnd = hwnd;
	m_msg = msg;
    m_bDoCallback = TRUE;
	Event(EV_QUIT);
}

BOOL CNNTP::Start(EVENT ev)
{
	UNUSED_PARAMETER(ev);

	// Make the connection to the remote host.  Let the connection
    // manager do all the work.
	m_connmgr.Connect(m_pszHost, m_ushPort, &m_sock, this, 0);
	NextState(&CNNTP::AwaitConnection);
    return TRUE;
}

BOOL CNNTP::AwaitConnection(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_connmgr.Abort();
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
            break;
		case EV_CONNECT:
			NextState(&CNNTP::AwaitGreeting);
			break;
		case EV_ERROR:
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		default:
			return FALSE;
	}
    return TRUE;
}

BOOL CNNTP::AwaitGreeting(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_CODE2:
			// Okay, we're connected and we have the greeting.
			PostMessage(m_hwnd, m_msg, RESULT_SUCCESS, 0);
			NextState(&CNNTP::Idle);
			break;
		case EV_ERROR:
		case EV_CODE1:
		case EV_CODE3:
		case EV_CODE4:
		case EV_CODE5:
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::Idle(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_GROUP:
			NextState(&CNNTP::SendGroup);
			break;
		case EV_NEXTHEADER:
			NextState(&CNNTP::SendNext);
            break;
		case EV_HEADER:
			NextState(&CNNTP::SendHeader);
			break;
		case EV_ARTICLE:
			NextState(&CNNTP::SendArticle);
			break;
		case EV_QUIT:
			NextState(&CNNTP::Quit);
            break;
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::SendGroup(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_NULLEVENT:
			// Send the group command.
			lstrcpy(m_szLineBuffer, "GROUP ");
			lstrcat(m_szLineBuffer, m_pszGroup);
			lstrcat(m_szLineBuffer, CRLF);
			delete[] m_pszGroup;
            m_pszGroup = NULL;
			iResult = m_sock.Send(m_szLineBuffer,
				lstrlen(m_szLineBuffer));
			if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
				NextState(&CNNTP::AwaitGroupResponse);
			}
			else {
				m_iError = iResult;
				m_result = RESULT_FATALERROR;
				NextState(&CNNTP::Exit);
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::AwaitGroupResponse(EVENT ev)
{
	long lTotal;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_CODE2:
			// Parse the return string
			sscanf(m_szReplyLineBuffer, "%*ld %ld", &lTotal);
			PostMessage(m_hwnd, m_msg, RESULT_SUCCESS,
				(LPARAM) lTotal);
			NextState(&CNNTP::Idle);
			break;
		case EV_CODE4:
			PostMessage(m_hwnd, m_msg, RESULT_FAILURE, 0);
			NextState(&CNNTP::Idle);
			break;
		case EV_ERROR:
		case EV_CODE1:
		case EV_CODE3:
		case EV_CODE5:
			// All other responses generate an error.
			m_result = RESULT_FATALERROR;
			NextState(&CNNTP::Quit);	// Try to quit.
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::SendNext(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_NULLEVENT:
			// Send the group command.
			lstrcpy(m_szLineBuffer, "NEXT" CRLF);
			iResult = m_sock.Send(m_szLineBuffer,
				lstrlen(m_szLineBuffer));
			if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
				NextState(&CNNTP::AwaitNextResponse);
			}
			else {
				m_iError = iResult;
				m_result = RESULT_FATALERROR;
				NextState(&CNNTP::Exit);
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::AwaitNextResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_CODE2:
			// Okay, the NEXT command succeeded.  Got get the
            // header itself.
			NextState(&CNNTP::SendHeader);
			break;
		case EV_CODE4:
			PostMessage(m_hwnd, m_msg, RESULT_FAILURE, 0);
			NextState(&CNNTP::Idle);
			break;
		case EV_ERROR:
		case EV_CODE1:
		case EV_CODE3:
		case EV_CODE5:
			// All other responses generate an error.
			m_result = RESULT_FATALERROR;
			NextState(&CNNTP::Quit);	// Try to quit.
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::SendHeader(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_NULLEVENT:
			// Send the header command.
			lstrcpy(m_szLineBuffer, "HEAD" CRLF);
			iResult = m_sock.Send(m_szLineBuffer,
				lstrlen(m_szLineBuffer));
			if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
				NextState(&CNNTP::OpenFile);
			}
			else {
				m_iError = iResult;
				m_result = RESULT_FATALERROR;
				NextState(&CNNTP::Exit);
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::SendArticle(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_NULLEVENT:
			// Send the header command.
			wsprintf(m_szLineBuffer, "ARTICLE %ld" CRLF, m_lArticle);
			iResult = m_sock.Send(m_szLineBuffer,
				lstrlen(m_szLineBuffer));
			if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
				NextState(&CNNTP::OpenFile);
			}
			else {
				m_iError = iResult;
				m_result = RESULT_FATALERROR;
				NextState(&CNNTP::Exit);
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::OpenFile(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_CODE2:
			// Okay, get ready to receive the data.  Open a temp
			// file to contain it.
			sscanf(m_szReplyLineBuffer, "%*ld %ld", &m_lArticle);
			// 4/20/96: Fixed GetTempFileName bug for Win32
			// systems.
#ifdef WIN32
			char szTempPath[256];
			GetTempPath(sizeof(szTempPath), szTempPath);
			GetTempFileName(szTempPath, "NWZ", 0, m_szFileName);
#else
			GetTempFileName(0, "NWZ", 0, m_szFileName);
#endif
			m_ofsFile.open(m_szFileName);
            m_bDataMode = TRUE;
			NextState(&CNNTP::RetrieveLine);
			break;
		case EV_CODE4:
			PostMessage(m_hwnd, m_msg, RESULT_FAILURE, 0);
			NextState(&CNNTP::Idle);
			break;
		case EV_ERROR:
		case EV_CODE1:
		case EV_CODE3:
		case EV_CODE5:
			// All other responses generate an error.
			m_result = RESULT_FATALERROR;
			NextState(&CNNTP::Quit);	// Try to quit.
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::RetrieveLine(EVENT ev)
{
	ARTICLEANDFILENAME * pArtAndFile;

	switch (ev) {
		case EV_ABORT:
			m_bDataMode = FALSE;
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the other
			// end.
			m_bDataMode = FALSE;
			m_result = RESULT_FAILURE;
			NextState(&CNNTP::Exit);
			break;
		case EV_ENDDATA:
			m_bDataMode = FALSE;
			// The data is done.  Close up the file
			m_ofsFile.close();
			// Post a message to the window.
			pArtAndFile = new ARTICLEANDFILENAME;
			pArtAndFile->lArticleNum = m_lArticle;
			pArtAndFile->pszFilename = DuplicateString(m_szFileName);
			PostMessage(m_hwnd, m_msg, RESULT_SUCCESS,
				(LPARAM)pArtAndFile);
			NextState(&CNNTP::Idle);
			break;
		case EV_DATA:
			// Write the line to the file, followed by a local
			// line terminator.
			if (m_szReplyLineBuffer[0] == '.') {
				// If the line starts with a period then it was
				// stuffed there, so skip over it when writing
				// to the file.
				m_ofsFile << &(m_szReplyLineBuffer[1]) << '\n';
			}
			else {
				m_ofsFile << m_szReplyLineBuffer << '\n';
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::Quit(EVENT ev)
{
	UNUSED_PARAMETER(ev);

    int iResult;

	// Send QUIT.
	lstrcpy(m_szLineBuffer, "QUIT" CRLF);
	iResult = m_sock.Send(m_szLineBuffer, lstrlen(m_szLineBuffer));
	if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
		NextState(&CNNTP::AwaitQuitResponse);
	}
	else {
		m_iError = iResult;
		m_result = RESULT_FAILURE;
		NextState(&CNNTP::Exit);
	}
	return TRUE;
}

BOOL CNNTP::AwaitQuitResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CNNTP::Exit);
			break;
		case EV_CLOSE:		// Just about anything is okay.
		case EV_ERROR:
		case EV_CODE1:
		case EV_CODE2:
		case EV_CODE3:
		case EV_CODE4:
		case EV_CODE5:
        	m_result = RESULT_SUCCESS;
			NextState(&CNNTP::Exit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

BOOL CNNTP::Exit(EVENT ev)
{
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

	// Close the socket.
	m_sock.Close();

	// Delete some allocated memory.
	delete[] m_pszHost;
	delete[] m_pszGroup;

	// Send a callback.
	if (m_bDoCallback) {
    	PostMessage(m_hwnd, m_msg, m_result, 0);
	}

	NextState(&CNNTP::Stop);
	return TRUE;
}

BOOL CNNTP::Stop(EVENT ev)
{
	UNUSED_PARAMETER(ev);
	// Simply ignore all events.
	return FALSE;
}

