/*
	File: CPOP3.CPP
	Description:
		A class for retrieving email messages via POP3 (RFC 1460).
	Copyright 1995, David G. Roberts

    4/24/96: Corrected SMTP/POP3 typo in CPOP3WithCallback.
*/

#include <ctype.h>  // need toupper()
#include <stdlib.h> // need srand()/rand()
#include <time.h>	// need time()
#include "cpop3.h"
#include "myassert.h"

const char CR = '\015';
const char LF = '\012';

// Used to include at the end of text.  The compiler will splice
// strings.
// Use like m_sock.Send("Foo" CRLF,...).  Note that we have to use a
// macro rather than a string constant because we need the
// preprocessor to perform a textual substitution so the string
// splicing will work.
#define CRLF "\015\012"

/*
	Function: CPOP3::Event
	Description:
		The state machine event dispatch loop.
*/
void CPOP3::Event(EVENT ev)
{
	while ((this->*m_pState)(ev)) {
		ev = EV_NULLEVENT;
	};
}

/*
	Function: CPOP3::StreamCallback
	Description:
		Invoked when events occur on the socket.
*/
void CPOP3::StreamCallback(int iToken, STREAMEVENT iEvent, int iError,
	char FAR *lpBuffer)
{
	UNUSED_PARAMETER(iToken);	// Eliminate the compiler warning.
    UNUSED_PARAMETER(lpBuffer);

	m_iError = iError;
	switch (iEvent) {
		case CStreamCallback::SE_CLOSE:
			Event(EV_CLOSE);
			break;
		case CStreamCallback::SE_DATAREADY:
			// Read the data.
			if (ReadReplyData()) {
				while (LineComplete()) {
					CopyLine();
					ShrinkReceiveBuffer();
					Event(EV_DATAREADY);
				}
			}
			else {
				// Something went wrong.
				WSCFatalError("Couldn't read from socket in CPOP3.");
			}
			break;
		case CStreamCallback::SE_DATASENT:
			// All commands to the server are one line.
			// Even if something blocks, we'll wait for a response
			// before sending another one, which means that the buffer
			// will be free again before we use it.
			break;
		default:
			// Error!  Something we weren't expecting.
			WSCFatalError(
				"Unexpected socket event in CPOP3::StreamCallback");
			break;
	}
}

/*
	Function: CPOP3::ReadReplyData
	Description:
		Calculates the amount of data the receiver buffer can
		receive at this time and reads it from the socket.
*/
BOOL CPOP3::ReadReplyData(void)
{
	// Figure out how much room is left in the buffer.
	int iBufferLen = MAX_REPLY_SIZE - m_iReceiveBufferLen;
	ASSERT(iBufferLen > 0);

	// Receive the data.
	int iResult = m_sock.Receive(
		&m_szReceiveBuffer[m_iReceiveBufferLen], &iBufferLen);

	if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
		// If no errors, update the length of the data in the buffer.
    	m_iReceiveBufferLen += iBufferLen;
		return TRUE;
	}
	else {
    	// Save the error and signal it.
		m_iError = iResult;
		return FALSE;
	}
}

/*
	Function: CPOP3::LineComplete
	Description:
		Determines if the buffer contains a line of text terminated
		with CRLF.
*/
BOOL CPOP3::LineComplete(void)
{
	ASSERT(m_iReceiveBufferLen <= MAX_REPLY_SIZE);

	for (int i = 0; i < m_iReceiveBufferLen; i++) {
		if ((m_szReceiveBuffer[i] == CR) &&
			((i + 1) < m_iReceiveBufferLen) &&
			(m_szReceiveBuffer[i + 1] == LF)) {
			return TRUE;
		}
	}
	return FALSE;
}

/*
	Function: CPOP3::CopyLine
	Description:
		Copies a line from the receive buffer to the reply line
		buffer.
*/
void CPOP3::CopyLine(void)
{
	ASSERT(m_iReceiveBufferLen <= MAX_REPLY_SIZE);

	for (int i = 0; i < m_iReceiveBufferLen; i++) {
		if ((m_szReceiveBuffer[i] == CR) &&
			((i + 1) < m_iReceiveBufferLen) &&
			(m_szReceiveBuffer[i + 1] == LF)) {
            m_szReplyLineBuffer[i] = '\0';
			break;
		}
		else {
			m_szReplyLineBuffer[i] = m_szReceiveBuffer[i];
		}
	}
}

/*
	Function: CPOP3::ShrinkReceiveBuffer
	Description:
		Deletes the first line of CRLF terminated text from the
		receiver buffer (presumably after it has been moved to the
		reply line buffer).
*/
void CPOP3::ShrinkReceiveBuffer(void)
{
	ASSERT(m_iReceiveBufferLen <= MAX_REPLY_SIZE);

	// Find the CRLF at the end of the line.
	for (int i = 0; i < m_iReceiveBufferLen; i++) {
		if ((m_szReceiveBuffer[i] == CR) &&
			((i + 1) < m_iReceiveBufferLen) &&
			(m_szReceiveBuffer[i + 1] == LF)) {
			// Set i equal to the character following the CRLF.
			i += 2;
			// Copy the characters following the CRLF to the beginning
			// of the line.
			for (int j = 0; j < m_iReceiveBufferLen - i; j++) {
				m_szReceiveBuffer[j] = m_szReceiveBuffer[i + j];
			}
			// Subtract number of characters removed from the current
			// number of characters in the buffer.
			m_iReceiveBufferLen -= i;
            // Break out of the for(i) loop.
			break;
		}
	}
}

/*
	Function: CPOP3::PositiveResponse
	Description:
		Returns TRUE if the reply line buffer beings with "+OK"
		and FALSE otherwise.
*/
BOOL CPOP3::PositiveResponse(void)
{
	// See if we have a "+OK..." response
	if (m_szReplyLineBuffer[0] == '+' &&
		toupper(m_szReplyLineBuffer[1]) == 'O' &&
		toupper(m_szReplyLineBuffer[2]) == 'K')
		return TRUE;
	else
		return FALSE;
}

/*
	Function: CPOP3::MessageTerminator
	Description:
		Returns TRUE if the line simply contains a single period
		character, indicating the end of the message text.
*/
BOOL CPOP3::MessageTerminator(void)
{
	if (m_szReplyLineBuffer[0] == '.' &&
		m_szReplyLineBuffer[1] == '\0')
		return TRUE;
	else
		return FALSE;
}

/*
	Function: CPOP3::NewFileName
	Description:
		Creates a random, unique filename to store a received
		message in.
*/
void CPOP3::NewFileName(char * pszPathName, char * pszFileName,
	char * pszFullName)
{
	char szTrialName[13];  // All file names will be 8.3 format.
	HFILE hfile;

	do {
		// Choose eight random letters.
		for (int i = 0; i < 8; i++) {
			int r = rand() % 26;
            szTrialName[i] = (char)( 'a' + r );
		}
		// Tack on a ".txt" extension.
		szTrialName[8] = '\0';
		lstrcat(szTrialName, ".txt");
		// Make the file name = the trial name.
		lstrcpy(pszFileName, szTrialName);
        // Build the full name from the path name and the trial name.
		lstrcpy(pszFullName, pszPathName);
		lstrcat(pszFullName, "\\");
		lstrcat(pszFullName, szTrialName);
		// See if the file already exists.
		OFSTRUCT ofstructTemp;
		hfile = OpenFile(pszFullName, &ofstructTemp, OF_EXIST);
		// If not, choose another name.
	} while (hfile != HFILE_ERROR);
}

/*
	Function: CPOP3::ConnMgrCallback
	Description:
		The callback function for the connection manager object.
		The function conditions the connection manager results and
		creates events for the CPOP3 state machine.
*/
void CPOP3::ConnMgrCallback(int iToken,
	CConnMgrCallback::RESULT result)
{
	UNUSED_PARAMETER(iToken);

	ASSERT(result != CConnMgrCallback::INVALID_RESULT);

    // Turn connection callback result into events.
	switch (result) {
		case CConnMgrCallback::OK_CONNECTED:
			Event(EV_CONNECT);
			break;
		case CConnMgrCallback::ERR_NOTCONNECTED:
			m_iError = m_connmgr.LastError();
			Event(EV_ERROR);
			break;
	}
}

/*
	Function: CPOP3::Abort
	Description:
		Called when the program wants to abort the current POP3
		session.  Injects an EV_ABORT event into the state machine.
		The bDoCallback argument determines whether the program
		wants a callback to be generated.  This might be useful if
		some code requires the CPOP3 callback in order to clean up
		allocated memory or other things.
*/
void CPOP3::Abort(BOOL bDoCallback)
{
	m_bDoCallback = bDoCallback;
	Event(EV_ABORT);
}

/*
	Function: CPOP3::RetrieveMessages
	Description:
		This starts the whole thing off.
*/
BOOL CPOP3::RetrieveMessages(LPCSTR lpszHost, u_short ushPort,
	LPCSTR lpszUser, LPCSTR lpszPassword, LPCSTR lpszPathName,
	CStringList * pslFileList)
{
	ASSERT(lpszHost != NULL);
	ASSERT(pslFileList != NULL);

	// Make sure we're currently stopped.
	// Store parameters away.
	m_pszHost = new char[lstrlen(lpszHost) + 1];
	lstrcpy(m_pszHost, lpszHost);
	m_ushPort = ushPort;
	m_pszUser = new char[lstrlen(lpszUser) + 1];
	lstrcpy(m_pszUser, lpszUser);
	m_pszPassword = new char[lstrlen(lpszPassword) + 1];
	lstrcpy(m_pszPassword, lpszPassword);
	m_pszPathName = new char[lstrlen(lpszPathName) + 1];
	lstrcpy(m_pszPathName, lpszPathName);
	m_pslFileList = pslFileList;

	// Initialize our state correctly.
	m_iReceiveBufferLen = 0;
    m_bDoCallback = TRUE;

	// Seed the random number generator so we can generate
	// unique file names to store messages in.
	time_t curtime = time(NULL);
	srand((unsigned)curtime);

	// Start us off.
	NextState(&CPOP3::Start);
	Event(EV_NULLEVENT);

    return TRUE;
}

/*
	Function: CPOP3::Start
	Description:
		The first state in the state machine.
*/
BOOL CPOP3::Start(EVENT ev)
{
	UNUSED_PARAMETER(ev);

	// Make the connection to the remote host.  Let the connection
    // manager do all the work.
	m_connmgr.Connect(m_pszHost, m_ushPort, &m_sock, this, 0);
	NextState(&CPOP3::AwaitConnection);
    return TRUE;
}

/*
	Function: CPOP3::AwaitConnection
	Description:
		Awaits the results of the connection manager's connection
		attempt.
*/
BOOL CPOP3::AwaitConnection(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_connmgr.Abort();
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
            break;
		case EV_CONNECT:
			NextState(&CPOP3::AwaitGreeting);
			break;
		case EV_ERROR:
			m_result = RESULT_FAILURE;
			NextState(&CPOP3::Exit);
			break;
		default:
			return FALSE;
	}
    return TRUE;
}

/*
	Function: CPOP3::AwaitGreeting
	Description:
		Waits for the line of greeting text from the remote
		POP3 server.
*/
BOOL CPOP3::AwaitGreeting(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CPOP3::Exit);
			break;
		case EV_DATAREADY:
			// Okay, we're connected and we have the greeting.
			// Ignore the greeting message (should be "+OK..."
			// but who knows).
			wsprintf(m_szLineBuffer, "USER %s" CRLF,
				(LPSTR)m_pszUser);
			iResult = m_sock.Send(m_szLineBuffer,
				lstrlen(m_szLineBuffer));
			if (iResult == WSANOERROR ||
				iResult == WSAEWOULDBLOCK) {
				NextState(&CPOP3::AwaitUserResponse);
			}
			else {
				m_iError = iResult;
				m_result = RESULT_FAILURE;
				NextState(&CPOP3::Exit);
			}
            break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CPOP3::AwaitUserResponse
	Description:
		Waits for the server's reply to the USER command.
*/
BOOL CPOP3::AwaitUserResponse(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CPOP3::Exit);
			break;
		case EV_DATAREADY:
			if (PositiveResponse()) {
				// We got an "+OK..." response.  Send the password.
				wsprintf(m_szLineBuffer, "PASS %s" CRLF,
					(LPSTR)m_pszPassword);
				iResult = m_sock.Send(m_szLineBuffer,
					lstrlen(m_szLineBuffer));
				if (iResult == WSANOERROR ||
					iResult == WSAEWOULDBLOCK) {
					NextState(&CPOP3::AwaitPassResponse);
				}
				else {
					m_iError = iResult;
					m_result = RESULT_FAILURE;
					NextState(&CPOP3::Exit);
				}
			}
			else {
				// All other responses generate a failure.
				m_result = RESULT_FAILURE;
				NextState(&CPOP3::Quit);	// Now go send QUIT.
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CPOP3::AwaitPassResponse
	Description:
		Waits for the server's reply to the PASS command.
*/
BOOL CPOP3::AwaitPassResponse(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CPOP3::Exit);
			break;
		case EV_DATAREADY:
			if (PositiveResponse()) {
				// We got an "+OK..." response.
                // Ask for the mailbox statistics.
				lstrcpy(m_szLineBuffer, "STAT" CRLF);
				iResult = m_sock.Send(m_szLineBuffer,
					lstrlen(m_szLineBuffer));
				if (iResult == WSANOERROR ||
					iResult == WSAEWOULDBLOCK) {
					NextState(&CPOP3::AwaitStatResponse);
				}
				else {
					m_iError = iResult;
					m_result = RESULT_FAILURE;
					NextState(&CPOP3::Exit);
				}
			}
			else {
				// All other responses generate a failure.
				m_result = RESULT_FAILURE;
				NextState(&CPOP3::Quit);	// Now go send QUIT.
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CPOP3::AwaitStatResponse
	Description:
		Waits for the server's reply to the STAT command.  When the
		reply comes, the function extracts the message count from
		the line of reply text.
*/
BOOL CPOP3::AwaitStatResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CPOP3::Exit);
			break;
		case EV_DATAREADY:
			if (PositiveResponse()) {
				// We got an "+OK..." response.
				// Parse the response for the message count.
				if (sscanf(m_szReplyLineBuffer, "+OK %d",
					&m_iTotalMessages) != 1) {
					// If the response didn't contain a count of
					// messages following the +OK, something's wrong.
					m_result = RESULT_FAILURE;
					NextState(&CPOP3::Exit);
				}
				else {
					// Set the current message count to zero.
					m_iRetrieved = 0;
					NextState(&CPOP3::StartRetrieve);
				}
			}
			else {
				// All other responses generate a failure.
				m_result = RESULT_FAILURE;
				NextState(&CPOP3::Quit);	// Now go send QUIT.
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CPOP3::StartRetrieve
	Description:
		Starts the retrieval process by creating and opening a file
		to hold the message and sending the RETR command to the
		server.
*/
BOOL CPOP3::StartRetrieve(EVENT ev)
{
	UNUSED_PARAMETER(ev);

	// Have we retrieved all the messages.
	if (m_iRetrieved >= m_iTotalMessages) {
		// Yes, so just quit gracefully with a success indication.
		// Note that this will happen even if there are no messages.
		// The caller should be prepared to find no message file names
        // in the file name list.
		m_result = RESULT_SUCCESS;
		NextState(&CPOP3::Quit);
        return TRUE;
	}

	// Find a unique file name, open a stream for it.
	char szFullName[260];
	NewFileName(m_pszPathName, m_szFileName, szFullName);
	m_ofsFile.open(szFullName);

	// Send a RETR command to get the current message.
	wsprintf(m_szLineBuffer, "RETR %d" CRLF, m_iRetrieved + 1);
	int iResult = m_sock.Send(m_szLineBuffer,
		lstrlen(m_szLineBuffer));
	if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
		NextState(&CPOP3::AwaitRetrResponse);
	}
	else {
		m_iError = iResult;
		m_result = RESULT_FAILURE;
		NextState(&CPOP3::Exit);
	}
	return TRUE;
}

/*
	Function: CPOP3::AwaitRetrResponse
	Description:
		Waits for the server's reply to the RETR command.
*/
BOOL CPOP3::AwaitRetrResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CPOP3::Exit);
			break;
		case EV_DATAREADY:
			if (PositiveResponse()) {
            	NextState(&CPOP3::RetrieveLine);
			}
			else {
				// All other responses generate a failure.
				m_result = RESULT_FAILURE;
				NextState(&CPOP3::Quit);	// Now go send QUIT.
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CPOP3::RetrieveLine
	Description:
		This state machine state retrieves each line of message
		text and writes it to the message file until the end of
		message indication is found (".CRLF").  It then sends
		the DELE command to delete the message.
*/
BOOL CPOP3::RetrieveLine(EVENT ev)
{
	int iResult;

	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CPOP3::Exit);
			break;
		case EV_DATAREADY:
			if (MessageTerminator()) {
				// The message is done.  Close up the file
				m_ofsFile.close();
				// Tell the server to delete the message from the
				// mailbox when the session concludes.
				wsprintf(m_szLineBuffer, "DELE %d" CRLF,
					m_iRetrieved + 1);
				iResult = m_sock.Send(m_szLineBuffer,
					lstrlen(m_szLineBuffer));
				if (iResult == WSANOERROR ||
					iResult == WSAEWOULDBLOCK) {
					NextState(&CPOP3::AwaitDeleResponse);
				}
				else {
					m_iError = iResult;
					m_result = RESULT_FAILURE;
					NextState(&CPOP3::Exit);
				}
			}
			else {
				// Write the line to the file, followed by a local
				// line terminator.
				if (m_szReplyLineBuffer[0] == '.') {
					// If the line starts with a period then it was
					// stuffed there, so skip over it when writing
                    // to the file.
					m_ofsFile << &(m_szReplyLineBuffer[1]) << '\n';
				}
				else {
					m_ofsFile << m_szReplyLineBuffer << '\n';
				}
				// No need to set the state.  Stay in RetrieveLine
                // until we get the message terminator.
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CPOP3::AwaitDeleResponse
	Description:
		Waits for the server's reply to the DELE command.
*/
BOOL CPOP3::AwaitDeleResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
			break;
		case EV_CLOSE:
			// Hmmm... unexpected close of the socket from the
			// other end.
			m_result = RESULT_FAILURE;
			NextState(&CPOP3::Exit);
			break;
		case EV_DATAREADY:
			if (PositiveResponse()) {
				m_iRetrieved++;
				m_pslFileList->Add(m_szFileName);
				NextState(&CPOP3::StartRetrieve);
			}
			else {
				// All other responses generate a failure.
				m_result = RESULT_FAILURE;
				NextState(&CPOP3::Quit);	// Now go send QUIT.
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CPOP3::Quit
	Description:
		Sends the QUIT command to the server to end the POP3 session.
*/
BOOL CPOP3::Quit(EVENT ev)
{
	UNUSED_PARAMETER(ev);

    int iResult;

	lstrcpy(m_szLineBuffer, "QUIT" CRLF);
	iResult = m_sock.Send(m_szLineBuffer, lstrlen(m_szLineBuffer));
	if (iResult == WSANOERROR || iResult == WSAEWOULDBLOCK) {
		NextState(&CPOP3::AwaitQuitResponse);
	}
	else {
		m_iError = iResult;
		m_result = RESULT_FAILURE;
		NextState(&CPOP3::Exit);
	}
	return TRUE;
}

/*
	Function: CPOP3::AwaitQuitResponse
	Description:
		Waits for the server's reply to the QUIT command.  Note that
		the state accepts any reply.  Theoretically the protocol
		only allows the server to reply with +OK, but it doesn't
		really matter at this point.
*/
BOOL CPOP3::AwaitQuitResponse(EVENT ev)
{
	switch (ev) {
		case EV_ABORT:
			m_result = RESULT_ABORT;
			NextState(&CPOP3::Exit);
			break;
		case EV_CLOSE:		// Just about anything is okay.
		case EV_DATAREADY:
			NextState(&CPOP3::Exit);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Function: CPOP3::Exit
	Description:
		The Exit state closes the socket, deallocates memory
		allocated when the state machine started, and calls
		DoCallback to perform the appropriate callback behavior.
*/
BOOL CPOP3::Exit(EVENT ev)
{
	UNUSED_PARAMETER(ev);	// Eliminate the compiler warning.

	// Close the socket.
	m_sock.Close();

	// Free up the stuff we allocated when we started.
	delete[] m_pszHost;
	delete[] m_pszUser;
	delete[] m_pszPassword;
	delete[] m_pszPathName;

	// Now, invoke the callback with the connection attempt result.
	if (m_bDoCallback) {
    	DoCallback();
	}

	NextState(&CPOP3::Stop);
	return TRUE;
}

/*
	Function: CPOP3::Stop
	Description:
		The Stop state simply devours any and all events sent to it
		without doing anything with them.  This causes any spurious
		events that may still be generated after the state machine
		completes to be ignored.
*/
BOOL CPOP3::Stop(EVENT ev)
{
	UNUSED_PARAMETER(ev);
	// Simply ignore all events.
	return FALSE;
}

/*********************************************************************
	Callbacks for specialized CPOP3 derivatives.
*********************************************************************/

void CPOP3WithWinMsg::DoCallback(void)
{
	PostMessage(m_hwnd, m_msg, m_iToken, m_result);
}

void CPOP3WithCallback::DoCallback(void)
{
	m_pCallback->POP3Callback(m_iToken, m_result);
}
