/*
	File: CSOCKET.CPP
	Description:
		Class to encapsulate a Windows socket.
	Copyright 1995, David G. Roberts

	4/18/96:  Added initialization of queue pointers to the
              private Accept constructor.
	11/17/95: Modified private version of CStreamSocket::CStreamSocket
	          used in the Accept function to register the new socket
			  with the event dispatch mechanism.
*/

#include "csocket.h"
#include "myassert.h"

/*
	Function: CStreamSocket::WinsockEventCallback
	Description:
		Receives Winsock events from the event dispatch
		mechansim.
	Inputs:
		Standard Windows message parameters.
	Output:
		None.
*/
void CStreamSocket::WinsockEventCallback(HWND hwnd, UINT msg,
	WPARAM wParam, LPARAM lParam)
{
	UNUSED_PARAMETER(hwnd);
	UNUSED_PARAMETER(msg);
    UNUSED_PARAMETER(wParam);

	// If we've removed the socket callback function for some reason,
	// then don't bother calling it.  Simply ignore the message.
	if (m_pCallback == NULL)
		return;
	// Crack apart the event code and error code.
	int iEvent = WSAGETSELECTEVENT(lParam);
	int iError = WSAGETSELECTERROR(lParam);
	switch (iEvent) {
		case FD_ACCEPT:
			ASSERT(m_iState == LISTENING);
			m_pCallback->StreamCallback(m_iToken,
				CStreamCallback::SE_ACCEPT, iError, NULL);
			break;
		case FD_CLOSE:
			m_pCallback->StreamCallback(m_iToken,
				CStreamCallback::SE_CLOSE, iError, NULL);
			break;
		case FD_CONNECT:
			ASSERT(m_iState == CONNECTING);
			if (iError == WSANOERROR) {
				// If no error, signal that we now want read, write,
				// and close events and move the CStreamSocket
				// state to CONNECTED.
				WSAAsyncSelect(m_s, WSCGetEventWindow(),
					WM_WSAASYNCSELECT,
					FD_READ | FD_WRITE | FD_CLOSE);
				m_iState = CONNECTED;
			}
			m_pCallback->StreamCallback(m_iToken,
				CStreamCallback::SE_CONNECT, iError, NULL);
			break;
		case FD_READ:
			// Invoke the callback and say that there's data to be
			// read. Following this, somebody should call the
            // Receive member function to actually get the data.
			m_pCallback->StreamCallback(m_iToken,
				CStreamCallback::SE_DATAREADY, iError, NULL);
			break;
		case FD_WRITE:
			// Winsock has some buffer space.  Do we have any data?
			while (m_pbdHead != NULL) {
				// Yes, so try to send it.
				int iSentLength;
				do {
					iSentLength = send(m_s,
						m_pbdHead->m_lpCurrentBuffer,
						m_pbdHead->m_iLength, 0);
					if (iSentLength != SOCKET_ERROR) {
						m_pbdHead->m_lpCurrentBuffer += iSentLength;
						m_pbdHead->m_iLength -= iSentLength;
					}
				} while (iSentLength != SOCKET_ERROR &&
					m_pbdHead->m_iLength > 0);
				if (m_pbdHead->m_iLength > 0) {
					// There's still more data left.  Figure out why.
					m_iLastError = WSAGetLastError();
					if (m_iLastError != WSAEWOULDBLOCK) {
						// Something bad happened.  Dequeue the
						// buffer and invoke the callback with the
						// bad news.
						char FAR * lpOriginalBuffer =
							m_pbdHead->m_lpOriginalBuffer;
						DeleteHead();
						m_pCallback->StreamCallback(m_iToken,
							CStreamCallback::SE_DATASENT,
							m_iLastError,
							lpOriginalBuffer);

					}
					else {
						// Else error == WSAEWOULDBLOCK, which means
						// that Winsock is just out of space.  We'll
						// just break out of the enclosing while loop
						// and continue to send the buffer when
						// Winsock sends us another FD_WRITE event.
						break;
					}
				}
				else {
					// We completed sending that buffer.  Dequeue
					// it and invoke the callback.
					char FAR * lpOriginalBuffer =
						m_pbdHead->m_lpOriginalBuffer;
					DeleteHead();
					m_iLastError = WSANOERROR;
					m_pCallback->StreamCallback(m_iToken,
						CStreamCallback::SE_DATASENT, m_iLastError,
						lpOriginalBuffer);
				}
			} // while (m_pbdHead != NULL)
			break;
		default:
			// The event is unknown, so simply ignore it.
			break;
	}
}

/*
	Function:
		CStreamSocket::CStreamSocket(SOCKET, CStreamCallback *, int)
	Description:
		A private constructor used to create a newly connected
		CStreamSocket in Accept.
	Inputs:
		s -- socket descriptor of new socket from Winsock.
		pCallback -- the callback function for the new CStreamSocket.
		iToken -- the token value of the new CStreamSocket.
	Outputs: None.
*/
CStreamSocket::CStreamSocket(SOCKET s, CStreamCallback * pCallback,
	int iToken)
	: m_s(s), m_iLastError(WSANOERROR), m_pCallback(pCallback),
	m_iToken(iToken), m_iState(CONNECTED),
	// 4/18/96: Added initialization of queue pointers.
	m_pbdHead(NULL), m_pbdTail(NULL) 
{
	// 11/17/95: Added the following two lines to register the newly
	// created socket with the event dispatch mechanism. Previously,
	// events were enabled on the socket but were simply discarded
	// by the event dispatch mechansim because the new socket was not
	// registered with it.
	SetCallback(pCallback, iToken);
	Register((WPARAM) m_s, WPARAM_TYPE_SOCKET);

	WSAAsyncSelect(m_s, WSCGetEventWindow(), WM_WSAASYNCSELECT,
    	FD_READ | FD_WRITE | FD_CLOSE);
}

/*
	Function: CStreamSocket::~CStreamSocket
	Description:
		Destructor for CStreamSocket.
	Inputs: None.
	Outputs: None.
*/
CStreamSocket::~CStreamSocket()
{
	// Close out the socket.
	if (m_s != INVALID_SOCKET)
		Close();
}

/*
	Function: CStreamSocket::Accept
	Description:
		Accepts a connection from a foreign host.  The socket should
		have already been put in a listening state by calling Listen.
	Inputs:
		pCallback -- the callback function to be associated with the
			newly created CStreamSocket.
		iToken -- the token value to be passed with the callback for
			the newly created CStreamSocket.
		lpsaAddr -- an optional parameter to hold the address of the
			foreign host.
		lpiAddrLen -- on input, the length of the data structure
			pointedto by lpsaAddr.  On exit, the length of the data
			written to the structure.
	Output:
		A pointer to a newly created CStreamSocket.  The socket can be
		destroyed with 'delete' at a later time.  If an error occurs,
		NULL is returned and the error code may be retrieved using
		LastError.  If the code is WSAEWOULDBLOCK, then there is no
		connection waiting.
*/
CStreamSocket * CStreamSocket::Accept(CStreamCallback * pCallback,
	int iToken, CSockAddr * pcsaAddr)
{
	ASSERT(m_s != INVALID_SOCKET && m_iState == LISTENING);

	// Try to accept the incoming connection.
	SOCKADDR saAddr;
	int iAddrLen = sizeof(saAddr);
	SOCKET s = accept(m_s, &saAddr, &iAddrLen);
	if (s == INVALID_SOCKET) {
		m_iLastError = WSAGetLastError();
		return NULL;
	}
	if (pcsaAddr != NULL)
		*pcsaAddr = saAddr;
	// If successful, create a new CStreamSocket
	CStreamSocket * pssNew = new CStreamSocket(s, pCallback, iToken);
	if (pssNew == NULL)
		return NULL;
	return pssNew;
}

/*
	Function: CStreamSocket::Bind
	Description:
		Binds a socket to an address.  If the socket has not yet
		been opened, the socket is first opened.
	Input:
		saBindAddr -- reference to the socket address to bind to.
	Output:
		WSANOERROR if no error occurred or another Winsock
		error code.
*/
int CStreamSocket::Bind(CSockAddr& saBindAddr)
{
	int iError;

	// If socket has not yet be opened, then open it.
	if (m_s == INVALID_SOCKET) {
		ASSERT(m_iState == IDLE);
		iError = Open();
		if (iError != WSANOERROR)
			return iError;
	}

	// Bind the socket to the specified address.
	iError = bind(m_s, (LPSOCKADDR) saBindAddr, sizeof(SOCKADDR));
	if (iError == SOCKET_ERROR) {
		m_iLastError = WSAGetLastError();
		return m_iLastError;
	}
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::Close
	Description:
		Closes a socket.
	Inputs: None.
	Output:
		WSANOERROR if no error occurred or another Winsock
		error code.
*/
int CStreamSocket::Close(void)
{
	// It's okay to close an already closed socket.
	if (m_s == INVALID_SOCKET)
		return m_iLastError = WSANOERROR;
	// Make sure we don't get any more callbacks.
	Deregister((WPARAM) m_s, WPARAM_TYPE_SOCKET);
	// Delete all the queued buffer descriptors.  Note that no
	// callbacks are given.  Presumably, the receiver of the that
	// callbacks knows we're being closed and won't miss them.
	while (m_pbdHead != NULL)
		DeleteHead();
	// Close the socket.
	if (closesocket(m_s) == SOCKET_ERROR)
		m_iLastError = WSAGetLastError();
	else
		m_iLastError = WSANOERROR;
	// We're now idle.
	m_s = INVALID_SOCKET;
	m_iState = IDLE;
	return m_iLastError;
}

/*
	Function: CStreamSocket::Connect
	Description:
		Start the connection process.  When the connection
		is set up, the application will receive a callback.
	Input:
		saForeignAddr -- the remote address to connect to.
	Output:
		WSANOERROR if no errors occurred or other Winsock
		error.
*/
int CStreamSocket::Connect(CSockAddr& saForeignAddr)
{
	int iError;

	// If socket has not yet be opened, then open it.
	if (m_s == INVALID_SOCKET) {
		ASSERT(m_iState == IDLE);
		iError = Open();
		if (iError != WSANOERROR)
			return iError;
	}
	// Set us up to receive connect events.
	iError = WSAAsyncSelect(m_s, WSCGetEventWindow(),
		WM_WSAASYNCSELECT, FD_CONNECT);
	if (iError == SOCKET_ERROR)
		return m_iLastError = WSAGetLastError();
	// Try to connect us.
	iError = connect(m_s, (LPSOCKADDR) saForeignAddr,
		sizeof(SOCKADDR));
	if (iError == SOCKET_ERROR) {
		m_iLastError = WSAGetLastError();
		if (m_iLastError == WSAEWOULDBLOCK) {
			// We would have blocked, so the connection is in
			// progress. Set our state to connecting and wait for the
			// callback event.
			m_iState = CONNECTING;
		}
		return m_iLastError;
	}
	// Somehow we got lucky and the call didn't block.  So indicate
	// that we're connected.
    m_iState = CONNECTED;
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::GetPeerName
	Description:
		Gets the address and port information of the remote host.
	Input:
		psaPeerName -- pointer to a CSockAddr object which will hold
			the result.
	Output:
		The peer address is placed in the structure pointed to by
		psaPeerName.
		The function returns WSANOERROR if no errors occurred or
		another Winsock error code.
*/
int CStreamSocket::GetPeerName(CSockAddr * psaPeerName)
{
	// See if we have a valid socket.
	if (m_s == INVALID_SOCKET) {
		return m_iLastError = WSAENOTSOCK;
	}

	int nLen = sizeof(SOCKADDR);
	int iError = getpeername(m_s, (LPSOCKADDR)(*psaPeerName), &nLen);
	if (iError == SOCKET_ERROR) {
		m_iLastError = WSAGetLastError();
		return m_iLastError;
	}
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::GetSockName
	Description:
		Gets the address and port that the socket is bound to.
		If the socket has not been opened, it is opened and bound
		to INADDR_ANY, port 0.
	Input:
		psaSockName -- pointer to a CSockAddr object which will hold
			the result.
	Output:
		The local address is placed in the structure pointed to by
		psaSockName.
		The function returns WSANOERROR if no errors occurred or
		another Winsock error code.
*/
int CStreamSocket::GetSockName(CSockAddr * psaSockName)
{
	int iError;

	// If socket has not yet be opened, then open it and bind it.
	if (m_s == INVALID_SOCKET) {
		ASSERT(m_iState == IDLE);
		iError = Open();
		if (iError != WSANOERROR)
			return iError;
		CSockAddr saTemp(INADDR_ANY, 0);
		iError = Bind(saTemp);
		if (iError != WSANOERROR)
			return iError;
	}

	// Get it's name.
	int nLen = sizeof(SOCKADDR);
	iError = getsockname(m_s, (LPSOCKADDR)(*psaSockName), &nLen);
	if (iError == SOCKET_ERROR) {
		m_iLastError = WSAGetLastError();
		return m_iLastError;
	}
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::GetSockOpt
	Description:
		Gets the value of a current socket option.
	Inputs:
		iLevel -- the level of the socket option.
		iOptName -- the name of the option.
		pOptVal -- pointer to a buffer to receive the option value.
		pnLen -- pointer to an integer holding the length of the
			buffer.
	Outputs:
		Sets *pnLen to the length of the data returned in the
		buffer.
		Directly returns a Winsock error value or WSANOERROR if
		no errors occurred.
*/
int CStreamSocket::GetSockOpt(int iLevel, int iOptName,
	char * pOptVal, int * pnLen)
{
	int iError;

	// If socket has not yet be opened, then open it.
	if (m_s == INVALID_SOCKET) {
		ASSERT(m_iState == IDLE);
		iError = Open();
		if (iError != WSANOERROR)
			return iError;
	}

	// Get the socket option information.
	iError = getsockopt(m_s, iLevel, iOptName, pOptVal, pnLen);
	if (iError == SOCKET_ERROR) {
		m_iLastError = WSAGetLastError();
		return m_iLastError;
	}
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::IoctlSocket
	Description:
		Changes the current mode of a socket.
	Inputs:
		lCommand -- the command to invoke on the socket.
		pArg -- pointer to an argument for the command.
	Outputs:
		Returns a Winsock error code or WSANOERROR if no errors
		occurred.
*/
int CStreamSocket::IoctlSocket(long lCommand, u_long * pArg)
{
	int iError;

	// If socket has not yet be opened, then open it.
	if (m_s == INVALID_SOCKET) {
		ASSERT(m_iState == IDLE);
		iError = Open();
		if (iError != WSANOERROR)
			return iError;
	}

	// Get the socket option information.
	iError = ioctlsocket(m_s, lCommand, pArg);
	if (iError == SOCKET_ERROR) {
		m_iLastError = WSAGetLastError();
		return m_iLastError;
	}
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::Listen
	Description:
		Put the socket into listening mode.
	Inputs:
		saServiceAddr -- the address to bind the socket to
			before listening.
		iQueueLen -- the length of the queue passed to
			Winsock "listen".
	Output:
		WSANOERROR if no errors occurred.
		Other Winsock error as necessary.
*/
int CStreamSocket::Listen(CSockAddr& saServiceAddr,
	int iQueueLen)
{
	int iError;

	// If socket has not yet be opened, then open it.
	if (m_s == INVALID_SOCKET) {
		ASSERT(m_iState == IDLE);
		iError = Open();
		if (iError != WSANOERROR)
			return iError;
	}
	// Bind the socket to the specified port and interface address.
	iError = bind(m_s, (LPSOCKADDR) saServiceAddr, sizeof(SOCKADDR));
	if (iError == SOCKET_ERROR)
		return m_iLastError = WSAGetLastError();
	// Set us up to listen with the specified connection queue length.
	iError = listen(m_s, iQueueLen);
	if (iError == SOCKET_ERROR)
		return m_iLastError = WSAGetLastError();
	// Make Winsock generate accept notification messages.
	iError = WSAAsyncSelect(m_s, WSCGetEventWindow(),
		WM_WSAASYNCSELECT, FD_ACCEPT);
	m_iState = LISTENING;
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::Open
	Description:
		Open a socket for communcation.
	Inputs: None.
	Output:
		WSANOERROR if no error occurred.
		Other Winsock error as returned from "socket" or
		"ioctlsocket".
*/
int CStreamSocket::Open(void)
{
	// Make sure we don't have something open.
	ASSERT(m_s == INVALID_SOCKET && m_iState == IDLE);
	// Create the socket.
	m_s = socket(AF_INET, SOCK_STREAM, 0);
	if (m_s == INVALID_SOCKET)
		return m_iLastError = WSAGetLastError();
	// Make the socket non-blocking.
	u_long ulFlag = 1;
	// Note, the compiler may give a warning about the FIONBIO
	// constant below.  Ignore the warning; everything is okay.
	int iError = ioctlsocket(m_s, FIONBIO, &ulFlag);
	if (iError == SOCKET_ERROR) {
		closesocket(m_s);
        m_s = INVALID_SOCKET;
		return m_iLastError = WSAGetLastError();
	}
	// Register the callbacks.
	Register((WPARAM) m_s, WPARAM_TYPE_SOCKET);
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::Receive
	Description:
		Get received data from the socket.
	Inputs:
		lpBuffer -- pointer to the buffer to receive the data.
		piBuffLength -- a pointer to an integer that
			contains the length of the buffer.  
		iFlags -- flags passed to Winsock "recv" function.
	Outputs:
		piBuffLength -- on exit, the integer
			is overwritten with the amount of data actually
			read from the socket.  If the connection is
			closed gracefully, the returned length will be
			zero.
		Directly returns the error code returned from "recv".
*/
int CStreamSocket::Receive(char FAR * lpBuffer, int * piBuffLength,
	int iFlags)
{
	int iReceivedLength = recv(m_s, lpBuffer, *piBuffLength, iFlags);
	if (iReceivedLength == SOCKET_ERROR) {
		*piBuffLength = 0;
		return m_iLastError = WSAGetLastError();
	}
	*piBuffLength = iReceivedLength;
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::Send
	Description:
		Send data using the socket.  If the result of the
		transmission attempt is WSAEWOULDBLOCK, the buffer
		containing the data is queued until the socket will
		accept the data.
	Inputs:
		lpBuffer -- pointer to the buffer to be sent.
		iLength -- length of the buffer.
		iFlags -- flags to be passed to Windows Sockets send
			function.
	Outputs:
		WSANOERROR -- no errors occurred.
		WSAEWOULDBLOCK -- some or all of the data could not be
			accepted at this time.  The buffer has been queued
			for later transmission and a callback will be
			generated when the data is sent.
		Other -- other Winsock errors as returned by "send".
*/
int CStreamSocket::Send(char FAR * lpBuffer, int iLength, int iFlags)
{
	ASSERT(m_iState == CONNECTED);

	// See if there's stuff on the queue.
	if (m_pbdHead != NULL) {
		// Yup, so we're already waiting to send stuff.  Simply chain
		// this buffer to the end.
		AddTail(lpBuffer, lpBuffer, iLength);
		return m_iLastError = WSAEWOULDBLOCK;
	}
	// There's no data on the queue, so simply try to send the data.
	char FAR * lpOriginalBuffer = lpBuffer;
	int iSentLength;
	do {
		iSentLength = send(m_s, lpBuffer, iLength, iFlags);
		if (iSentLength != SOCKET_ERROR) {
			lpBuffer += iSentLength;
			iLength -= iSentLength;
		}
	} while (iSentLength != SOCKET_ERROR && iLength > 0);
	if (iLength > 0) {
		// We didn't send all the data.  Find out why.
		m_iLastError = WSAGetLastError();
		if (m_iLastError == WSAEWOULDBLOCK) {
			// Winsock ran out of buffer space, so buffer the
			// remaining data.
			AddTail(lpOriginalBuffer, lpBuffer, iLength);
			return WSAEWOULDBLOCK;
		}
		else {
			// Something bad happened.  Tell the caller.
			return m_iLastError;
		}
	}
	else {
		// All the data got sent, so return with no error.  Note that
		// we don't call the callback routine because we could get in
		// a recursive loop if the callback routine tries to send more
		// data.  Depending on how many times we looped, we might run
		// out of stack space.  The caller should use a return value
		// of WSANOERROR in lieu of a callback to indicate that the
		// buffer is free.
		return m_iLastError = WSANOERROR;
	}
}

/*
	Function: CStreamSocket::SetSockOpt
	Description:
		Set a current socket option.
	Inputs:
		iLevel -- the level of the socket option.
		iOptName -- the name of the socket option.
		pOptVal -- pointer to a buffer containing the new value for
			the socket option.
		nLen -- the length of the data in the buffer.
	Output:
		Returns a Winsock error code or WSANOERROR if no errors
		occurred.
*/
int CStreamSocket::SetSockOpt(int iLevel, int iOptName,
	char * pOptVal, int nLen)
{
	int iError;

	// If socket has not yet be opened, then open it.
	if (m_s == INVALID_SOCKET) {
		ASSERT(m_iState == IDLE);
		iError = Open();
		if (iError != WSANOERROR)
			return iError;
	}

	// Get the socket option information.
	iError = setsockopt(m_s, iLevel, iOptName, pOptVal, nLen);
	if (iError == SOCKET_ERROR) {
		m_iLastError = WSAGetLastError();
		return m_iLastError;
	}
	return m_iLastError = WSANOERROR;
}

/*
	Function: CStreamSocket::AddTail
	Description:
		Add a buffer to the end of the sending queue.
	Inputs:
		lpOriginalBuffer -- pointer to the buffer as originally
			passed to Send.  We need this to identify the buffer
			to the application in the completion callback.
		lpCurrentBuffer -- pointer to the first byte of data
			remaining in the buffer.
		iLength -- the length of data remaining to be sent.
	Output:
		None.
*/
void CStreamSocket::AddTail(char FAR * lpOriginalBuffer,
	char FAR * lpCurrentBuffer, int iLength)
{
	BufferDescriptor * pbdNew = new BufferDescriptor(lpOriginalBuffer,
		lpCurrentBuffer, iLength);
	if (m_pbdHead == NULL) {
		// The list was empty.
		m_pbdHead = m_pbdTail = pbdNew;
	}
	else {
    	// Chain the buffer to the end.
		m_pbdTail->m_pbdNext = pbdNew;
		m_pbdTail = pbdNew;
	}
}

/*
	Function: CStreamSocket::DeleteHead
	Description:
		Remove a buffer from the head of the sending queue.
	Inputs: None.
	Outputs: None.
*/
void CStreamSocket::DeleteHead(void)
{
	ASSERT(m_pbdHead != NULL);

	BufferDescriptor * pbdNextHead = m_pbdHead->m_pbdNext;
	delete m_pbdHead;
	m_pbdHead = pbdNextHead;
	if (m_pbdHead == NULL)	// If we just deleted the last in
		m_pbdTail = NULL;	//   the chain...
}

