/*
	File: PARSDATE.CPP
	Description:
		Functions used for parsing an RFC822 or ctime style date
		string.
	History:
		1/1/95 -- Created, Dave Roberts
	Copyright 1995, David G. Roberts
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <windows.h>
#include "parsdate.h"

/* CONSTANTS */
static char *szDayName[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri",
	"Sat" };
static char *szLongDayName[7] = { "Sunday", "Monday", "Tuesday",
	"Wednesday", "Thursday", "Friday", "Saturday" };
static char *szMonthName[12] = {"Jan", "Feb", "Mar", "Apr", "May",
	"Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
static char *szLongMonthName[12] = {"January", "February", "March",
	"April", "May", "June", "July", "August", "September",
	"October", "November", "December" };
static int iDaysPerMonth[12] = { 31, 29, 31, 30, 31, 30, 31, 31,
	30, 31, 30, 31 };

// Timezone info.  Much of the TZ info here (including the comments) came
// from the date_util.c file of the source to the Elm mail system,
// Copyright (c) 1993 USENET Community Trust.
typedef struct {
	char	*szName;	// Name of timezone
	int		iOffset;	// Offset east of GMT in military format
} TZ_INFO;

static TZ_INFO tzInfo[] = {
	// From RFC822
	{ "UT", 0 }, { "UCT", 0 }, { "GMT", 0 },
	{ "EST", -500 }, { "EDT", -400 },
	{ "CST", -600 }, { "CDT", -500 },
	{ "MST", -700 }, { "MDT", -600 },
	{ "PST", -800 }, { "PDT", -700 },
	{ "HST", -1000 },					// hawaii
	{ "Z", 0}, // Military time ("Zulu" = GMT), rest of militaries not used

	// popular European timezones
	{ "WET", 0 },		// western european
	{ "MET", 100 },		// middle european
	{ "EET", 200 },		// eastern european
	{ "BST", 100 },		// ??? british summer time

	// Canadian timezones
	{ "AST", -400 }, { "ADT", -300 },	// atlantic
	{ "NST", -330 }, { "NDT", -230 },	// newfoundland
	{ "YST", -900 }, { "YDT", -800 },	// yukon

	// Asian timezones
	{ "JST", 900 },		// japan
	{ "SST", 800 },		// singapore

	// South-Pacific timezones
	{ "NZST", 1200 }, { "NZDT", 1300 },	// new zealand
	{ "WST", 800 }, { "WDT", 900 }		// western australia
};

// Date format codes.
#define	DF_RFC822	(1)
#define DF_CTIME	(2)

// ParseDate state numbers.
#define	ST_WDAY		(0)
#define ST_COMMA	(1)
#define ST_MDAY		(2)
#define ST_DASH1	(3)
#define ST_MONTH	(4)
#define ST_DASH2	(5)
#define ST_YEAR		(6)
#define ST_HOUR		(7)
#define ST_COLON1	(8)
#define ST_CORRECT4	(9)
#define ST_MIN		(10)
#define ST_COLON2	(11)
#define	ST_SEC		(12)
#define ST_ZONE		(13)

/* MACROS */
#define IsALPHA(c)	((65 <= c && c <= 90) || (97 <= c && c <= 122))
#define IsDIGIT(c)	(48 <= c && c <= 57)
#define IsLWSP(c)	(c == 32 || c == 9)
#define DIM(x)		(sizeof(x)/sizeof(x[0]))

/*
	Function: ConvertTimeZone
	Description:
		Converts a timezone string into a numeric offset in minutes
		for the zone.
		Strings can specify a number of different named timezones
		(e.g., EST, PDT, etc.) or a numeric offset in string format
		(e.g., +100, -800, +1).
*/
int ConvertTimeZone(char * szTZ)
{
	int iResult;
	int i;
	int iLen;

	if (szTZ == NULL || szTZ[0] == '\0')
		return 0;	// Null strings = GMT.

	if (((szTZ[0] == '+' || szTZ[0] == '-') && IsDIGIT(szTZ[1])) ||
		IsDIGIT(szTZ[0])) {
		// This one looks to be numeric.
		iResult = atoi(szTZ);
		iLen = lstrlen(szTZ);
		if (iLen == 2 || iLen == 3)	// If we got a 2-digit offset
			iResult *= 100;			// then convert it to 4-digit offset.
	}
	else {
		// Search for a named timezone.
		if (szTZ[0] == '-')	// Elm mentions that some bad mailers put
			szTZ++;			// '-' in front of a zone name, so skip it.

		iResult = 0;		// Default.
		// Search the zone name list.
		for (i = 0; i < DIM(tzInfo); i++) {
			if (lstrcmpi(szTZ, tzInfo[i].szName) == 0) {
				iResult = tzInfo[i].iOffset;
				break;
			}
		} // end for (i...
	} // end else...

	// Now convert timezone from military format offset to offset in
	// minutes.
	iResult = (iResult / 100) * 60 + (iResult % 100);
	return iResult;
}

/*
	Function: ConvertDay
	Description:
		Returns number of month corresponding to string argument.
		January = 0.
*/
int ConvertDay(char * szDay)
{
	int i;

	// Compare against short names.
	for (i = 0; i < DIM(szDayName); i++) {
		// Use windows case insensitive string comparison function.
		if (lstrcmpi(szDay, szDayName[i]) == 0)
			return i;
	}
	// Compare against long names.
	for (i = 0; i < DIM(szLongDayName); i++) {
		// Use windows case insensitive string comparison function.
		if (lstrcmpi(szDay, szLongDayName[i]) == 0)
			return i;
	}
	return -1;
}

/*
	Function: ConvertMonth
	Description:
		Returns number of month corresponding to string argument.
		January = 0.
*/
int ConvertMonth(char * szMonth)
{
	int i;

	// Compare against short names.
	for (i = 0; i < DIM(szMonthName); i++) {
		// Use windows case insensitive string comparison function.
		if (lstrcmpi(szMonth, szMonthName[i]) == 0)
			return i;
	}
	// Compare against long names.
	for (i = 0; i < DIM(szLongMonthName); i++) {
		// Use windows case insensitive string comparison function.
		if (lstrcmpi(szMonth, szLongMonthName[i]) == 0)
			return i;
	}
	return -1;
}

/*
	Function: GetDateToken
	Description:
		Gets the next token from a date string.  A date token is
		defined to be a sequence of letters or numbers, but not both, or
		a punctuation or control character by itself.  The function
		returns a pointer to the next character that was not consumed.
		The scanned token is returned in szBuffer and is up to uMaxLen
		in length.
*/
char FAR * GetDateToken(char FAR * lpszDate, char * szBuffer,
	unsigned uMaxLen)
{
	int i;
	int c;
	BOOL bAlphaMode;

	// Skip any leading white space
	while (IsLWSP(*lpszDate))
		lpszDate++;

	// See if we've found the end of the string.
	if (*lpszDate == '\0') {
    	// If so, then return null token and don't advance.
		szBuffer[0] = '\0';
		return lpszDate;
	}

	// See if we got a punctuation, control, or space character.
	if (!IsALPHA(*lpszDate) && !IsDIGIT(*lpszDate)) {
		// If so, return just that single character.
		szBuffer[0] = *lpszDate;
		szBuffer[1] = '\0';
		return lpszDate + 1;
	}
	else if (IsALPHA(*lpszDate))	// Scanning alphas or digits?
		bAlphaMode = TRUE;
	else
		bAlphaMode = FALSE;

	// Loop until we encounter a character that is not alpha or digit,
	// or we run out of room in the token buffer.
	i = 0;
	while (1) {
		c = *lpszDate;
		if ((unsigned)i == (uMaxLen - 1) || 
			bAlphaMode && !IsALPHA(c) ||
			!bAlphaMode && !IsDIGIT(c))
			break;
		szBuffer[i] = (char)c;
		i++;
		lpszDate++;
	}
	// Null terminate token buffer string.
	szBuffer[i] = '\0';

	return lpszDate;
}

time_t ParseDateString(char FAR * lpszDate)
{
	char	szToken[21];	// Input token (20 + terminator)
	int		iState;			// Parser state
	int		iFormat;		// Major date format
	BOOL	bAdvance;		// Should machine advance to next token?
	BOOL	bDone;			// Have we parsed a complete date yet?
	int		iDOW, iDay,		// Variables to hold results of parse.
			iMonth, iYear,
			iHour, iMin,
			iSec, iZoneOffset;
	struct tm tmTime;		// Temp structure needed for mktime conversion.
	time_t	tLocalCorrection; // Needed because mktime assumes time is
							  // specified in local time.
	time_t	tResult;		// The result of our hard work.

	iState		= ST_WDAY;		// Start off looking for a weekday name.
	bAdvance	= TRUE;			// Need to get initial token.
	bDone		= FALSE;		// We're just getting started.

	// Loop until we've found all the parts.  Premature end of the date
	// string causes a direct return with a value of -1.
	while (!bDone) {
		if (bAdvance) {
			// Get next token from the input.
			lpszDate = GetDateToken(lpszDate, szToken, sizeof(szToken));
		}
		switch (iState) {
			case ST_WDAY:	// Parse weekday name.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				if (IsDIGIT(szToken[0])) {
					// If the day name starts with a digit, we know we've
					// got an RFC822-style date, so move on to day of month.
					iFormat		= DF_RFC822;
					iState		= ST_MDAY;
					bAdvance	= FALSE;	// Don't advance.
				}
				else {
					// Convert day of week.
					iDOW = ConvertDay(szToken);
					if (0 <= iDOW && iDOW <= 11) {
						// If it's a valid day name, continue.
						iState		= ST_COMMA;
						bAdvance	= TRUE;
					}
					else {
						// If it's not a valid day name, then perhaps
						// we have a ctime-style date with a missing
						// day name.  Assume it's a month name instead.(??)
						iFormat		= DF_CTIME;
						iState		= ST_MONTH;
						bAdvance	= FALSE;	// Don't advance.
					}
				}
				break;	// Go to next state.
			case ST_COMMA:	// Check for a comma following weekday name.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				if (szToken[0] == ',' && szToken[1] == '\0') {
					// If we have a comma, then it's an RFC822-style date.
					iFormat		= DF_RFC822;
					iState		= ST_MDAY;
					bAdvance	= TRUE;
				}
				else {
					// Otherwise it's a ctime-style date.
					iFormat		= DF_CTIME;
					iState		= ST_MONTH;
					bAdvance	= FALSE;	// Don't advance.
				}
				break;	// Go to next state.
			case ST_MDAY:	// Parse day of month.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				iDay = atoi(szToken);
				if (iDay < 1 || iDay > 31)	// Sanity check.
					return -1;
				if (iFormat == DF_RFC822) {
					iState		= ST_DASH1;
					bAdvance	= TRUE;
				}
				else if (iFormat == DF_CTIME) {
					// This is a ctime-style date so we have already
					// parsed the month.  Do a bit more sanity checking.
					// Note that a compiler may flag this line with a
					// "Possible use before definition" warning which
                    // can be ignored.
					if (iDay > iDaysPerMonth[iMonth])
						return -1;
					iState		= ST_HOUR;
					bAdvance	= TRUE;
				}
				else
					return -1;	// Problem: invalid date format!!
				break;	// Go to next state.
			case ST_DASH1:	// Skip optional dash between day and month.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				if (szToken[0] == '-' && szToken[1] == '\0')
					bAdvance = TRUE;
				else
					bAdvance = FALSE;
				iState = ST_MONTH;
				break;	// Go to next state.
			case ST_MONTH:	// Parse month name.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				iMonth = ConvertMonth(szToken);
				if (iMonth < 0 || iMonth > 11) // Sanity check.
					return -1;
				if (iFormat == DF_RFC822) {
					// We now have both date and month info.  Do a bit
					// more sanity checking on day.
					if (iDay > iDaysPerMonth[iMonth])
						return -1;
					iState		= ST_DASH2;
					bAdvance	= TRUE;
				}
				else if (iFormat == DF_CTIME) {
					iState		= ST_MDAY;
					bAdvance	= TRUE;
				}
				else
					return -1;	// Problem: invalid date format!!
				break;	// Go to next state.
			case ST_DASH2:	// Skip optional dash between month and year.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				if (szToken[0] == '-' && szToken[1] == '\0')
					bAdvance = TRUE;
				else
					bAdvance = FALSE;
				iState = ST_YEAR;
				break;	// Go to next state.
			case ST_YEAR:	// Parse year.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				iYear = atoi(szToken);
				if (iYear > 100)	// Just use the last two digits
					iYear %= 100;
				if (iFormat == DF_RFC822) {
					iState		= ST_HOUR;
					bAdvance	= TRUE;
				}
				else if (iFormat == DF_CTIME) {
                	bDone		= TRUE; // We're done!
				}
				else
					return -1;	// Problem: invalid date format!!
				break;	// Go to next state.
			case ST_HOUR:	// Parse hour.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				iHour = atoi(szToken);
				if (iHour < 0)	// Sanity check.
					return -1;
				if (iHour <= 23) {
					// Hour appears to be in range.
					iState		= ST_COLON1;
					bAdvance	= TRUE;
				}
				else {
					// Hour is out of range, so we have a military
					// style 4-digit hours-minutes (HHMM).
					iState		= ST_CORRECT4;
					bAdvance	= TRUE;
				}
				break;	// Go to next state.
			case ST_COLON1:	// Check for colon between hour and minutes.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				if (szToken[0] == ':' && szToken[1] == '\0') {
					iState		= ST_MIN;
					bAdvance	= TRUE;
				}
				else {
					iState		= ST_CORRECT4;
					bAdvance	= FALSE;	// Don't advance.
				}
				break;	// Go to next state.
			case ST_CORRECT4: // Correct for 4-digit hours/mins (HHMM)
				iMin	= iHour % 100;
				iHour	= iHour / 100;
				if (iMin > 59 || iHour > 23)	// Sanity check.
					return -1;
				iSec		= 0;	// No seconds.
				iState		= ST_ZONE;
				bAdvance	= FALSE;
				break;
			case ST_MIN:	// Parse minutes.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				iMin = atoi(szToken);
				if (iMin < 0 || iMin > 59)	// Sanity check.
					return -1;
				iState		= ST_COLON2;
				bAdvance	= TRUE;
				break;
			case ST_COLON2:	// Check for colon between minutes and seconds.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				if (szToken[0] == ':' && szToken[1] == '\0') {
					iState		= ST_SEC;
					bAdvance	= TRUE;
				}
				else {
					iSec		= 0;	// No seconds.
					iState		= ST_ZONE;
					bAdvance	= FALSE;
				}
				break;	// Go to next state.
			case ST_SEC:	// Parse seconds.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				iSec = atoi(szToken);
				if (iSec < 0 || iSec > 59)	// Sanity check
					return -1;
				iState		= ST_ZONE;
				bAdvance	= TRUE;
				break;
			case ST_ZONE:	// Parse timezone.
				if (szToken[0] == '\0') // Check for premature end.
					return -1;
				// Ctime-style dates have the year following the
				// optional timezone string.  If the timezone starts
				// with a digit rather than '+' or '-' and we're parsing
				// a ctime-style date, then assume the timezone isn't
				// present and we have a year.
				if (IsDIGIT(szToken[0]) && iFormat == DF_CTIME) {
					iZoneOffset	= 0; // Assume GMT if no other info.
					iState		= ST_YEAR;
					bAdvance	= FALSE;
				}
				else {
					// See if character parsed was a '+' or a '-'.
					if ((szToken[0] == '+' || szToken[0] == '-') &&
						szToken[1] == '\0') {
						// If so, grab the rest of the timezone following
						// and append it to the current token.
						lpszDate = GetDateToken(lpszDate, szToken + 1,
							sizeof(szToken) - 1);
					}
					iZoneOffset = ConvertTimeZone(szToken);
					if (iFormat == DF_RFC822) {
						bDone = TRUE;	// We're done!
					}
					else if (iFormat == DF_CTIME) {
						iState		= ST_YEAR;
						bAdvance	= TRUE;
					}
					else
						return -1;	// Problem: invalid date format!!
				}
				break;		// Go to next state.
			default:
				return -1;	// The state machine is off in the weeds!!!
		}
	}

	// Now convert all the parsed information into a time_t value.

	// First convert 01 Jan 70 00:00:00 to a time_t value.  It would be
	// nice if the tm structure contained a timezone field that mktime
	// could use to interpret the rest of the time specified.  There is,
	// no such field, however, and so mktime interprets the tm structure
	// fields as a local time and adds in an offset to get a GMT time_t
	// value.  The following code figures out what this offset is so that
	// we can subtract it out later and add in our own offset based on the
	// timezone value we just parsed in the date.
	tmTime.tm_sec	= 0;
	tmTime.tm_min	= 0;
	tmTime.tm_hour	= 0;
	tmTime.tm_mday	= 1;
	tmTime.tm_mon	= 0;
	tmTime.tm_year	= 70;
	tmTime.tm_wday	= 0;	// Don't need this.
	tmTime.tm_yday	= 0;	// Don't need this.
	tmTime.tm_isdst	= 0;	// Don't need this.
	tLocalCorrection = mktime(&tmTime);

	tmTime.tm_sec	= iSec;
	tmTime.tm_min	= iMin;
	tmTime.tm_hour	= iHour;
	tmTime.tm_mday	= iDay;
	tmTime.tm_mon	= iMonth;
	tmTime.tm_year	= iYear;
	tmTime.tm_wday	= 0;	// Don't need this.
	tmTime.tm_yday	= 0;	// Don't need this.
	tmTime.tm_isdst	= 0;	// Don't need this.

	// Convert time to seconds since the epoch.
	tResult = mktime(&tmTime);

	// Now adjust time to GMT, taking into account the local bias
	// introduced by mktime and the parsed timezone value.
	tResult -= (tLocalCorrection + (iZoneOffset * 60));

	return tResult;
}

