/*
	File: RFC822.CPP
	Description:
		A bunch of functions used for RFC 822 stuff.
	Copyright 1995, David G. Roberts

	11/17/95: Modified GetCurrentRFC822Date to work correctly under
	          Windows '95.
*/

#include <time.h>
#include <windows.h>
#include "winsock.h"
#include "rfc822.h"
#include "myassert.h"

#define ABS(x) ((x >= 0) ? x : -x)

/*
	Function: GetCurrentRFC822Date
	Description:
		Gets the current date in RFC 822 format. The input buffer
		must be large enough to hold the resulting date (size it at
		least as large as MAX_DATE_LENGTH, defined in RFC822.H).
*/
void GetCurrentRFC822Date(char * pszBuffer)
{
// Split into Win32 and Win16 versions because the VC++ 2.0 library
// didn't correctly deal with the time zone variables under
// Windows 95. For some strange reason, the whole date string would
// be set to NULL.
#ifdef WIN32
	const char * szDays[] = { "Sun", "Mon", "Tue", "Wed",
		"Thu", "Fri", "Sat" };
	const char * szMonths[] = { "Jan", "Feb", "Mar", "Apr", "May",
		"Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

	// Get the current local time.
	SYSTEMTIME curtime;
	GetLocalTime(&curtime);
	// Figure out the timezone information.
	TIME_ZONE_INFORMATION tzinfo;
	GetTimeZoneInformation(&tzinfo);
	// Convert timezone bias into 4-digit format.
	// The following equation to set TZSign purposefully flips the
	// sign of the Bias field. Windows Bias values have the opposite
	// sign used in RFC822. For example, in Windows, Pacific Standard
	// Time is +0800, while RFC822 says it should be -0800. 
	char TZSign = (tzinfo.Bias <= 0) ? '+' : '-';
	int HourBias = ABS(tzinfo.Bias) / 60;
	int MinuteBias = ABS(tzinfo.Bias) % 60;
	// Fill out the buffer.
	wsprintf(pszBuffer, "%s, %02d %s %02d %02d:%02d:%02d %c%02d%02d",
		(LPSTR)szDays[curtime.wDayOfWeek],
		curtime.wDay,
		(LPSTR)szMonths[curtime.wMonth - 1],
		curtime.wYear,
		curtime.wHour,
		curtime.wMinute,
		curtime.wSecond,
		TZSign, 
		HourBias, 
		MinuteBias);
		
#else // Win16
	struct tm * ptm;
	time_t curtime;
	const char * szDays[] = { "Sun", "Mon", "Tue", "Wed",
		"Thu", "Fri", "Sat" };
	const char * szMonths[] = { "Jan", "Feb", "Mar", "Apr", "May",
		"Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

	// Set the _tzname, _daylight, and _timezone variables.
	tzset();
	// Get the current time.
	time(&curtime);
	ptm = localtime(&curtime);
	// Fill out the buffer.
	wsprintf(pszBuffer, "%s, %02d %s %02d %02d:%02d:%02d %s",
		(LPSTR)szDays[ptm->tm_wday],
		ptm->tm_mday,
		(LPSTR)szMonths[ptm->tm_mon],
		ptm->tm_year,
		ptm->tm_hour,
		ptm->tm_min,
		ptm->tm_sec,
		(LPSTR)_tzname[_daylight]);
#endif
}

/*
	Function: GetMessageID
	Description:
		Creates a text string message ID, virtually guaranteed to be
		unique on this host.  The message ID string consists of a
		string of digits representing the current date and time, a
		message counter value, an "at-sign" (@), and the host name.
		The current date and time and the counter value make the
		string unique.  There is a remote possibility that if you
		had another program running on the same machine at the same
		time that it could send a message with the same message ID,
		but this possibility is very remote.
*/
void GetMessageID(char * pszBuffer)
{
	struct tm * ptm;
	time_t curtime;
	static unsigned long counter = 0;
	char szHostName[100];

	// Get the current time.
	time(&curtime);
	ptm = localtime(&curtime);
	gethostname(szHostName, sizeof(szHostName));
	wsprintf(pszBuffer, "%02d%02d%02d%02d%02d%02d.%lu@%s",
		ptm->tm_mon + 1, ptm->tm_mday, ptm->tm_year, ptm->tm_hour,
		ptm->tm_min, ptm->tm_sec, counter, (LPSTR)szHostName);
	counter++;
}

/*********************************************************************
	Parsing Functions
*********************************************************************/

// Standard RFC 822 specials
const char szSpecials[] = "()<>@,;:\\\".[]";
// MIME (RFC 1521) tspecials
const char szTSpecials[] = "()<>@,;:\\\"/[]?=";

/*
	Function: IsInSet
	Description:
		Returns TRUE if the specified character is a part of the
		given character set.
*/
BOOL IsInSet(const int c, const char * szSet)
{
	int i;

	for (i = 0; szSet[i] != '\0'; i++)
		if (szSet[i] == c) return TRUE;
	return FALSE;
}

/*
	Function: SkipLWSP
	Description:
		Returns the first character of the specified string
		that is not a linear whitespace character.
*/
char * SkipLWSP(char * pszBuffer)
{
	while (IsLWSP(*pszBuffer))
		pszBuffer++;
	return pszBuffer;
}

/*
	Function: LexAtom
	Description:
		Gets an atom from the specified header.  Returns the
		first character that could not be consumed. The buffer length
		is specified in nLen and must be long enough to hold the
		entire token.
*/
char * LexAtom(char * pszHeader, char * pszBuffer, int nLen)
{
	int iHeader = 0;
	int iBuffer = 0;
	char c = pszHeader[iHeader++];
	while (!IsSPECIAL(c) && !IsSPACE(c) && !IsCTL(c) &&
		iBuffer < (nLen - 1)) {
		pszBuffer[iBuffer++] = c;
		c = pszHeader[iHeader++];
	}
	pszBuffer[iBuffer] = '\0';
	return pszHeader + iHeader - 1;
}

/*
	Function: LexQText
	Description:
		Gets quoted text from the specified header.  Returns the
		first character that could not be consumed. The buffer length
		is specified in nLen and must be long enough to hold the
		entire token.
*/
char * LexQText(char * pszHeader, char * pszBuffer, int nLen)
{
	int iHeader = 0;
	int iBuffer = 0;
	char c = pszHeader[iHeader++];
	while (c != '"' && iBuffer < (nLen - 1)) {
		// Skip over quoted pairs.
		if (c == '\\') {
			pszBuffer[iBuffer++] = pszHeader[iHeader++];
		}
		else {
			pszBuffer[iBuffer++] = c;
		}
		c = pszHeader[iHeader++];
	}
	pszBuffer[iBuffer] = '\0';
	return pszHeader + iHeader - 1;
}

/*
	Function: LexDText
	Description:
		Gets a domain literal from the specified header. Returns the
		first character that could not be consumed. The buffer length
		is specified in nLen and must be long enough to hold the
		entire token.
*/
char * LexDText(char * pszHeader, char * pszBuffer, int nLen)
{
	int iHeader = 0;
	int iBuffer = 0;
	char c = pszHeader[iHeader++];
	while (c != '[' && c != ']' && iBuffer < (nLen - 1)) {
		// Skip over quoted pairs.
		if (c == '\\') {
			pszBuffer[iBuffer++] = pszHeader[iHeader++];
		}
		else {
			pszBuffer[iBuffer++] = c;
		}
		c = pszHeader[iHeader++];
	}
	pszBuffer[iBuffer] = '\0';
	return pszHeader + iHeader - 1;
}

/*
	Function: LexCText
	Description:
		Gets comment text from the specified header.  Returns the
		first character that could not be consumed. The buffer length
		is specified in nLen and must be long enough to hold the
		entire token.
*/
char * LexCText(char * pszHeader, char * pszBuffer, int nLen)
{
	int iHeader = 0;
	int iBuffer = 0;
	int iNestLevel = 0;
	char c = pszHeader[iHeader++];
	while ((c != ')' || iNestLevel != 0) && iBuffer < (nLen - 1)) {
		// Skip over quoted pairs.
		if (c == '\\') {
			pszBuffer[iBuffer++] = pszHeader[iHeader++];
		}
		else {
			pszBuffer[iBuffer++] = c;
			if (c == '(')
				iNestLevel++;
			if (c == ')')
            	iNestLevel--;
		}
		c = pszHeader[iHeader++];
	}
	pszBuffer[iBuffer] = '\0';
	return pszHeader + iHeader - 1;
}

/*
	Function: LexMIMEToken
	Description:
		Gets MIME token from the specified header.  A MIME token
		is defined in RFC 1521 and is slightly different than an
		RFC 822 atom.  The function returns the first character in
        the stream that could not be consumed.
*/
char * LexMIMEToken(char * pszHeader, char * pszBuffer, int nLen)
{
	int iHeader = 0;
	int iBuffer = 0;
	char c = pszHeader[iHeader++];
	while (!IsTSPECIAL(c) && !IsSPACE(c) && !IsCTL(c) &&
		iBuffer < (nLen - 1)) {
		pszBuffer[iBuffer++] = c;
		c = pszHeader[iHeader++];
	}
	pszBuffer[iBuffer] = '\0';
	return pszHeader + iHeader - 1;
}

/*
	Function: GetToken
	Description:
		Gets an RFC 822 token of some sort from the buffer.  The
		buffer is filled with the token and the integer specified
		by piType is set to the token type.  The next character
		following the token is returned.  If an error occurs,
		*piType is set to TOKEN_ERROR and the function return value 
		indicates where the error occurred.  If bMIME is true, the
		function looks for MIME tokens rather than RFC 822 atoms.
		Note that the enclosing characters of quoted text,
		domain literals, and comments ( "\"[]()" ) is stripped
		from the token before it is returned.

		Note that pszBuffer must be non-NULL and nLen must 2 or more
		(2 for special or CTL and terminating null).
*/
char * GetToken(BOOL bMIME, char * pszHeader, char * pszBuffer,
	int nLen, int * piType)
{
	ASSERT(pszBuffer != NULL);
	ASSERT(nLen >= 2);

	// Skip over any leading white space characters.
	pszHeader = SkipLWSP(pszHeader);
	if (*pszHeader == '\0') {
    	// We're at the end of the string.
		*pszBuffer = '\0';
		*piType = TOKEN_END;
		return pszHeader;
	}
	else if (*pszHeader == '"') {
		// Parse some quoted text.
		pszHeader = LexQText(pszHeader + 1, pszBuffer, nLen);
		// See if we stopped because of terminating quote.
		if (*pszHeader == '"') {
			// Yes, so everything was okay.
			*piType = TOKEN_QSTRING;
			return pszHeader + 1;
		}
		else {
			// Nope.  Signal error.
			*piType = TOKEN_ERROR;
			return pszHeader;
		}
	}
	else if (*pszHeader == '[') {
		// Parse a domain literal.
		pszHeader = LexDText(pszHeader + 1, pszBuffer, nLen);
		// See if we stopped because of terminating ']'.
		if (*pszHeader == ']') {
			// Yes, so everything was okay.
			*piType = TOKEN_DOMLIT;
			return pszHeader + 1;
		}
		else {
			// Nope.  Signal error.
			*piType = TOKEN_ERROR;
			return pszHeader;
		}
	}
	else if (*pszHeader == '(') {
		// Parse a comment.
		pszHeader = LexCText(pszHeader + 1, pszBuffer, nLen);
		// See if we stopped because of terminating ')'.
		if (*pszHeader == ')') {
			// Yes, so everything was okay.
			*piType = TOKEN_DOMLIT;
			return pszHeader + 1;
		}
		else {
			// Nope.  Signal error.
			*piType = TOKEN_ERROR;
			return pszHeader;
		}
	}
	else if (IsCTL(*pszHeader)) {
		*piType = TOKEN_CTL;
		pszBuffer[0] = *pszHeader;
		pszBuffer[1] = '\0';
        return pszHeader + 1;
	}
	else if (bMIME) {
		// Parse MIME tokens and tspecials.
		if (IsTSPECIAL(*pszHeader)) {
			*piType = TOKEN_TSPECIAL;
			pszBuffer[0] = *pszHeader;
			pszBuffer[1] = '\0';
			return pszHeader + 1;
		}
		else {
			pszHeader = LexMIMEToken(pszHeader, pszBuffer, nLen);
			*piType = TOKEN_MIMETOKEN;
            return pszHeader;
		}
	}
	else {
		// Parse RFC 822 atoms and specials.
		if (IsSPECIAL(*pszHeader)) {
			*piType = TOKEN_SPECIAL;
			pszBuffer[0] = *pszHeader;
			pszBuffer[1] = '\0';
			return pszHeader + 1;
		}
		else {
			pszHeader = LexAtom(pszHeader, pszBuffer, nLen);
			*piType = TOKEN_ATOM;
            return pszHeader;
        }
	}
}

