/*
	File: MAILBOX.CPP
	Description:
		Implements the mailbox window for Mail Client.
	Copyright 1995, David G. Roberts
*/

#include <ctype.h>
#include <dos.h>
#include <io.h>
#include <stdlib.h>
#include <time.h>
#include <fstream.h>
#include <windows.h>
#include <windowsx.h>
#include "mailcli.rh"
#include "myassert.h"
#include "cwinsock.h"
#include "clist.h"
#include "cstrlist.h"
#include "cpop3.h"
#include "rfc822.h"
#include "parsdate.h"
#include "utility.h"

#define DECLARE
#include "mailclig.h"

/*
	CONSTANTS
*/
#define LBCTLID		1
#define WM_CPOP3DONE	WM_USER

/*
	MACROS
*/
#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))

/*
	TYPES
*/

// A little collection structure to hold summary data about
// a particular message.  Note that the collection deletes all the
// string members upon destruction.  Make copies beforehand!
struct SUMMARYINFO {
	char *	pszFileName;
	char *	pszFrom;
	char *	pszSubject;
	time_t	time;

	SUMMARYINFO() : pszFileName(NULL), pszFrom(NULL),
		pszSubject(NULL), time(0) {};
	SUMMARYINFO(char * pszfn, char * pszfr, char * pszsub, time_t d) :
		pszFileName(pszfn), pszFrom(pszfr), pszSubject(pszsub),
		time(d) {};
	~SUMMARYINFO() {
		if (pszFileName != NULL) delete[] pszFileName;
		if (pszFrom != NULL) delete[] pszFrom;
		if (pszSubject != NULL) delete[] pszSubject;
	};
};

// A list class to manage SUMMARYINFO data.  Derived from CList.
class CSummaryList : public CList {
protected:
	virtual int Compare(void * pvThing1, void * pvThing2);
	virtual void DeleteData(void * pvThing);
public:
	CSummaryList() { };
	~CSummaryList();
	void Add(SUMMARYINFO * psi) { CList::Add((void *) psi); };
	void Insert(SUMMARYINFO * psi) { CList::Insert((void *) psi); };
	SUMMARYINFO * First(void)
		{ return (SUMMARYINFO *) CList::First(); };
	SUMMARYINFO * Next(void)
		{ return (SUMMARYINFO *) CList::Next(); };
};

CSummaryList::~CSummaryList()
{
	// Delete all the data in the list.
	SUMMARYINFO * psi = First();
	while (psi != NULL) {
		delete psi;
		psi = Next();
	}
}

int CSummaryList::Compare(void * pvThing1, void *pvThing2)
{
	SUMMARYINFO * psi1 = (SUMMARYINFO *) pvThing1;
	SUMMARYINFO * psi2 = (SUMMARYINFO *) pvThing2;

	if (psi1->time < psi2->time)
		return -1;
	else if (psi1->time == psi2->time)
		return 0;
	else
		return 1;
}

void CSummaryList::DeleteData(void * pvThing)
{
	delete (SUMMARYINFO *) pvThing;
}

// A little collection structure to hold some data that is associated
// with the mailbox window.
struct MAILBOXDATA {
	HWND				hwndLB;
	BOOL				bRetrieving;
	CSummaryList *		pslSummaryList;
	CPOP3WithWinMsg *	pcpop3;
	CStringList *		pslFileList;

	MAILBOXDATA() : hwndLB(NULL), bRetrieving(FALSE),
		pslSummaryList(new CSummaryList), pcpop3(NULL),
		pslFileList(NULL) {};
	~MAILBOXDATA();
};

MAILBOXDATA::~MAILBOXDATA()
{
	if (bRetrieving) {
		bRetrieving = FALSE;
		pcpop3->Abort(FALSE);
		delete pcpop3;
		delete pslFileList;
	}
	delete pslSummaryList;
}

/*
	LOCAL FUNCTIONS
*/
static void OnCommand(HWND hwnd, int id, HWND hwndCtl,
	UINT codeNotify);
static void OnCPOP3Done(HWND, UINT, WPARAM, LPARAM);
static BOOL OnCreate(HWND, CREATESTRUCT FAR*);
static void OnDestroy(HWND);
static void OnMDIActivate(HWND, BOOL, HWND, HWND);
static BOOL OnQueryEndSession(HWND hwnd);
static void OnSize(HWND, UINT, int, int);

// MENU HANDLERS
static void OnMailDelete(HWND);
static void OnMailRetrieve(HWND);
static void OnMailView(HWND);

// UTILITY FUNCTIONS
void ReadMailbox(HWND);
void WriteMailbox(HWND);
void CreateMailboxItems(HWND);
SUMMARYINFO * GetSummaryInfo(char *);
char * GetHeaderData(char *, char *);

/*********************************************************************
	WINDOW AND DIALOG BOX PROCEDURES
*********************************************************************/

/*
	Function: MailboxWndProc
	Description:
		Window procedure for the mailbox window class.
*/
LRESULT CALLBACK MailboxWndProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	switch (msg) {
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
		HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
		HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
		HANDLE_MSG(hwnd, WM_MDIACTIVATE, OnMDIActivate);
		HANDLE_MSG(hwnd, WM_QUERYENDSESSION, OnQueryEndSession);
		HANDLE_MSG(hwnd, WM_SIZE, OnSize);
		case WM_CPOP3DONE:
			OnCPOP3Done(hwnd, msg, wParam, lParam);
        	return 0;
	}
    return DefMDIChildProc(hwnd, msg, wParam, lParam);
}

/*********************************************************************
	MESSAGE HANDLING FUNCTIONS
*********************************************************************/

void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	if (hwndCtl == NULL) {
		// It's a window message.
		switch (id) {
			case IDM_MAILRETRIEVE:
				OnMailRetrieve(hwnd);
				return;
			case IDM_MAILVIEWMESSAGE:
				OnMailView(hwnd);
				return;
			case IDM_MAILDELETEMESSAGE:
				OnMailDelete(hwnd);
				return;
		}
	}
	else {
		ASSERT(id == LBCTLID);
		// It's from our control.
		if (codeNotify == LBN_DBLCLK) {
			OnMailView(hwnd);
			return;
		}
	}
	// Make sure the default child proc sees things we don't
	// process ourselves!
	FORWARD_WM_COMMAND(hwnd, id, hwndCtl, codeNotify,
		DefMDIChildProc);
}

void OnCPOP3Done(HWND hwnd, UINT msg, WPARAM wToken, LPARAM lResult)
{
	UNUSED_PARAMETER(msg);
	UNUSED_PARAMETER(wToken);

	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *)GetWindowLong(hwnd, 0);

	// We're done retrieving.  Reenable the menu item and delete
	// the POP3 object.
	lpdata->bRetrieving = FALSE;
	EnableMenuItem(g_hMailboxMenu, IDM_MAILRETRIEVE, MF_ENABLED);
	delete lpdata->pcpop3;
	lpdata->pcpop3 = NULL;
	// If something went wrong, tell the user.
	if (lResult != CPOP3::RESULT_SUCCESS) {
		MessageBox(hwnd, "Error retrieving POP mail.", "POP Error",
			MB_OK | MB_ICONEXCLAMATION);
		delete lpdata->pslFileList;
		lpdata->pslFileList = NULL;
		return;
	}
	char * pszFileName = lpdata->pslFileList->First();
	// See if we retrieved any messages.
	if (pszFileName == NULL) {
		MessageBox(hwnd, "No messages retrieved.", "POP Result",
			MB_OK | MB_ICONEXCLAMATION);
		delete lpdata->pslFileList;
		lpdata->pslFileList = NULL;
		return;
	}
	int iMsgCount = 0;
	while (pszFileName != NULL) {
		iMsgCount++;
		// Create some summary data from the file.
		SUMMARYINFO * psiInfo = GetSummaryInfo(pszFileName);
		// Insert it into the summary list (which will order it
		// by the time field).
        lpdata->pslSummaryList->Insert(psiInfo);
        // Get the next file name.
		pszFileName = lpdata->pslFileList->Next();
	}
	// Update the mailbox on disk.
    WriteMailbox(hwnd);
	// Update the list box.
	CreateMailboxItems(hwnd);
    // Tell the user what we retrieved.
	char szMessage[80];
	if (iMsgCount == 1)
		lstrcpy(szMessage, "1 message retrieved.");
	else
		wsprintf(szMessage, "%d messages retrieved.", iMsgCount);
	MessageBox(hwnd, szMessage, "POP Result",
		MB_OK | MB_ICONEXCLAMATION);
	delete lpdata->pslFileList;
	lpdata->pslFileList = NULL;
}

BOOL OnCreate(HWND hwnd, CREATESTRUCT FAR* lpCreateStruct)
{
	UNUSED_PARAMETER(lpCreateStruct);

	// Allocate some private window data for ourselves.
	MAILBOXDATA FAR * lpdata = new MAILBOXDATA;
	SetWindowLong(hwnd, 0, (LONG) lpdata);
	// Create a listbox control to do our work for us.
	lpdata->hwndLB = CreateWindow(
		"LISTBOX",					// class name
		NULL,						// window name
		WS_CHILD | WS_VISIBLE |		// window styles
		LBS_NOTIFY | WS_VSCROLL,
		0,0,0,0,					// location and size
		hwnd,						// parent
		(HMENU) LBCTLID,			// child id, not really a menu
		g_hInstance,				// program instance
		NULL);						// creation data
	// Set the listbox control's font to a fixed pitch.
	HFONT hfont = GetStockFont(SYSTEM_FIXED_FONT);
	SetWindowFont(lpdata->hwndLB, hfont, FALSE);

	// Disable the menu item that allows creation of this window
	// type.  This means there can only be one mailbox window at
	// a time.
	EnableMenuItem(g_hFrameMenu, IDM_FILEOPENMAILBOX, MF_GRAYED);
	EnableMenuItem(g_hMailboxMenu, IDM_FILEOPENMAILBOX, MF_GRAYED);
	EnableMenuItem(g_hViewerMenu, IDM_FILEOPENMAILBOX, MF_GRAYED);
	EnableMenuItem(g_hEditorMenu, IDM_FILEOPENMAILBOX, MF_GRAYED);

	// Read the mailbox off disk and set it up.
	ReadMailbox(hwnd);
	CreateMailboxItems(hwnd);

	return TRUE;
}

void OnDestroy(HWND hwnd)
{
	// Deallocate our private window data before we go away.
	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);
	delete lpdata;

    // Reenable File|Open Mailbox
	EnableMenuItem(g_hFrameMenu, IDM_FILEOPENMAILBOX, MF_ENABLED);
	EnableMenuItem(g_hMailboxMenu, IDM_FILEOPENMAILBOX, MF_ENABLED);
	EnableMenuItem(g_hViewerMenu, IDM_FILEOPENMAILBOX, MF_ENABLED);
	EnableMenuItem(g_hEditorMenu, IDM_FILEOPENMAILBOX, MF_ENABLED);
}

void OnMDIActivate(HWND hwnd, BOOL fActive, HWND hwndActivate,
	HWND hwndDeactivate)
{
	UNUSED_PARAMETER(hwndActivate);
	UNUSED_PARAMETER(hwndDeactivate);

	if (fActive) {
		// Put our menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hMailboxMenu,
				g_hMailboxWindowMenu));
		MAILBOXDATA FAR * lpdata =
			(MAILBOXDATA FAR*) GetWindowLong(hwnd, 0);
		// Enable/disable the Mail|Send menu item.
		if (lpdata->bRetrieving) {
			EnableMenuItem(g_hMailboxMenu, IDM_MAILRETRIEVE,
				MF_GRAYED);
		}
		else {
			EnableMenuItem(g_hMailboxMenu, IDM_MAILRETRIEVE,
				MF_ENABLED);
		}
		// Make sure keystrokes go to the listbox control.
		SetFocus(lpdata->hwndLB);
	}
	else {
		// We're being deactivated, so put the main menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hFrameMenu, g_hFrameWindowMenu));
	}
	DrawMenuBar(g_hFrame);
}

BOOL OnQueryEndSession(HWND hwnd)
{
	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);
	if (lpdata->bRetrieving) {
		SendMessage(g_hClient, WM_MDIRESTORE, (WPARAM) hwnd, 0);
		int iResult = MessageBox(hwnd,
			"Retrieving mail.  Close anyway?",
			"Close Anyway?", MB_YESNO | MB_ICONQUESTION);
		if (iResult == IDYES)
			return TRUE;
		else
        	return FALSE;
	}
	else {
		return TRUE;
	}
}

void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
	UNUSED_PARAMETER(state);

	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);
	MoveWindow(lpdata->hwndLB, 0, 0, cx, cy, TRUE);
	// Must make sure we forward this to the default child proc
	// as well!!
	FORWARD_WM_SIZE(hwnd, state, cx, cy, DefMDIChildProc);
}

/*********************************************************************
	MENU HANDLERS
*********************************************************************/

void OnMailRetrieve(HWND hwnd)
{
	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);

	lpdata->pcpop3 = new CPOP3WithWinMsg(hwnd, WM_CPOP3DONE, 0);
	lpdata->pslFileList = new CStringList;
	lpdata->pcpop3->RetrieveMessages(g_pszPOPServer, g_ushPOPPort,
		g_pszMailbox, g_pszPassword, g_szMessageDir,
		lpdata->pslFileList);
    lpdata->bRetrieving = TRUE;
	EnableMenuItem(g_hMailboxMenu, IDM_MAILRETRIEVE, MF_GRAYED);
}

void OnMailView(HWND hwnd)
{
	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);
	// Get the current selection index from the list box.
	int iCurSel = ListBox_GetCurSel(lpdata->hwndLB);
	if (iCurSel < 0) {
		MessageBox(hwnd, "You must make a selection first.",
			"Whoops!", MB_OK | MB_ICONHAND);
		return;
	}
	// Find the right summary info entry in the list that
	// corresponds to the selection index.
	SUMMARYINFO * psi = lpdata->pslSummaryList->First();
	for (int i = 0; i < iCurSel; i++, psi = lpdata->pslSummaryList->Next())
		; // Null loop body.
	ASSERT(psi != NULL);
	// Create the viewer window and give it the message filename.
	MDICREATESTRUCT mdicreate;
	mdicreate.szClass	= g_szViewerClass;
	mdicreate.szTitle	= "Viewer";
	mdicreate.hOwner	= g_hInstance;
	mdicreate.x			= CW_USEDEFAULT;
	mdicreate.y			= CW_USEDEFAULT;
	mdicreate.cx		= CW_USEDEFAULT;
	mdicreate.cy		= CW_USEDEFAULT;
	mdicreate.style		= 0;
	mdicreate.lParam	=
		(LPARAM)(LPSTR)DuplicateString(psi->pszFileName);

	// Send a message to the client window to create the new child.
	SendMessage(g_hClient, WM_MDICREATE, 0,
		(LPARAM) (LPMDICREATESTRUCT) &mdicreate);

	return;
}

void OnMailDelete(HWND hwnd)
{
	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);
	// Get the current selection index from the list box.
	int iCurSel = ListBox_GetCurSel(lpdata->hwndLB);
	if (iCurSel < 0) {
		MessageBox(hwnd, "You must make a selection first.",
			"Whoops!", MB_OK | MB_ICONHAND);
		return;
	}
	// Find the right summary info entry in the list that
	// corresponds to the selection index.
	SUMMARYINFO * psi = lpdata->pslSummaryList->First();
	for (int i = 0; i < iCurSel; i++, psi =
		lpdata->pslSummaryList->Next())
		; // Null loop body.
	ASSERT(psi != NULL);
	// Delete the disk message file.
	char szFullName[MAX_PATH_LENGTH];
	lstrcpy(szFullName, g_szMessageDir);
	lstrcat(szFullName, "\\");
	lstrcat(szFullName, psi->pszFileName);
    unlink(szFullName);
    // Delete the node.
    lpdata->pslSummaryList->Delete(psi);
	// Update the mailbox on disk.
    WriteMailbox(hwnd);
	// Update the list box.
	CreateMailboxItems(hwnd);
}

/*********************************************************************
	UTILITY FUNCTIONS
*********************************************************************/

void ReadMailbox(HWND hwnd)
{
	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);
	ifstream ifsFile(g_szMailboxFile);
	const int LINESIZE = 256;
	char szLine[LINESIZE];

	while (ifsFile.peek() != EOF) {
		SUMMARYINFO * psi = new SUMMARYINFO;
		ifsFile.getline(szLine, LINESIZE);
		psi->pszFileName = DuplicateString(szLine);
		ifsFile.getline(szLine, LINESIZE);
		psi->pszFrom = DuplicateString(szLine);
		ifsFile.getline(szLine, LINESIZE);
		psi->pszSubject = DuplicateString(szLine);
		ifsFile.getline(szLine, LINESIZE);
		psi->time = (time_t) atol(szLine);
        lpdata->pslSummaryList->Insert(psi);
	}
	ifsFile.close();
}

void WriteMailbox(HWND hwnd)
{
	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);
	ofstream ofsFile(g_szMailboxFile);

	SUMMARYINFO * psi = lpdata->pslSummaryList->First();
	while (psi != NULL) {
		ofsFile << psi->pszFileName << '\n';
		ofsFile << psi->pszFrom << '\n';
		ofsFile << psi->pszSubject << '\n';
		ofsFile << (long) psi->time << '\n';
		psi = lpdata->pslSummaryList->Next();
	}
	ofsFile.close();
}

void CreateMailboxItems(HWND hwnd)
{
	const char *Months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	const int MAXFROM = 20;
	const int MAXSUBJECT = 50;

	// Clear out the list box.
	MAILBOXDATA FAR * lpdata =
		(MAILBOXDATA FAR *) GetWindowLong(hwnd, 0);
	HWND hwndLB = lpdata->hwndLB;
	SetWindowRedraw(hwndLB, FALSE);
	SendMessage(hwndLB, LB_RESETCONTENT, 0, 0);

	// Create a summary line for each item in the summary info list.
	SUMMARYINFO * psi = lpdata->pslSummaryList->First();
	while (psi != NULL) {
		// Format the summary line string using the information
        // in the summary record.
		char szSummaryLine[100];
		char szTruncFrom[MAXFROM + 1];
		char szTruncSubject[MAXSUBJECT + 1];
		struct tm * ptm = localtime(&(psi->time));
		lstrcpyn(szTruncFrom, psi->pszFrom, MAXFROM);
        lstrcpyn(szTruncSubject, psi->pszSubject, MAXSUBJECT);
		wsprintf(szSummaryLine, "%02d-%s-%02d  %-20s  %-50s",
			ptm->tm_mday, (LPSTR)Months[ptm->tm_mon], ptm->tm_year,
			(LPSTR)szTruncFrom, (LPSTR)szTruncSubject);
		ListBox_AddString(hwndLB, szSummaryLine);
		psi = lpdata->pslSummaryList->Next();
	}
	SetWindowRedraw(hwndLB, TRUE);
}

SUMMARYINFO * GetSummaryInfo(char * pszFileName)
{
	SUMMARYINFO * psiResult = new SUMMARYINFO;

	// Save off the file name.
	psiResult->pszFileName = DuplicateString(pszFileName);

	// Get the From: header.
	char * pszHeaderData = GetHeaderData("From:", pszFileName);
	if (pszHeaderData == NULL) {
		// If we couldn't find the header, create a zero-length
		// null terminated string.
		pszHeaderData = new char[1];
		*pszHeaderData = '\0';
	}
	psiResult->pszFrom = pszHeaderData;

	// Get the subject header.
	pszHeaderData = GetHeaderData("Subject:", pszFileName);
	if (pszHeaderData == NULL) {
		// If we couldn't find the header, create a zero-length
		// null terminated string.
		pszHeaderData = new char[1];
		*pszHeaderData = '\0';
	}
	psiResult->pszSubject = pszHeaderData;

	// Get the date header.
	pszHeaderData = GetHeaderData("Date:", pszFileName);
	if (pszHeaderData == NULL) {
		// If there was no date, pick a default time.
		psiResult->time = 0;
	}
	else {
    	// Else parse the date to return a time_t value.
		psiResult->time = ParseDateString(pszHeaderData);
        delete[] pszHeaderData;
	}

    // Return the SUMMARYINFO structure to the caller.
    return psiResult;
}

char * GetHeaderData(char * pszHeader, char * pszFileName)
{
	int i;
	char szFullName[MAX_PATH_LENGTH];
	lstrcpy(szFullName, g_szMessageDir);
	lstrcat(szFullName, "\\");
	lstrcat(szFullName, pszFileName);

	ifstream ifsMessage(szFullName);
	char szLineBuffer[1001];

	while (ifsMessage.peek() != EOF) {
		// Slurp up a line of data from the file.
		ifsMessage.getline(szLineBuffer, sizeof(szLineBuffer));
		// See if the line was simply a CRLF.
		if (lstrlen(szLineBuffer) == 0) {
			// If so, we reached the end of the headers before
			// finding the one we were looking for.  Exit the
			// while loop and return NULL.
			return NULL;
		}
		// See if this is the header we want.
		BOOL bMatch = TRUE;
		for (i = 0; i < lstrlen(pszHeader); i++) {
			if (tolower(pszHeader[i]) != tolower(szLineBuffer[i])) {
				// Nope, something didn't match.
				bMatch = FALSE;
				break;
			}
		}
		// If we matched, break out.
		if (bMatch)
			break;
	}
	// At this point, i==lstrlen(pszHeader).
	// Advance i past any whitespace that might have followed
	// the colon after the header name.
	while (IsLWSP(szLineBuffer[i]))
		i++;
	// Now, shrink the line down, eliminating the header name,
	// colon, and any following white space but retaining all
	// the info we really care about.
	char * pszResult = new char[lstrlen(szLineBuffer) - i + 1];
	int j = 0;
	while (szLineBuffer[i] != '\0') {
		pszResult[j++] = szLineBuffer[i++];
	}
    pszResult[j] = '\0';
	// Now, check to see if the following line is folded.  If
	// so, we should append it on to the current data.
	while (ifsMessage.peek() != EOF) {
		// We still have more file, so read in another line.
		ifsMessage.getline(szLineBuffer, sizeof(szLineBuffer));
		// See if it starts with linear white space.  If not,
		// then the line is not folded further and we're done.
		if (!IsLWSP(szLineBuffer[0]))
			break;
		// Okay, the line is folded.
		// See how long the combined line will be when we've
		// appended this line to the previous ones.
		int iLength = lstrlen(pszResult) + lstrlen(szLineBuffer);
		char * pszTemp = new char[iLength + 1];
		lstrcpy(pszTemp, pszResult);
		lstrcat(pszTemp, szLineBuffer);
		delete[] pszResult;
		pszResult = pszTemp;
	}

	return pszResult;
}


