/*
	File: MAILEDIT.CPP
	Description:
		Implements the editor window class for Mail Client.
	Copyright 1995, David G. Roberts
	
	11/11/95: Fixed GetTempFileName bug in OnMailSend for Win 32.
*/

#include <dos.h>
#include <fstream.h>
#include <stdlib.h>
#include <windows.h>
#include <windowsx.h>
#include "mailcli.rh"
#include "myassert.h"
#include "cwinsock.h"
#include "csmtp.h"
#include "caddrlst.h"
#include "rfc822.h"
#include "cuuendec.h"
#include "cbase64.h"

#define DECLARE
#include "mailclig.h"

/*
	CONSTANTS
*/
#define EDITCTLID		1
#define WM_CSMTPDONE	WM_USER
const char cszSendingText[] = ": Sending...";

/*
	TYPES
*/

/*
	Structure: EDITORDATA
	Description:
		A collection of private data allocated in WM_CREATE
		processing for the editor MDI child window.  A pointer to
		this structure is stored in the Windows data structures.
*/
struct EDITORDATA {
	HWND		hwndEdit;
	char *		pszTempFile;
	CAddrList * palAddrList;
	CSMTPWithWinMsg * pcsmtp;
	ifstream	ifsMsgStream;
	BOOL		bSending;

	EDITORDATA() : hwndEdit(NULL), pszTempFile(NULL),
		palAddrList(NULL), pcsmtp(NULL), bSending(FALSE) {};
	~EDITORDATA();
};

EDITORDATA::~EDITORDATA()
{
	// Note: delete the CSMTP object before the address list and
	// before we close the stream.  The CSMTP object could be using
	// these things.
	if (bSending) {
		bSending = FALSE;
		delete pcsmtp;
		pcsmtp = NULL;
		ifsMsgStream.close();
		unlink(pszTempFile);
		delete[] pszTempFile;
		pszTempFile = NULL;
		delete palAddrList;
		palAddrList = NULL;
	}

}

// A macro to help us the get the pointer from the extra window
// storage area that compensates for the pointer size differences
// of different memory models.  Note that we have to the size test
// in real code rather than a preprocessor test because MSVC 1.5 
// doesn't support the "sizeof" operator in preprocessor
// expressions.
#define GETEDITORDATA(hwnd) \
	( (sizeof(EDITORDATA*) != 2) ? \
	(EDITORDATA *)GetWindowLong(hwnd, 0) : \
	(EDITORDATA *)GetWindowWord(hwnd, 0) )

/*
	Structure: ADDRESSDLGINFO
	Description:
		The set of data returned from the addressing info dialog.
		The structure simply collects this info together so that
		a simple pointer can be passed to the dialog.
*/
struct ADDRESSDLGINFO {
	CAddrList *	pToAddrList;
	CAddrList *	pCcAddrList;
	char *		pszSubject;
	char *		pszFullName;

	ADDRESSDLGINFO() : pToAddrList(NULL), pCcAddrList(NULL),
		pszSubject(NULL), pszFullName(NULL) {};
	// Make sure memory gets released.
	~ADDRESSDLGINFO() {
		if (pToAddrList != NULL) delete pToAddrList;
		if (pCcAddrList != NULL) delete pCcAddrList;
		if (pszSubject != NULL) delete[] pszSubject;
		if (pszFullName != NULL) delete[] pszFullName;
	};
};

/*
	LOCAL FUNCTIONS
*/
static void OnCommand(HWND hwnd, int id, HWND hwndCtl,
	UINT codeNotify);
static BOOL OnCreate(HWND, CREATESTRUCT FAR*);
static void OnCSMTPDone(HWND, UINT, WPARAM, LPARAM);
static void OnDestroy(HWND);
static void OnMDIActivate(HWND, BOOL, HWND, HWND);
static BOOL OnQueryEndSession(HWND hwnd);
static void OnSize(HWND, UINT, int, int);

// MENU HANDLERS
static void OnMailSend(HWND);
static void OnMailUuencode(HWND hwnd);

// UTILITY FUNCTIONS
static BOOL AddressStringToList(char *, CAddrList *);
static void OutputAddrList(char *, CAddrList *, CAddrList *,
	ofstream&);
static void OutputMessageText(HWND, ofstream&);
static void ConvertToQP(char *, int, char *);
static void QuoteCharacter(unsigned char, char *);
static void CreateBoundary(char *, int);
static void OutputAttachment(char *, ofstream&);

/*********************************************************************
	WINDOW AND DIALOG BOX PROCEDURES
*********************************************************************/

/*
	Function: EditorWndProc
	Description:
		Window procedure for the editor MDI child window.
		Dispatches all the window messages to the correct handling
		routines.
*/
LRESULT CALLBACK EditorWndProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	switch (msg) {
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
		HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
		HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
		HANDLE_MSG(hwnd, WM_MDIACTIVATE, OnMDIActivate);
		HANDLE_MSG(hwnd, WM_QUERYENDSESSION, OnQueryEndSession);
		HANDLE_MSG(hwnd, WM_SIZE, OnSize);
		case WM_CSMTPDONE:
			OnCSMTPDone(hwnd, msg, wParam, lParam); return 0;
	}
    return DefMDIChildProc(hwnd, msg, wParam, lParam);
}

/*
	Function: AddressDlgProc
	Description:
		The dialog procedure for the addressing info dialog.
		Most of the code is devoted to extracting all the info
		when the user hits the OK button.
*/
BOOL CALLBACK AddressDlgProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	int id;
	ADDRESSDLGINFO FAR * lpAddrInfo;
	int iTextLen;
	char * pszText;
	BOOL bParsedOK;
	OPENFILENAME ofn;
	static char *szFilter[] = { "All Files (*.*)", "*.*", "", "" };
	char szFullName[MAX_PATH_LENGTH];
	char szFileName[MAX_FILENAME_LENGTH];
	BOOL bResult;

	switch (msg) {
		case WM_COMMAND:
			id = GET_WM_COMMAND_ID(wParam, lParam);
			switch (id) {
				case IDOK:
					// Retrieve the pointer we stuffed away when we
					// were first created.
					lpAddrInfo = (ADDRESSDLGINFO FAR *)
						GetWindowLong(hwnd, DWL_USER);
					// Get text length of To: box.
					iTextLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_TO));
					// Allocate string.
					pszText = new char[iTextLen + 1];
					// Get text from To: box into string.
					GetDlgItemText(hwnd, IDD_TO, pszText,
						iTextLen + 1);
					// Allocate a list of addresses.
					lpAddrInfo->pToAddrList = new CAddrList;
					// Parse the addresses.
					bParsedOK = AddressStringToList(pszText,
						lpAddrInfo->pToAddrList);
					delete[] pszText;
					// If things didn't go right, then delete the
					// allocated memory and return.
					if (!bParsedOK) {
						delete lpAddrInfo->pToAddrList;
                        SetEditErrorFocus(hwnd, IDD_TO);
						return TRUE;
					}

					// Now do the same for the Cc text.
					// Get text length of Cc: box.
					iTextLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_CC));
					// Allocate string.
					pszText = new char[iTextLen + 1];
					// Get text from To: box into string.
					GetDlgItemText(hwnd, IDD_CC, pszText,
						iTextLen + 1);
					// Allocate a list of addresses.
					lpAddrInfo->pCcAddrList = new CAddrList;
					// Parse the addresses.
					bParsedOK = AddressStringToList(pszText,
						lpAddrInfo->pCcAddrList);
					delete[] pszText;
					// If things didn't go right, then delete the
					// allocated memory and return.
					if (!bParsedOK) {
						delete lpAddrInfo->pToAddrList;
						delete lpAddrInfo->pCcAddrList;
						SetEditErrorFocus(hwnd, IDD_CC);
						return TRUE;
					}
					// Now extract the subject text.
					iTextLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_SUBJECT));
					lpAddrInfo->pszSubject = new char[iTextLen + 1];
					GetDlgItemText(hwnd, IDD_SUBJECT,
						lpAddrInfo->pszSubject, iTextLen + 1);
					// Now extract the attachment filename text.
					iTextLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_ATTACHMENT));
					lpAddrInfo->pszFullName = new char[iTextLen + 1];
					GetDlgItemText(hwnd, IDD_ATTACHMENT,
						lpAddrInfo->pszFullName, iTextLen + 1);

					EndDialog(hwnd, TRUE);
					return TRUE;
				case IDCANCEL:
					EndDialog(hwnd, FALSE);
					return TRUE;
				case IDD_BROWSE:
					// Set up the OPENFILENAME structure in
					// preparation for browsing.

					ofn.lStructSize			= sizeof(OPENFILENAME);
					ofn.hwndOwner			= hwnd;
					ofn.hInstance			= NULL;
					ofn.lpstrFilter			= szFilter[0];
					ofn.lpstrCustomFilter	= NULL;
					ofn.nMaxCustFilter		= 0;
					ofn.nFilterIndex		= 1;
					ofn.lpstrFile			= szFullName;
					ofn.nMaxFile			= sizeof(szFullName);
					lstrcpy(szFullName, "");
					ofn.lpstrFileTitle		= szFileName;
					ofn.nMaxFileTitle		= sizeof(szFileName);
                    lstrcpy(szFileName, "");
					ofn.lpstrInitialDir		= NULL;
					ofn.lpstrTitle			= "Attach File";
					ofn.Flags				= OFN_HIDEREADONLY;
					ofn.nFileOffset			= 0;
					ofn.nFileExtension		= 0;
					ofn.lpstrDefExt			= NULL;
					ofn.lCustData			= 0;
					ofn.lpfnHook			= NULL;
					ofn.lpTemplateName		= NULL;

					bResult = GetOpenFileName(&ofn);

					if (bResult) {
						SetWindowText(GetDlgItem(hwnd, IDD_ATTACHMENT),
							szFullName);
					}
					break;
			} // End switch (id)
			break;
		case WM_INITDIALOG:
			// Stuff away the pointer to the user data.
			SetWindowLong(hwnd, DWL_USER, (LONG) lParam);
			// Set the focus to the To: edit control so the user can
			// immediately start typing.
			SetFocus(GetDlgItem(hwnd, IDD_TO));
			return 0;  // Return 0 because we set focus to IDD_TO.
	} // End switch (msg)
	return FALSE;
}

/*********************************************************************
	MESSAGE HANDLING FUNCTIONS
*********************************************************************/

/*
	Function: OnCommand
	Description:
		Handles WM_COMMAND messages for the MDI child window.
*/
void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	EDITORDATA * pdata = GETEDITORDATA(hwnd);

	if (hwndCtl == NULL) {
		// It's a window message.
		switch (id) {
			case IDM_EDITCUT:
				SendMessage(pdata->hwndEdit, WM_CUT, 0, 0);
				return;
			case IDM_EDITCOPY:
				SendMessage(pdata->hwndEdit, WM_COPY, 0, 0);
				return;
			case IDM_EDITPASTE:
				SendMessage(pdata->hwndEdit, WM_PASTE, 0, 0);
				return;
			case IDM_MAILSEND:
				OnMailSend(hwnd);
				return;
			case IDM_MAILUUENCODE:
				OnMailUuencode(hwnd);
                return;
		}
	}
	else {
		ASSERT(id == EDITCTLID);
		// It's from our control.
		if (codeNotify == EN_ERRSPACE) {
			MessageBox(hwnd, "Editor out of space.", "Error",
				MB_OK | MB_ICONEXCLAMATION);
			return;
		}
	}
	// Make sure the default child proc sees things we don't
	// process ourselves!
	FORWARD_WM_COMMAND(hwnd, id, hwndCtl, codeNotify,
		DefMDIChildProc);
}

/*
	Function: OnCreate
	Description:
		Handles WM_CREATE messages.  The function creates the
		editor control child window and allocates a private data
		area associated with the MDI child window to hold a set
		of private variables.
*/
BOOL OnCreate(HWND hwnd, CREATESTRUCT FAR* lpCreateStruct)
{
	UNUSED_PARAMETER(lpCreateStruct);

	// Allocate some private window data for ourselves.
	EDITORDATA * pdata = new EDITORDATA;

	// Note that we must do the following test with real code
	// rather than in the preprocessor because MSVC 1.5
	// doesn't support the "sizeof" operator in preprocessor
	// expressions! Gak!  Note that one or the other of the
	// following SetWindowXXX statements may generate a
	// warning.  The warning may involve the fact that the condition
	// is always true, one of the lines will never be executed, or
	// that the pointer is being converted to shorter or longer type
    // than it should be.
	if (sizeof(EDITORDATA *) != 2)
		SetWindowLong(hwnd, 0, (LONG) pdata);
	else
		SetWindowWord(hwnd, 0, (WORD) pdata);

	// Create a multi-line edit control to do our work for us.
	pdata->hwndEdit = CreateWindow(
		"EDIT",						// class name
		NULL,						// window name
		WS_CHILD | WS_VISIBLE |		// window styles
		WS_HSCROLL | WS_VSCROLL |  ES_LEFT |
		ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL,
		0,0,0,0,					// location and size
		hwnd,						// parent
		(HMENU) EDITCTLID,			// child id, not really a menu
		g_hInstance,				// program instance
		NULL);						// creation data
	// Set the edit control's font to a fixed pitch.
	HFONT hfont = GetStockFont(SYSTEM_FIXED_FONT);
	SetWindowFont(pdata->hwndEdit, hfont, FALSE);
	return TRUE;
}

/*
	Function: OnCSMTPDone
	Description:
		Processes the private WM_CSMTPDONE window message that is
		sent by the CSMTP object when it completes its operation.
		This function takes care of deleting all the various data
		that was needed for CSMTP to operate.  If CSMTP was
		successful, this function deletes the editor child window
		(self destructs), otherwise it informs the user that an
		error has occurred and allows the user to retry the operation
		after correcting the problem.
*/
void OnCSMTPDone(HWND hwnd, UINT msg, WPARAM wToken, LPARAM lResult)
{
	UNUSED_PARAMETER(msg);
	UNUSED_PARAMETER(wToken);

	EDITORDATA * pdata = GETEDITORDATA(hwnd);

	pdata->bSending = FALSE;
	delete pdata->pcsmtp;
	pdata->pcsmtp = NULL;
	pdata->ifsMsgStream.close();
	unlink(pdata->pszTempFile);
	delete[] pdata->pszTempFile;
	pdata->pszTempFile = NULL;
	delete pdata->palAddrList;
	pdata->palAddrList = NULL;

	EnableMenuItem(g_hEditorMenu, IDM_MAILSEND, MF_ENABLED);

	if (lResult == CSMTP::RESULT_SUCCESS ||
		lResult == CSMTP::RESULT_ABORT) {
		SendMessage(hwnd, WM_CLOSE, 0, 0);
	}
	else {
		int iOldLength = GetWindowTextLength(hwnd);
		char * pszNewTitle = new char[iOldLength + 1];
		GetWindowText(hwnd, pszNewTitle, iOldLength + 1);
		int iNewLength = iOldLength - lstrlen(cszSendingText);
		pszNewTitle[iNewLength] = '\0';
		SetWindowText(hwnd, pszNewTitle);
		delete[] pszNewTitle;

		pdata->bSending = FALSE;

		SendMessage(g_hClient, WM_MDIRESTORE, (WPARAM) hwnd, 0);
        SendMessage(g_hClient, WM_MDIACTIVATE, (WPARAM) hwnd, 0);
		MessageBox(hwnd,
			"Could not send message!  Check configuration and "
			"addressing information and try again.", "Error",
			MB_OK | MB_ICONEXCLAMATION);
	}
}

/*
	Function: OnDestroy
	Description:
		Handles WM_DESTROY processing.  Simply deletes the private
		data associated with the window before it is destroyed.
*/
void OnDestroy(HWND hwnd)
{
	// Deallocate our private window data before we go away.
	EDITORDATA * pdata = GETEDITORDATA(hwnd);
	delete pdata;
}

/*
	Function: OnMDIActivate
	Description:
		Handles WM_MDIACTIVATE messages.  The function takes care of
		putting the editor menu in place and enabling/disabling
		the "Send" menu choice based on whether a SMTP session is
		already in progress for this message.  If the window is
		being deactivated, the frame menu is put in place.
*/
void OnMDIActivate(HWND hwnd, BOOL fActive, HWND hwndActivate,
	HWND hwndDeactivate)
{
	UNUSED_PARAMETER(hwndActivate);
	UNUSED_PARAMETER(hwndDeactivate);

	if (fActive) {
		// Put our menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hEditorMenu,
				g_hEditorWindowMenu));
		EDITORDATA * pdata = GETEDITORDATA(hwnd);
		// Enable/disable the Mail|Send menu item.
		if (pdata->bSending) {
			EnableMenuItem(g_hEditorMenu, IDM_MAILSEND, MF_GRAYED);
		}
		else {
			EnableMenuItem(g_hEditorMenu, IDM_MAILSEND, MF_ENABLED);
		}
		// Make sure keystrokes go to the edit control.
		SetFocus(pdata->hwndEdit);
	}
	else {
		// We're being deactivated, so put the main menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hFrameMenu,
				g_hFrameWindowMenu));
	}
	DrawMenuBar(g_hFrame);
}

/*
	Function: OnQueryEndSession
	Description:
		Handles WM_QUERYENDSESSION window messages.  The function
		throws up a warning message box if the user tries to close
		the window while the message is being sent via SMTP.
*/
BOOL OnQueryEndSession(HWND hwnd)
{
	EDITORDATA * pdata = GETEDITORDATA(hwnd);
	if (pdata->bSending) {
		SendMessage(g_hClient, WM_MDIRESTORE, (WPARAM) hwnd, 0);
		int iResult = MessageBox(hwnd,
			"Message still being sent.  Close anyway?",
			"Close Anyway?", MB_YESNO | MB_ICONQUESTION);
		if (iResult == IDYES)
			return TRUE;
		else
        	return FALSE;
	}
	else {
		return TRUE;
	}
}

/*
	Function: OnSize
	Description:
		Handle WM_SIZE window messages.  Makes sure that the edit
		control inside the window gets resized when the user
		resizes the child MDI window.
*/
void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
	UNUSED_PARAMETER(state);

	EDITORDATA * pdata = GETEDITORDATA(hwnd);
	MoveWindow(pdata->hwndEdit, 0, 0, cx, cy, TRUE);
	// Must make sure we forward this to the default child proc
	// as well!!
	FORWARD_WM_SIZE(hwnd, state, cx, cy, DefMDIChildProc);
}

/*********************************************************************
	MENU HANDLERS
*********************************************************************/

/*
	Function: OnMailSend
	Description:
		The function that kicks off the sending of a message.  The
		function first displays a dialog box that allows the user to
		edit the addess list and subject of the message.  Next, the
		function creates the RFC822 compliant message in a temporary
		file, which is then handed to the SMTP object to deliver.
*/
void OnMailSend(HWND hwnd)
{
	ADDRESSDLGINFO adi;

	// Force the user to be configured before progressing.
	if (!g_bConfigured) {
		MessageBox(hwnd, "You must configure the mail parameters "
			"before sending mail.", "Can't Send Message",
			MB_OK | MB_ICONSTOP);
		return;
	}

	// Put up the addressing dialog box.
	FARPROC lpfnAddressProc =
		MakeProcInstance((FARPROC) AddressDlgProc, g_hInstance);
	BOOL bResult = DialogBoxParam(g_hInstance, "ADDRESSING", hwnd,
		(DLGPROC) lpfnAddressProc,
		(LPARAM)(ADDRESSDLGINFO FAR *) &adi);
	FreeProcInstance(lpfnAddressProc);

	// Don't do anything if the user hit Cancel.
	if (bResult == FALSE)
		return;

	// Make sure it's addressed to somebody.
	if (adi.pToAddrList->First() == NULL &&
		adi.pCcAddrList->First() == NULL) {
		MessageBox(hwnd, "You must supply at least one destination "
			"address.", "Can't Send Message", MB_OK | MB_ICONSTOP);
		return;
	}

	// See if there is an attachment and if so, whether the file
	// exists or not.
	BOOL bAttachment;
	if (lstrcmp(adi.pszFullName, "") != 0) {
		bAttachment = TRUE;
		OFSTRUCT ofs;
		HFILE hfile = OpenFile(adi.pszFullName, &ofs, OF_EXIST);
		if (hfile == HFILE_ERROR) {
			MessageBox(hwnd, "Attachment file can't be opened.",
				"Can't Send Message", MB_OK | MB_ICONSTOP);
			return;
		}
	}
	else {
		bAttachment = FALSE;
	}

	// Capture the mouse, 'cause this could take more than an
	// instant.
	SetCapture(g_hFrame);
	HCURSOR hcursorSave = SetCursor(LoadCursor(NULL, IDC_WAIT));

	// Block sending.
	EDITORDATA * pdata = GETEDITORDATA(hwnd);
	pdata->bSending = TRUE;
	EnableMenuItem(g_hEditorMenu, IDM_MAILSEND, MF_GRAYED);

	// Fiddle with the window text and make the window iconic.
	int iOldLength = GetWindowTextLength(hwnd);
	int iNewLength = iOldLength + lstrlen(cszSendingText) + 1;
	char * pszNewTitle = new char[iNewLength];
	GetWindowText(hwnd, pszNewTitle, iOldLength + 1);
	lstrcat(pszNewTitle, cszSendingText);
	SetWindowText(hwnd, pszNewTitle);
	delete[] pszNewTitle;
	ShowWindow(hwnd, SW_MINIMIZE);
	SendMessage(g_hClient, WM_MDINEXT, (WPARAM)hwnd, 0);

	// Create the RFC822 message from the body text and the addressing
	// information the user has entered.

	// Get a temp file name from Windows.
	pdata->pszTempFile = new char[MAX_PATH_LENGTH];
	// 11/11/95: Fixed bug getting temp path in Win32.
#ifdef WIN32
	char TempPath[MAX_PATH_LENGTH];
	GetTempPath(sizeof(TempPath), TempPath);
	GetTempFileName(TempPath, "MAIL", 0, pdata->pszTempFile);
#else
	GetTempFileName(0, "MAIL", 0, pdata->pszTempFile);
#endif
	// Open a standard C++ output stream.
	ofstream ofsMsgStream(pdata->pszTempFile);

	// Compose the message.  Note that we use '\n' line terminators
	// rather than the explicit "\r\n" that RFC 822 requires.  The
	// standard end-of-line convention in DOS is actually "\r\n" but
	// not in C/C++.  The standard C/C++ libraries handle converting
	// our '\n' to "\r\n" as the file is written.  When it's read back
	// in in the CSMTP object, the "\r\n" in the file is converted
	// back to just '\n', which the CSMTP object then converts back
	// to "\r\n" for SMTP purposes.  It's a torturous route, but it
    // works. :-)
	char szDate[MAX_DATE_LENGTH];
	GetCurrentRFC822Date(szDate);
	ofsMsgStream << "Date: " << szDate << '\n';
	ofsMsgStream << "From: " << g_pszUserName << " <" <<
		g_pszUserAddress << ">\n";

	// Allocate the combined To: & CC: list that will be passed to the
	// SMTP object for delivery.
	pdata->palAddrList = new CAddrList;

	// Output the To: list.
	OutputAddrList("To: ", adi.pToAddrList, pdata->palAddrList,
		ofsMsgStream);

	// Output the Cc: list.
	OutputAddrList("Cc: ", adi.pCcAddrList, pdata->palAddrList,
		ofsMsgStream);

	// Write out the subject.
	if (lstrcmp(adi.pszSubject, "") != 0) {
		ofsMsgStream << "Subject: " << adi.pszSubject << '\n';
	}

	// Create a message ID.
	char szMessageID[MAX_MESSAGE_ID_LENGTH];
	GetMessageID(szMessageID);
	ofsMsgStream << "Message-ID: <" << szMessageID << ">\n";

	// We're doing MIME.
	ofsMsgStream << "MIME-Version: 1.0\n";

	// Some advertising... :-)
	ofsMsgStream << "X-Mail-Agent: Mail Client\n";
	ofsMsgStream << "X-Mail-Agent: From the book, \"Developing for "
		"the Internet with WinSock\"\n";
	ofsMsgStream << "X-Mail-Agent: By Dave Roberts\n";

	if (bAttachment == FALSE) {
		OutputMessageText(pdata->hwndEdit, ofsMsgStream);
	}
	else {
		// Create some boundary text.
		char szBoundary[30];
		CreateBoundary(szBoundary, sizeof(szBoundary));
		// Output the multipart/mixed content type with the
		// boundary.
		ofsMsgStream << "Content-Type: multipart/mixed;\n" <<
			"    boundary=\"" << szBoundary << "\"\n";
		ofsMsgStream << '\n';	// Blank line between header/body.
		// Output the boundary which starts a body part.
		ofsMsgStream << '\n' << "--" << szBoundary << '\n';
		// Output the text portion of the file.
		OutputMessageText(pdata->hwndEdit, ofsMsgStream);
		// Output the boundary which separates the text and the
		// attachment.
		ofsMsgStream << '\n' << "--" << szBoundary << '\n';
		// Output the attachment.
		OutputAttachment(adi.pszFullName, ofsMsgStream);
		// Output the boundary which terminates all body parts.
		ofsMsgStream << '\n' << "--" << szBoundary << "--\n";
	}

	ofsMsgStream.close();

	// Release the mouse.
	SetCursor(hcursorSave);
    ReleaseCapture();

	// Create the CSMTP object and let 'er rip!
	pdata->ifsMsgStream.open(pdata->pszTempFile);
	pdata->pcsmtp = new CSMTPWithWinMsg(hwnd, WM_CSMTPDONE, 0);
	pdata->pcsmtp->SendMessage(g_pszSMTPServer, g_ushSMTPPort,
		&(pdata->ifsMsgStream), g_pszUserAddress,
		pdata->palAddrList);
}

/*
	Function: OnMailUuencode
	Description:
		Handle the UUENCODE menu choice.  UUENCODE a file and
		insert it into the editing buffer.
*/
void OnMailUuencode(HWND hwnd)
{
	char szBuffer[82];
	CUUENC cuuenc(hwnd);

	// Capture the mouse, 'cause this could take more than an
	// instant.
	SetCapture(g_hFrame);
	HCURSOR hcursorSave = SetCursor(LoadCursor(NULL, IDC_WAIT));

	EDITORDATA * pdata = GETEDITORDATA(hwnd);

	// Turn off redraw so this goes a lot faster.
	SetWindowRedraw(pdata->hwndEdit, FALSE);

	// Move the caret to the start of the current line.
	// This ensures that all the uuencoded text is contained on
	// separate lines.  In particular, it ensures that the "begin "
	// text doesn't have text on the line in front of it, thus
	// screwing up decoding.
	int iLineStart = (int) SendMessage(pdata->hwndEdit, EM_LINEINDEX,
		(WPARAM)-1, 0);	// -1 = this line.
	SendMessage(pdata->hwndEdit, EM_SETSEL,
		GET_EM_SETSEL_MPS(iLineStart, iLineStart));

	// Encode a file.
	int iResult = cuuenc.EncodeLine(szBuffer, sizeof(szBuffer));
	while (iResult == CUUENC::RESULT_OK) {
		lstrcat(szBuffer, "\r\n");
		SendMessage(pdata->hwndEdit, EM_REPLACESEL, 0,
			(LPARAM)(LPCSTR) szBuffer);
		iResult = cuuenc.EncodeLine(szBuffer, sizeof(szBuffer));
	}

	// Turn on redraw again.
	SetWindowRedraw(pdata->hwndEdit, TRUE);

	// Release the captured mouse.
	SetCursor(hcursorSave);
	ReleaseCapture();

	// Figure out what happened.
	if (iResult == CUUENC::RESULT_DONE) {
		// Handle the last line.
		lstrcat(szBuffer, "\r\n");
		SendMessage(pdata->hwndEdit, EM_REPLACESEL, 0,
			(LPARAM)(LPCSTR) szBuffer);
	}
	else if (iResult == CUUENC::RESULT_ERROR) {
    	// Deal with any error that cropped up.
		MessageBox(hwnd, "An error occurred during encoding.",
			"Error", MB_OK | MB_ICONEXCLAMATION);
	}
}

/*********************************************************************
	UTILITY FUNCTIONS
*********************************************************************/

/*
	Function: AddressStringToList
	Description:
		Take a string containing a series of comma-separated
		email addresses, break them into individual strings,
		and add them to an address list.
*/
BOOL AddressStringToList(char * pszAddressString, CAddrList * pAddrList)
{
	char szAddress[80];
	int iAddrStringIndex = 0;

	// Do this until we hit the end of the input string.
	while (pszAddressString[iAddrStringIndex] != '\0') {
		int iAddressIndex = 0;
		char c = pszAddressString[iAddrStringIndex];
		// Copy characters from the input string to the address
		// string until we hit a null or a comma.
		while (c != '\0' && c != ',' &&
			iAddressIndex < sizeof(szAddress)) {
			// Copy all non-whitespace characters to the holding
			// string.
			if (c != ' ' && c != '\t') {
				szAddress[iAddressIndex] = c;
				iAddressIndex++;
			}
			iAddrStringIndex++;
			c = pszAddressString[iAddrStringIndex];
		}
		if (iAddressIndex == sizeof(szAddress)) {
			// If we ran out of room, error.
			return FALSE;
		}
		// Terminate the holding string.
		szAddress[iAddressIndex] = '\0';
		// Move past the comma of the input string.
		if (c != '\0') {
			iAddrStringIndex++;
		}
        // If we got a valid string, add it to the list.
		if (iAddressIndex > 0) {
			pAddrList->Add(szAddress);
		}
	}
	return TRUE;
}

/*
	Function: OutputAddrList
	Description:
		Output a list of addresses in an email address header
		to an output stream.
*/
void OutputAddrList(char * pszHeader, CAddrList * palHeaderList,
	CAddrList * palTotalList, ofstream &ofsMsgStream)
{
	// Create the To: list.
	if (palHeaderList->First() != NULL) {
		char szLine[73];  // 72 + null
		lstrcpy(szLine, pszHeader);
		lstrcat(szLine, palHeaderList->First());
		int iLineLen = lstrlen(szLine);
		palTotalList->Add(palHeaderList->First());
		char * pszNextAddr = palHeaderList->Next();
		if (pszNextAddr == NULL) {
			ofsMsgStream << szLine << '\n';
		}
		else {
			lstrcat(szLine, ",");
			iLineLen++;
			while (pszNextAddr != NULL) {
				while (pszNextAddr != NULL &&
					/* space + comma */
					((iLineLen + lstrlen(pszNextAddr) + 2) <
					 sizeof(szLine))) {
					 // Add a space after the previous comma.
					 lstrcat(szLine, " ");
					 iLineLen++;
					 lstrcat(szLine, pszNextAddr);
					 iLineLen += lstrlen(pszNextAddr);
					 palTotalList->Add(pszNextAddr);
					 pszNextAddr = palHeaderList->Next();
					 if (pszNextAddr != NULL) {
						// If there's another one, add a comma.
						lstrcat(szLine, ",");
						iLineLen++;
					 }
				}
				ofsMsgStream << szLine << '\n';
				lstrcpy(szLine, "   ");  // Add 3 spaces.
				iLineLen = lstrlen(szLine);
			} // end while
		} // end else
	} // end if
}

/*
	Function: OutputMessageText
	Description:
		Outputs a text mail body part to a stream.  Includes
		Content-Type and Content-Transfer-Encoding headers
		describing the text.  Handles ISO 8859-1 text by using
		quoted-printable encoding.
*/
void OutputMessageText(HWND hwndEdit, ofstream& ofsMsgStream)
{
	// Search the message text to see if any of the characters have
	// the high bit set.  If so, then the character set is ISO 8859-1.
	BOOL bISOCharset = FALSE;
	int iNumLines = (int)SendMessage(hwndEdit, EM_GETLINECOUNT, 0, 0);
    int i;
	for (i = 0; i < iNumLines; i++) {
		char szLineBuffer[500];
		// Get the line from the edit control.
		// First indicate the size of the buffer in the first couple
		// of bytes (pretty wierd way to do this, Microsoft...).
		// Leave 1 byte for the terminator which EM_GETLINE won't put
		// in.
		*(WORD *)szLineBuffer = sizeof(szLineBuffer) - 1;
		int iLength = (int) SendMessage(hwndEdit, EM_GETLINE, i,
			(LPARAM) (char FAR *) szLineBuffer);
		// EM_GETLINE doesn't terminate the string, so add a null.
		szLineBuffer[iLength] = '\0';
		for (int j = 0; szLineBuffer[j] != '\0'; j++) {
			if (0x80 & szLineBuffer[j]) {
				bISOCharset = TRUE;
				break;
			}
		}
		if (bISOCharset == TRUE)
			break;
	}

	ofsMsgStream << "Content-Type: text/plain; charset=";
	ofsMsgStream << (bISOCharset ? "iso-8859-1\n" : "us-ascii\n");
	ofsMsgStream << "Content-Transfer-Encoding: ";
	ofsMsgStream << (bISOCharset ? "quoted-printable\n" : "7bit\n");

	// Blank line separating headers and body.
	ofsMsgStream << '\n';

	// Output the lines of the message.
	iNumLines = (int) SendMessage(hwndEdit, EM_GETLINECOUNT, 0, 0);
	for (i = 0; i < iNumLines; i++) {
		char szLineBuffer[500];
		// Get the line from the edit control.
		// First indicate the size of the buffer in the first couple
		// of bytes (pretty wierd way to do this, Microsoft...).
		// Leave 1 byte for the terminator which EM_GETLINE won't put
		// in.
		*(WORD *)szLineBuffer = sizeof(szLineBuffer) - 1;
		int iLength = (int) SendMessage(hwndEdit, EM_GETLINE, i,
			(LPARAM) (char FAR *) szLineBuffer);
		// EM_GETLINE doesn't terminate the string, so add a null.
		szLineBuffer[iLength] = '\0';
		if (bISOCharset) {
			// Check for characters with high-bit set and convert
			// them to quoted printable.
			char szSecondBuffer[500];
			ConvertToQP(szSecondBuffer, sizeof(szSecondBuffer),
				szLineBuffer);
			ofsMsgStream << szSecondBuffer << '\n';
		}
		else {
			// It's ASCII, so simply write it to the file.
			ofsMsgStream << szLineBuffer << '\n';
		}
	}
}

/*
	Function: ConvertToQP
	Description:
		Converts a line of 8-bit characters into quoted-printable
		encoding.
*/
void ConvertToQP(char * pszAfter, int nLen, char * pszBefore)
{
	int iAfter = 0;
	int iBefore = 0;
	int nLineLength = 0;
	// Note that we stop a bit before we overrun the after buffer
	// because we could have one character generate 3 quoted
	// characters, a soft line break could be inserted, and
    // we still need to insert a null terminator.
	while ((iAfter < nLen - 6) && pszBefore[iBefore] != '\0') {
		char c = pszBefore[iBefore];
		// Is it a whitespace character?
		if (IsLWSP(c)) {
			// Quoted printable encoding doesn't allow a space or tab
			// at the end of a line.  See if there are more
			// non-whitespace characters following.  If not, encode
			// this character as "=09" or "=20".
			int iTestIndex = iBefore + 1;
			BOOL bNonLWSPFollows = FALSE;
			while (pszBefore[iTestIndex] != '\0') {
				if (!IsLWSP(pszBefore[iTestIndex])) {
					bNonLWSPFollows = TRUE;
					break;
				}
				iTestIndex++;
			}
			if (bNonLWSPFollows) {
				// Non-whitespace follows, so we're okay.
				pszAfter[iAfter] = c;
				iAfter++;
				nLineLength++;
			}
			else {
				// Just whitespace at the end of the line, so quote
				// this one.
				QuoteCharacter(c, &pszAfter[iAfter]);
				iAfter += 3;
				nLineLength += 3;
			}
		}
		// Is 33 <= c <= 60 or 62 <= c <= 126?
		else if ((33 <= c && c <= 60) || (62 <= c && c <= 126)) {
			// It's a normal character and doesn't have to be
			// quoted.
			pszAfter[iAfter] = c;
			iAfter++;
            nLineLength++;
		}
		else {
			QuoteCharacter(c, &pszAfter[iAfter]);
			iAfter += 3;
			nLineLength += 3;
		}

		// Check the line length to ensure
		if (nLineLength >= 75 && pszBefore[iBefore + 1] != '\0') {
			pszAfter[iAfter++] = '=';
			pszAfter[iAfter++] = '\n';
            nLineLength = 0;
		}
		iBefore++;
	}
	pszAfter[iAfter] = '\0';
}

/*
	Function: QuoteCharacter
	Description:
		Convert a single character into quoted-printable encoding.
*/
void QuoteCharacter(unsigned char c, char * pszBuffer)
{
	const char szHexDigit[] = "0123456789ABCDEF";
	pszBuffer[0] = '=';
	pszBuffer[1] = szHexDigit[(c >> 4) & 0xF];
	pszBuffer[2] = szHexDigit[c & 0xF];
}

/*
	Function: CreateBoundary
	Description:
		Creates a virtually unique message body separation boundary.
		Uses pseudo-random characters as filler to reduce the
		chance that a body part contains the boundary text somewhere.
*/
void CreateBoundary(char * szBuffer, int nLen)
{
	// From MIME RFC definition.
	const char szBCharsNoSpace[] =
		"0123456789abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ'()+_,-./:=?";
	const char szStart[] = "=_ Boundary ";
	// Fill up as much of the buffer as we can with the starting
	// text.
	int i = 0;
	while (i < nLen - 1 && szStart[i] != '\0') {
		szBuffer[i] = szStart[i];
		i++;
	}
	// Then fill the rest of the buffer with random stuff to make
    // it virtually unique.
	while (i < nLen - 1) {
		szBuffer[i] = szBCharsNoSpace[rand() %
			sizeof(szBCharsNoSpace)];
		i++;
	}
    szBuffer[i] = '\0';
}

/*
	Function: OutputAttachment
	Description:
		Output the attachment body part to a stream.  The function
		outputs Content-Type and Content-Transfer-Encoding
		headers and converts the specified file to BASE64 encoding.
*/
void OutputAttachment(char * pszFileName, ofstream& ofsMsgStream)
{
	// Output the header info.
	ofsMsgStream << "Content-Type: application/octet-stream\n";
	ofsMsgStream << "Content-Transfer-Encoding: base64\n";
	ofsMsgStream << '\n';

	// Create an encoder object and have it open the attached file.
	CBase64Enc b64encoder;
	int iResult = b64encoder.OpenFile(pszFileName);
	if (iResult != CBase64Enc::RESULT_OK)
		return;
	// Encode all the data in the file and write it to the output
    // stream one line at a time.
	char szLine[80];
	iResult = b64encoder.EncodeLine(szLine, sizeof(szLine));
	while (iResult == CBase64Enc::RESULT_OK) {
		ofsMsgStream << szLine << '\n';
		iResult = b64encoder.EncodeLine(szLine, sizeof(szLine));
	}
}
