/*
	File: MAILVIEW.CPP
	Description:
		Implements the viewer window class for Mail Client.
	Copyright 1995, David G. Roberts
*/

#include <string.h>
#include <dos.h>
#include <fstream.h>
#include <windows.h>
#include <windowsx.h>
#include "mailcli.rh"
#include "myassert.h"
#include "cwinsock.h"
#include "csmtp.h"
#include "caddrlst.h"
#include "rfc822.h"
#include "cuuendec.h"
#include "cmime.h"
#include "cbase64.h"

#define DECLARE
#include "mailclig.h"

/*
	CONSTANTS
*/
#define EDITCTLID		1

/*
	MACROS
*/
#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))

/*
	TYPES
*/

// VIEWERDATA holds some data associated with the viewer window.
struct VIEWERDATA {
	HWND		hwndEdit;
	CMIMEMsg *	pMessage;

    VIEWERDATA::VIEWERDATA() { pMessage = NULL; };
	VIEWERDATA::~VIEWERDATA() { delete pMessage; };
};

/*
	LOCAL FUNCTIONS
*/
static void OnCommand(HWND hwnd, int id, HWND hwndCtl,
	UINT codeNotify);
static BOOL OnCreate(HWND, CREATESTRUCT FAR*);
static void OnDestroy(HWND);
static void OnMDIActivate(HWND, BOOL, HWND, HWND);
static void OnSize(HWND, UINT, int, int);

/*
	MENU HANDLERS
*/
static void OnMailUudecode(HWND);

/*
	UTILITY FUNCTIONS
*/
void ViewMessagePart(HWND);
void ConvertQPToText(char *, char *, int);
int ConvertHexDigit(char);

/*********************************************************************
	WINDOW AND DIALOG BOX PROCEDURES
*********************************************************************/

/*
	Function: ViewerWndProc
	Description:
		Window proc for the viewer window.
*/
LRESULT CALLBACK ViewerWndProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	switch (msg) {
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
		HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
		HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
		HANDLE_MSG(hwnd, WM_MDIACTIVATE, OnMDIActivate);
		HANDLE_MSG(hwnd, WM_SIZE, OnSize);
	}
	return DefMDIChildProc(hwnd, msg, wParam, lParam);
}

/*********************************************************************
	MESSAGE HANDLING FUNCTIONS
*********************************************************************/

/*
	Function: OnCommand
	Description:
		Handles WM_COMMAND messages.
*/
void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	VIEWERDATA FAR *lpdata = (VIEWERDATA FAR *)GetWindowLong(hwnd, 0);

	if (hwndCtl == NULL) {
		// It's a window message.
		switch (id) {
			case IDM_EDITCOPY:
				SendMessage(lpdata->hwndEdit, WM_COPY, 0, 0);
				return;
			case IDM_MAILFIRSTBODYPART:
				lpdata->pMessage->FirstPart();
				ViewMessagePart(hwnd);
				return;
			case IDM_MAILNEXTBODYPART:
				if (lpdata->pMessage->NextPart() == NULL) {
					MessageBox(hwnd, "There are no more parts.",
                    	"Notice", MB_OK | MB_ICONSTOP);
				}
				else
					ViewMessagePart(hwnd);
				return;
			case IDM_MAILUUDECODE:
				OnMailUudecode(hwnd);
				return;
		}
	}
	else {
		ASSERT(id == EDITCTLID);
		// It's from our control.
		if (codeNotify == EN_ERRSPACE) {
			MessageBox(hwnd, "Viewer out of space.", "Error",
				MB_OK | MB_ICONEXCLAMATION);
			return;
		}
	}
	// Make sure the default child proc sees things we don't
	// process ourselves!
	FORWARD_WM_COMMAND(hwnd, id, hwndCtl, codeNotify,
		DefMDIChildProc);
}

/*
	Function: OnCreate
	Description:
		Creates the editor window in which the messages are
		displayed, parses the message, and displays the first
		body part.
*/
BOOL OnCreate(HWND hwnd, CREATESTRUCT FAR* lpCreateStruct)
{
	// Allocate some private window data for ourselves.
	VIEWERDATA FAR * lpdata = new VIEWERDATA;
	SetWindowLong(hwnd, 0, (LONG) lpdata);

	// Create a multi-line edit control to do our work for us.
	lpdata->hwndEdit = CreateWindow(
		"EDIT",						// class name
		NULL,						// window name
		WS_CHILD | WS_VISIBLE |		// window styles
		WS_HSCROLL | WS_VSCROLL | ES_LEFT |
		ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL | ES_READONLY,
		0,0,0,0,					// location and size
		hwnd,						// parent
		(HMENU) EDITCTLID,			// child id, not really a menu
		g_hInstance,				// program instance
		NULL);						// creation data

	// Set the edit control's font to a fixed pitch.
	HFONT hfont = GetStockFont(SYSTEM_FIXED_FONT);
	SetWindowFont(lpdata->hwndEdit, hfont, FALSE);

	// Create the full file name from the message directory name
    // and the individual file name.
	char FAR * lpszFileName = (LPSTR)(((MDICREATESTRUCT FAR *)
			(lpCreateStruct->lpCreateParams))->lParam);
	char szFullName[MAX_PATH_LENGTH];
	lstrcpy(szFullName, g_szMessageDir);
	lstrcat(szFullName, "\\");
	lstrcat(szFullName, lpszFileName);
	delete[] lpszFileName;

	// Create the MIME message object.
	lpdata->pMessage = new CMIMEMsg;
	if (lpdata->pMessage->Parse(szFullName) != TRUE) {
		// Handle possible parsing error.
		MessageBox(hwnd, "Error parsing message.", "Error",
			MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	// Move the first part of the message and view it.
	lpdata->pMessage->FirstPart();
	ViewMessagePart(hwnd);

	return TRUE;
}

/*
	Function: OnDestroy
	Description:
		Handles a WM_DESTROY message.
*/
void OnDestroy(HWND hwnd)
{
	// Deallocate our private window data before we go away.
	VIEWERDATA FAR *lpdata = (VIEWERDATA FAR *)GetWindowLong(hwnd, 0);
	delete lpdata;
}
/*
	Function: OnMDIActivate
	Description:
		Takes care of putting the viewer menu into the menu
		bar when the viewer window becomes active.
*/
void OnMDIActivate(HWND hwnd, BOOL fActive, HWND hwndActivate,
	HWND hwndDeactivate)
{
	UNUSED_PARAMETER(hwndActivate);
	UNUSED_PARAMETER(hwndDeactivate);

	if (fActive) {
		// Put our menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hViewerMenu,
				g_hViewerWindowMenu));
		VIEWERDATA FAR *lpdata =
			(VIEWERDATA FAR*)GetWindowLong(hwnd, 0);
		// Make sure keystrokes go to the edit control.
		SetFocus(lpdata->hwndEdit);
	}
	else {
		// We're being deactivated, so put the main menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hFrameMenu,
				g_hFrameWindowMenu));
	}
	DrawMenuBar(g_hFrame);
}

/*
	Function: OnSize
	Description:
		Notifies the editor child window whenever the viewer window
		gets resized.
*/
void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
	UNUSED_PARAMETER(state);

	VIEWERDATA FAR *lpdata = (VIEWERDATA FAR *)GetWindowLong(hwnd, 0);
	MoveWindow(lpdata->hwndEdit, 0, 0, cx, cy, TRUE);
	// Must make sure we forward this to the default child proc
	// as well!!
	FORWARD_WM_SIZE(hwnd, state, cx, cy, DefMDIChildProc);
}

/*********************************************************************
	MENU HANDLERS
*********************************************************************/

/*
	Function: OnMailUudecode
	Description:
		Handles a UUDECODE command from the menu bar.
*/
void OnMailUudecode(HWND hwnd)
{
	int iResult;

	// Capture the mouse, 'cause this could take more than an
	// instant.
	SetCapture(g_hFrame);
	HCURSOR hcursorSave = SetCursor(LoadCursor(NULL, IDC_WAIT));

	CUUDEC uudec(hwnd);
	VIEWERDATA FAR *lpdata = (VIEWERDATA FAR *)GetWindowLong(hwnd, 0);

    // Figure out how many lines there are in this viewer.
	int iCount = (int) SendMessage(lpdata->hwndEdit, EM_GETLINECOUNT,
		0, 0);
	for (int i = 0; i < iCount; i++) {
		char szLineBuffer[1000];
		// Get the line from the edit control.
		// First indicate the size of the buffer in the first couple
		// of bytes (pretty weird way to do this, Microsoft...).
		// Leave 1 byte for the terminator which EM_GETLINE won't put
        // in.
		*(WORD *)szLineBuffer = sizeof(szLineBuffer) - 1;
		int iLength = (int) SendMessage(lpdata->hwndEdit, EM_GETLINE,
			i, (LPARAM) (char FAR *) szLineBuffer);
		// EM_GETLINE doesn't terminate the string, so add a null.
		szLineBuffer[iLength] = '\0';
		iResult = uudec.DecodeLine(szLineBuffer);
		if (iResult != CUUDEC::RESULT_OK)
        	break;
	}

	// Release the mouse.
	SetCursor(hcursorSave);
	ReleaseCapture();

	// If something went wrong, signal it.
	if (iResult == CUUDEC::RESULT_ERROR) {
		MessageBox(hwnd, "An error occurred during decoding.",
			"Error", MB_OK | MB_ICONEXCLAMATION);
	}
}

/*********************************************************************
	UTILITY FUNCTIONS
*********************************************************************/

/*
	Function: ViewMessagePart
	Description:
		Given the current message body part, the function displays
		it in the viewer window if it's a text message.  If it's
		a different message type, the function offers to save it
		in a file.
*/
void ViewMessagePart(HWND hwnd)
{
	int iResult;
	BOOL bResult;
    int iLength;
	char szRawBuffer[1000];
	char szDecodedBuffer[1000];
	CBase64Dec * pBase64Dec = new CBase64Dec;
	static char *szFilter[] = { "All Files (*.*)", "*.*", "", "" };
	char szFullName[MAX_PATH_LENGTH];
	char szFileName[MAX_FILENAME_LENGTH];

	VIEWERDATA FAR *lpdata = (VIEWERDATA FAR *)GetWindowLong(hwnd, 0);

	CMIMEBodyPart * pCurrentPart = lpdata->pMessage->CurrentPart();

	// Delete all the data in the viewer currently.
	SendMessage(lpdata->hwndEdit, EM_SETSEL,
		GET_EM_SETSEL_MPS(0, -1));
	SendMessage(lpdata->hwndEdit, EM_REPLACESEL, 0, (LPARAM) "");

	// Open up the body part so we can read it.
	if (pCurrentPart->Open() != TRUE) {
		MessageBox(hwnd, "Could not open body part.", "Error",
			MB_OK | MB_ICONEXCLAMATION);
		return;
	}

	// Figure out what the current body part type is.
	// Whatever it is, it shouldn't be a multipart node.  The
	// FirstPart and NextPart functions of CMIMEMsg should skip
	// over these and only give us the constituent parts of a
    // multipart.
	ASSERT(lstrcmpi(pCurrentPart->Type(), "multipart") != 0);

	if (lstrcmpi(pCurrentPart->Type(), "text") == 0) {
		// Figure out what character set it is using.
		CMIMEParamTuple * pTuple = pCurrentPart->FirstParam();
		while (pTuple != NULL &&
			lstrcmpi(pTuple->Param(), "charset") != 0) {
			pTuple = pCurrentPart->NextParam();
		}
		// See if we can handle it.
		if (pTuple != NULL &&
			lstrcmpi(pTuple->Value(), "us-ascii") != 0 &&
			lstrcmpi(pTuple->Value(), "iso-8859-1") != 0) {
			MessageBox(hwnd, "This message body part uses an "
				"unknown or unsupported character set.  It may not "
				"display correctly.", "Warning",
				MB_OK | MB_ICONEXCLAMATION);
		}
		// If it's not text/plain, throw up a disclaimer.
		if (lstrcmpi(pCurrentPart->SubType(), "plain") != 0) {
			MessageBox(hwnd, "This message part contains an unknown "
				"subtype of type text.  It may not display "
				"correctly.",
                "Warning", MB_OK | MB_ICONASTERISK);
		}

		// Put the text into the viewer.
		BOOL bFoundSeparator = FALSE;
		int iResult = pCurrentPart->FirstLine(szRawBuffer,
			sizeof(szRawBuffer));
		while (iResult >= 0) {
			// Decode buffer.
			if (bFoundSeparator) {
				if (lstrcmpi(pCurrentPart->Encoding(), "base64")
					== 0) {
					iLength = pBase64Dec->DecodeLine(szRawBuffer,
						lstrlen(szRawBuffer),
						(uchar *)szDecodedBuffer,
						sizeof(szDecodedBuffer));
					szDecodedBuffer[iLength] = '\0';
				}
				else if (lstrcmpi(pCurrentPart->Encoding(),
					"quoted-printable") == 0) {
					ConvertQPToText(szRawBuffer, szDecodedBuffer,
						sizeof(szDecodedBuffer));
				}
				else {
					lstrcpy(szDecodedBuffer, szRawBuffer);
				}
			}
			else {
				if (lstrcmp(szRawBuffer, "\r\n") == 0)
					bFoundSeparator = TRUE;
				lstrcpy(szDecodedBuffer, szRawBuffer);
			}
			// Stuff the decoded text into the edit control
			SendMessage(lpdata->hwndEdit, EM_REPLACESEL, 0,
				(LPARAM) szDecodedBuffer);
			// Get the next line of raw text.
			iResult = pCurrentPart->NextLine(szRawBuffer,
				sizeof(szRawBuffer));
		}
	}
	else {
		// Treat everything else like application/octet-stream.

		// Put something in the edit box.
		SendMessage(lpdata->hwndEdit, EM_REPLACESEL, 0,
				(LPARAM) "*** NON-TEXT DATA ***\r\n");

		// Query the user if we should save it.
		iResult = MessageBox(hwnd,
			"This body part contains non-text data.  "
			"Would you like to save it to a file?", "Save?",
			MB_YESNO | MB_ICONQUESTION);
		if (iResult == IDNO)
			goto exit;

		// Open up a common file dialog.
		szFullName[0] = szFileName[0] = '\0';
		OPENFILENAME ofn;
		memset(&ofn, 0, sizeof(OPENFILENAME));

		ofn.lStructSize		= sizeof(OPENFILENAME);
		ofn.hwndOwner		= hwnd;
		ofn.lpstrFilter		= szFilter[0];
		ofn.lpstrFile		= szFullName;
		ofn.nMaxFile		= sizeof(szFullName);
		ofn.lpstrFileTitle	= szFileName;
		ofn.nMaxFileTitle	= sizeof(szFileName);
		ofn.lpstrTitle		= "Save Non-Text Attachment";
		ofn.Flags			= OFN_HIDEREADONLY |
								OFN_NOREADONLYRETURN |
								OFN_OVERWRITEPROMPT;

		bResult = GetSaveFileName(&ofn);

		if (bResult == 0)
			goto exit;

		// Open up the file.
		OFSTRUCT ofs;
		HFILE hfile = OpenFile(szFullName, &ofs,
			OF_WRITE | OF_CREATE);
		if (hfile == HFILE_ERROR) {
			MessageBox(hwnd, "Could not open the output file.",
            	"Error", MB_OK | MB_ICONEXCLAMATION);
			goto exit;
		}

		// Get each line, decode it, and write it to the file.
		BOOL bFoundSeparator = FALSE;
		int iResult = pCurrentPart->FirstLine(szRawBuffer,
			sizeof(szRawBuffer));
		while (iResult >= 0) {
			// Decode buffer.
			if (bFoundSeparator) {
				if (lstrcmpi(pCurrentPart->Encoding(), "base64")
					== 0) {
					iLength = pBase64Dec->DecodeLine(szRawBuffer,
						lstrlen(szRawBuffer),
						(uchar *)szDecodedBuffer,
						sizeof(szDecodedBuffer));
				}
				else if (lstrcmpi(pCurrentPart->Encoding(),
					"quoted-printable") == 0) {
					ConvertQPToText(szRawBuffer, szDecodedBuffer,
						sizeof(szDecodedBuffer));
					iLength = lstrlen(szDecodedBuffer);
				}
				else {
					lstrcpy(szDecodedBuffer, szRawBuffer);
				}
				// Write it out.
				_lwrite(hfile, (LPCSTR) szDecodedBuffer, iLength);
			}
			else {
				if (lstrcmp(szRawBuffer, "\r\n") == 0)
					bFoundSeparator = TRUE;
			}
			// Get the next line of raw text.
			iResult = pCurrentPart->NextLine(szRawBuffer,
				sizeof(szRawBuffer));
		}
		_lclose(hfile);

	}

exit:
	// Close up the body part now that it's in the viewer.
	pCurrentPart->Close();

	// Put the caret at the beginning of the control and scroll it
	// into view.
	SendMessage(lpdata->hwndEdit, EM_SETSEL,
		GET_EM_SETSEL_MPS(0, 0));
#ifdef WIN32
	SendMessage(lpdata->hwndEdit, EM_SCROLLCARET, 0, 0);
#endif

	delete pBase64Dec;
}

/*
	Function: ConvertQPToText
	Description:
		Decodes quoted-printable encoded data.
*/
void ConvertQPToText(char * pszInput, char * pszOutput, int nLen)
{
	int iInputIndex = 0;
	int iOutputIndex = 0;

	while (iOutputIndex < (nLen - 1) &&
		pszInput[iInputIndex] != '\0') {
		char c = pszInput[iInputIndex++];
		if (c != '=') {
			pszOutput[iOutputIndex++] = c;
			continue;
		}
		// We found a '=' character.  See if we have a soft
		// line break or a hex-encoded character.
		c = pszInput[iInputIndex++];
		if (c == CR) {
			// Soft break.
			break;
		}
		else {
			pszOutput[iOutputIndex++] =
				(char) (16 * ConvertHexDigit(c) +
				ConvertHexDigit(pszInput[iInputIndex++]));
		}
	}
    pszOutput[iOutputIndex] = '\0';
}

/*
	Function: ConvertHexDigit
	Description:
		Given an ASCII character, the function returns its
		hex value, if it has one, or -1 if the character is not
		a hex digit.
*/
int ConvertHexDigit(char c)
{
	if ('0' <= c && c <= '9')
		return c - '0';
	if ('A' <= c && c <= 'F')
		return c - 'A' + 10;
	if ('a' <= c && c <= 'f')
		return c - 'a' + 10;
	return -1;
}
