/*
	File: MAILCLI.CPP
	Description:
		A simple mail client.
	Copyright 1995, David G. Roberts
*/

#ifdef __BORLANDC__
#include <dir.h>
#else
#include <direct.h>
#define mkdir _mkdir
#define getcwd _getcwd
#endif
#include <stdlib.h>
#include <windows.h>
#include <windowsx.h>
#include "ctl3d.h"
#include "winsock.h"
#include "cwinsock.h"
#include "mailcli.rh"
#include "myassert.h"
#include "tracer.h"

/*
	CONSTANTS
*/
#define DEFAULT_SMTP_PORT	25

#define FRAME_WINDOW_MENU_POS	2
#define EDITOR_WINDOW_MENU_POS	3
#define MAILBOX_WINDOW_MENU_POS	2
#define VIEWER_WINDOW_MENU_POS	3

#define MAILBOX_NAME		"mailbox.txt"
#define MESSAGE_DIR			"messages"

/*
	TYPES
*/
typedef int INT;
typedef BOOL FAR * LPBOOL;

/*
	GLOBALS
*/
#include "mailclig.h"

/*
	FUNCTION PROTOTYPES
*/
void		CleanupInstance(HINSTANCE);
BOOL CALLBACK ConfigureDlgProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK FrameWndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK EditorWndProc(HWND, UINT, WPARAM, LPARAM);
BOOL		InitWinsock(HINSTANCE, HINSTANCE);
BOOL		InitApp(HINSTANCE);
BOOL		InitInstance(HINSTANCE);
void		InitError(void);
LRESULT CALLBACK MailboxWndProc(HWND, UINT, WPARAM, LPARAM);
INT WINAPI	WinMain(HANDLE, HANDLE, LPSTR, INT);
LRESULT CALLBACK ViewerWndProc(HWND, UINT, WPARAM, LPARAM);

/*
	MESSAGE HANDLER PROTOTYPES
*/
static void	OnClose(HWND hwnd);
static void	OnCommand(HWND, int, HWND, UINT);
BOOL OnCreate(HWND, CREATESTRUCT FAR*);

/*
	WM_COMMAND HANDLER PROTOTYPES
*/
static void	OnFileNewMessage(HWND);
static void	OnFileClose(HWND);
static void	OnFileExit(HWND);
static void OnFileOpenMailbox(HWND);
static void	OnMailConfigure(HWND);
static void OnWindowCloseAll(HWND);
static BOOL CALLBACK CloseEnumProc(HWND, LPARAM);
static void	OnHelpAbout(HWND);

/*********************************************************************
	WINMAIN
*********************************************************************/

/*
	Function: WinMain
	Description:
		Good ol' WinMain.
*/
INT WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	LPSTR lpszCmdParam, INT nCmdShow)
{
	UNUSED_PARAMETER(lpszCmdParam);
	Tracer::Instance()->SetFlags(Tracer::TR_ALL);

	// Save off our instance handle for later.
	g_hInstance = hInstance;

	// Use the current directory to fix some paths.
	_fullpath(g_szMessageDir, MESSAGE_DIR, sizeof(g_szMessageDir));
	_fullpath(g_szMailboxFile, MAILBOX_NAME, sizeof(g_szMailboxFile));

	// Make sure a message directory exits.
	// If not, create it.  An error will result if the
	// directory already exists.  Ignore it.
	mkdir(g_szMessageDir);

	// If no other instance is running, initialize the application
	// window classes.
	if (!hPrevInstance) {
		if (!InitApp(hInstance)) {
			InitError();
			return 0;
		}
	}

	// Load all our menus.
	g_hFrameMenu	= LoadMenu(hInstance, "FRAMEMENU");
	g_hMailboxMenu	= LoadMenu(hInstance, "MAILBOXMENU");
	g_hViewerMenu	= LoadMenu(hInstance, "VIEWERMENU");
	g_hEditorMenu	= LoadMenu(hInstance, "EDITORMENU");

	// Get the handle of the Window submenu for each MDI menu.
	g_hFrameWindowMenu = GetSubMenu(g_hFrameMenu,
		FRAME_WINDOW_MENU_POS);
	g_hEditorWindowMenu = GetSubMenu(g_hEditorMenu,
		EDITOR_WINDOW_MENU_POS);
	g_hMailboxWindowMenu = GetSubMenu(g_hMailboxMenu,
		MAILBOX_WINDOW_MENU_POS);
	g_hViewerWindowMenu = GetSubMenu(g_hViewerMenu,
		VIEWER_WINDOW_MENU_POS);

	// Load the accellerators.
    g_hAccel = LoadAccelerators(hInstance, g_szAppName);

	// Initialize all instance specific stuff.  In particular, create
    // the frame window.
	if (!InitInstance(hInstance)) {
		InitError();
		return 0;
	}

	// Since initializing Winsock can take some time (because we
	// get the default Finger port using blocking calls), put
	// up an hourglass cursor so the user knows we're working on
	// something and not just out to lunch.
	SetCapture(g_hFrame);
	SetCursor(LoadCursor(NULL, IDC_WAIT));

	// Make sure Winsock initialized and get the event dispatch
	// mechanism up and running.
	if (!InitWinsock(hInstance, hPrevInstance)) {
		ReleaseCapture();
		return 0;
	}

	ReleaseCapture();

	// Show the frame window and cause it to be repainted.
	ShowWindow(g_hFrame, nCmdShow);
	UpdateWindow(g_hFrame);

	// Do the message loop.
	MSG msg;
	while (GetMessage((LPMSG)&msg, NULL, 0, 0)) {
		if (!TranslateMDISysAccel(g_hClient, &msg) &&
			!TranslateAccelerator(g_hFrame, g_hAccel, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	DestroyMenu(g_hFrameMenu);
	DestroyMenu(g_hMailboxMenu);
	DestroyMenu(g_hViewerMenu);
	DestroyMenu(g_hEditorMenu);

	// Clean up before we exit.
	CleanupInstance(hInstance);

	return msg.wParam;
}

/*
	Function: InitError
	Description:
		Handles putting up a message box for the application if it
		cannot be initialized.
*/
void InitError(void)
{
	char szMessage[80];

	wsprintf(szMessage, "ERROR: Cannot initialize %s.",
		(LPSTR)g_szFullAppName);
	// Use NULL for hwnd because application probably doesn't have
    // its window created yet.
	MessageBox(NULL, szMessage, g_szAppName, MB_ICONSTOP | MB_OK);
}

/*
	Function: InitApp
	Description:
		Registers the application's window classes and performs
		initialization that must be done for all application
		instances.  The function returns TRUE if successful and no
		errors occur, and FALSE otherwise.
*/
BOOL InitApp(HINSTANCE hInstance)
{
	WNDCLASS wc;

	// Register the frame window class.

	wc.style			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= FrameWndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= hInstance;
	wc.hIcon			= LoadIcon(hInstance, "APPICON");
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground	= (HBRUSH) (COLOR_APPWORKSPACE + 1);
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= g_szFrameClass;

	if (!RegisterClass(&wc))
		return FALSE;

	// Register the mailbox child window class.
	wc.style			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= MailboxWndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= sizeof(void FAR *);
	wc.hInstance		= hInstance;
	wc.hIcon			= LoadIcon(hInstance, "MAILBOXICON");
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground	= (HBRUSH) (COLOR_WINDOW + 1);
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= g_szMailboxClass;

	if (!RegisterClass(&wc))
		return FALSE;

	// Register the message viewer child window class.
	wc.style			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= ViewerWndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= sizeof(void FAR *);
	wc.hInstance		= hInstance;
	wc.hIcon			= LoadIcon(hInstance, "VIEWERICON");
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground	= (HBRUSH) (COLOR_WINDOW + 1);
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= g_szViewerClass;

	if (!RegisterClass(&wc))
		return FALSE;

	// Register the message editor child window class.
	wc.style			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= EditorWndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= sizeof(void FAR *);
	wc.hInstance		= hInstance;
	wc.hIcon			= LoadIcon(hInstance, "EDITORICON");
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground	= (HBRUSH) (COLOR_WINDOW + 1);
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= g_szEditorClass;

	if (!RegisterClass(&wc))
		return FALSE;

	return TRUE;
}

/*
	Function: InitInstance
	Description:
		Performs instance specific initialization, like the
		creation of the frame window, etc.  The function returns
		TRUE if successful and no errors occur, and FALSE otherwise.
*/
BOOL InitInstance(HINSTANCE hInstance)
{
	g_hFrame = CreateWindow(
		g_szFrameClass,					// window class name
		g_szFullAppName,				// window title text
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN, // window style
		CW_USEDEFAULT, CW_USEDEFAULT,	// position = default
		CW_USEDEFAULT, CW_USEDEFAULT,	// size = default
		NULL,							// no parent window
		g_hFrameMenu,					// menu = window class default
		hInstance,						// window owner
		NULL);							// no additional creation data

	if (!g_hFrame)
		return FALSE;

	// Initialize Microsoft 3D controls (CTL3D)
	Ctl3dRegister(hInstance);
	Ctl3dAutoSubclass(hInstance);

	return TRUE;
}

/*
	Function: InitWinsock
	Description:
		Make sure Winsock is up and running and that the Winsock event
		dispatch window is created.  Further, perform any actions that
		have to take place before the application window is created.
*/
BOOL InitWinsock(HINSTANCE hInstance, HINSTANCE hPrevInstance)
{
	// Make sure that we didn't have any problems
	// starting up Winsock.
	if (g_wsWinsock.LastError() != WSANOERROR) {
		ReportWSAError("Error: Can't startup Winsock.",
			g_wsWinsock.LastError(), NULL, g_hInstance);
		return FALSE;
	}

	// Create the Winsock message dispatch window.
	if (!WSCCreateEventWindow(hInstance, hPrevInstance)) {
		MessageBox(g_hFrame,
			"Error: Could not create the Winsock message window.",
			"Winsock Error", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	return TRUE;
}

/*
	Function: CleanupInstance
	Description:
		Do any instance specific clean-up before the application
		quits.
*/
void CleanupInstance(HINSTANCE hInstance)
{
	// Clean up Microsoft 3D controls (CTL3D)
	Ctl3dUnregister(hInstance);
}

/*********************************************************************
	WINDOW AND DIALOG PROCS
*********************************************************************/

/*
	Function: FrameWndProc
	Description:
		Window procedure for the application frame window.
*/
LRESULT CALLBACK FrameWndProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	switch (msg) {
		HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
		HANDLE_MSG(hwnd, WM_CLOSE, OnClose);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
		case WM_DESTROY:
			// Delete all the config info.
			delete[] g_pszSMTPServer;
			delete[] g_pszPOPServer;
			delete[] g_pszMailbox;
			delete[] g_pszPassword;
			delete[] g_pszUserName;
			delete[] g_pszUserAddress;
            // Post the quit message.
			PostQuitMessage(0);
			return 0;
		default:
			return DefFrameProc(hwnd, g_hClient, msg, wParam, lParam);
	}
}

/*
	Function: ConfigureDlgProc
	Description:
		Dialog proc for the configuration dialog.
*/
BOOL CALLBACK ConfigureDlgProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	int		id;
	BOOL	bTranslated;
	int		nLen;			

	UNUSED_PARAMETER(lParam);

	switch (msg) {
		case WM_INITDIALOG:
			// Initialize dialog edit boxes.
			SetDlgItemText(hwnd, IDD_SMTPSERVER, g_pszSMTPServer);
			SetDlgItemInt(hwnd, IDD_SMTPPORT, g_ushSMTPPort, FALSE);
			SetDlgItemText(hwnd, IDD_POPSERVER, g_pszPOPServer);
			SetDlgItemInt(hwnd, IDD_POPPORT, g_ushPOPPort, FALSE);
			SetDlgItemText(hwnd, IDD_MAILBOX, g_pszMailbox);
			SetDlgItemText(hwnd, IDD_PASSWORD, g_pszPassword);
			SetDlgItemText(hwnd, IDD_NAME, g_pszUserName);
			SetDlgItemText(hwnd, IDD_ADDRESS, g_pszUserAddress);
			return TRUE;
		case WM_COMMAND:
			// Explode wParam and lParam into constituent parts.
			id = GET_WM_COMMAND_ID(wParam, lParam);
			switch (id) {
				case IDOK:
					// Verify all the fields.
					if (!(VerifyDlgText(hwnd, IDD_SMTPSERVER) &&
						VerifyDlgInt(hwnd, IDD_SMTPPORT, FALSE) &&
						VerifyDlgText(hwnd, IDD_POPSERVER) &&
						VerifyDlgInt(hwnd, IDD_POPPORT, FALSE) &&
						VerifyDlgText(hwnd, IDD_MAILBOX) &&
						VerifyDlgText(hwnd, IDD_PASSWORD) &&
						VerifyDlgText(hwnd, IDD_NAME) &&
						VerifyDlgText(hwnd, IDD_ADDRESS))) {
							return TRUE;
					}
					// Now that we've verified everything, retrieve
					// it all "for real."
					// First the SMTP server...
					delete[] g_pszSMTPServer;
					nLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_SMTPSERVER));
					g_pszSMTPServer = new char[nLen + 1];
					GetDlgItemText(hwnd, IDD_SMTPSERVER,
						g_pszSMTPServer, nLen + 1);
					// ...Then the SMTP port...
					g_ushSMTPPort = (u_short) GetDlgItemInt(hwnd,
						IDD_SMTPPORT, (LPBOOL) &bTranslated, FALSE);
					// ...Then the POP server...
					delete[] g_pszPOPServer;
					nLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_POPSERVER));
					g_pszPOPServer = new char[nLen + 1];
					GetDlgItemText(hwnd, IDD_POPSERVER,
						g_pszPOPServer, nLen + 1);
					// ...Then the POP port...
					g_ushPOPPort = (u_short) GetDlgItemInt(hwnd,
						IDD_POPPORT, (LPBOOL) &bTranslated, FALSE);
					// ...Then the mailbox name...
					delete[] g_pszMailbox;
					nLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_MAILBOX));
					g_pszMailbox = new char[nLen + 1];
					GetDlgItemText(hwnd, IDD_MAILBOX, g_pszMailbox,
						nLen + 1);
					// ...Then the password string...
					delete[] g_pszPassword;
					nLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_PASSWORD));
					g_pszPassword = new char[nLen + 1];
					GetDlgItemText(hwnd, IDD_PASSWORD, g_pszPassword,
						nLen + 1);
					// ...Then the user name string...
					delete[] g_pszUserName;
					nLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_NAME));
					g_pszUserName = new char[nLen + 1];
					GetDlgItemText(hwnd, IDD_NAME,
						g_pszUserName, nLen + 1);
					// ...Then the user address string.
					delete[] g_pszUserAddress;
					nLen = Edit_GetTextLength(GetDlgItem(hwnd,
						IDD_ADDRESS));
					g_pszUserAddress = new char[nLen + 1];
					GetDlgItemText(hwnd, IDD_ADDRESS,
						g_pszUserAddress, nLen + 1);
					// Now we're done.
					EndDialog(hwnd, TRUE);
					return TRUE;
				case IDCANCEL:
					EndDialog(hwnd, FALSE);
					return TRUE;
				case IDD_DEFAULTSMTPPORT:
					SetDlgItemInt(hwnd, IDD_SMTPPORT,
						g_ushSMTPDefaultPort, FALSE);
					return TRUE;
				case IDD_DEFAULTPOPPORT:
					SetDlgItemInt(hwnd, IDD_POPPORT,
						g_ushPOPDefaultPort, FALSE);
					return TRUE;
			} // switch (id)
	} // switch (msg)
	return FALSE;
}

/*********************************************************************
	WINDOW MESSAGE HANDLERS
*********************************************************************/

/*
	Function: OnClose
	Description:
		Handles a WM_CLOSE message to the frame window.
*/
void OnClose(HWND hwnd)
{
	// Send ourselves a Close All menu command message to close up
	// the children.
	SendMessage(hwnd, WM_COMMAND, IDM_WINDOWCLOSEALL, 0);
	if (GetWindow(g_hClient, GW_CHILD) == NULL) {
		// If the client window has no children, then all the children
		// are closed and we can continue.  Call the default window
		// proc to have it post a destroy message.
		FORWARD_WM_CLOSE(hwnd, ForwardDefFrameProc);
	}
	// One of the client's children didn't close, so simply return.
    // The user can't exit.
}

/*
	Function: OnCommand
	Description:
		Handle WM_COMMAND messages.
*/
void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	// Dispatch command
	switch (id) {
		case IDM_FILENEWMESSAGE:
			OnFileNewMessage(hwnd);
			break;
		case IDM_FILECLOSE:
			OnFileClose(hwnd);
			break;
		case IDM_FILEEXIT:
			OnFileExit(hwnd);
			break;
		case IDM_FILEOPENMAILBOX:
			OnFileOpenMailbox(hwnd);
			break;
		case IDM_MAILCONFIGURE:
			OnMailConfigure(hwnd);
			break;
		case IDM_WINDOWCASCADE:
			SendMessage(g_hClient, WM_MDICASCADE, 0, 0);
			break;
		case IDM_WINDOWARRANGE:
			SendMessage(g_hClient, WM_MDIICONARRANGE, 0, 0);
			break;
		case IDM_WINDOWTILE:
			SendMessage(g_hClient, WM_MDITILE, 0, 0);
            break;
		case IDM_WINDOWCLOSEALL:
			OnWindowCloseAll(hwnd);
			break;
		case IDM_HELPABOUT:
			OnHelpAbout(hwnd);
			break;
		default:
			// ...Or send it to the active child window and see if it
			// processes it.
			HWND hwndChild = GetMDIActiveChild(g_hClient);
			if (IsWindow(hwndChild)) {
				FORWARD_WM_COMMAND(hwndChild, id, hwndCtl, codeNotify,
					SendMessage);
			}
            break;
	}
    // Always forward it to DefFrameProc after we're done.
	FORWARD_WM_COMMAND(hwnd, id, hwndCtl, codeNotify,
		ForwardDefFrameProc);
}

/*
	Function: OnCreate
	Description:
		Handle creation of the main frame window.  Creates the MDI
		client window to manage MDI children and initializes the mail
		configuration information.
*/
BOOL OnCreate(HWND hwnd, CREATESTRUCT FAR* lpCreateStruct)
{
	UNUSED_PARAMETER(lpCreateStruct);

	CLIENTCREATESTRUCT ClientCreate;

	// Create the MDI client window.
	ClientCreate.hWindowMenu = g_hFrameWindowMenu;
	ClientCreate.idFirstChild = IDM_FIRSTCHILD;
	g_hClient = CreateWindow(
		"MDICLIENT",			// window class name
		NULL,					// no title for the client window
		WS_CHILD | WS_CLIPCHILDREN | WS_VISIBLE, // window style
		0, 0,					// position at 0,0
		0, 0,					// no size
		hwnd,					// parent is the frame window
		(HMENU) 1,				// Not really a menu. A child
								// window number when doing MDI.
		g_hInstance,			// owning instance
		(LPSTR)&ClientCreate);	// additional creation data

	// Set up the default ports for SMTP and POP.
	g_ushSMTPDefaultPort = 25;
	g_ushPOPDefaultPort = 110;

	// Initialize the configuration information.
	const int STR_BASE = 10;
	const int STR_GROW_QUANTUM = 10;
	int iStrLen;
	int iReadLen;

	// Assume we're going to do everything right.
    g_bConfigured = TRUE;

    // Get the SMTP server name.
	iStrLen = STR_BASE;
	g_pszSMTPServer = new char[1];
	do {
		delete[] g_pszSMTPServer;
		iStrLen += STR_GROW_QUANTUM;
		g_pszSMTPServer = new char[iStrLen];
		iReadLen = GetPrivateProfileString("SMTP", "Server", "",
			g_pszSMTPServer, iStrLen, g_szIniFile);
	} while (iReadLen >= iStrLen - 1);
	if (lstrcmp(g_pszSMTPServer, "") == 0) {
		// If a null string got in the file or there was no entry,
        // then we still aren't configured.
		g_bConfigured = FALSE;
	}

	// Get the SMTP port number.
	g_ushSMTPPort = (u_short) GetPrivateProfileInt("SMTP", "Port", 0,
		g_szIniFile);
	if (g_ushSMTPPort == 0) {
		g_ushSMTPPort = g_ushSMTPDefaultPort;
		g_bConfigured = FALSE;
	}

	// Get the POP server name.
	iStrLen = STR_BASE;
	g_pszPOPServer = new char[1];
	do {
		delete[] g_pszPOPServer;
		iStrLen += STR_GROW_QUANTUM;
		g_pszPOPServer = new char[iStrLen];
		iReadLen = GetPrivateProfileString("POP", "Server", "",
			g_pszPOPServer, iStrLen, g_szIniFile);
	} while (iReadLen >= iStrLen - 1);
	if (lstrcmp(g_pszPOPServer, "") == 0) {
		// If a null string got in the file or there was no entry,
		// then we still aren't configured.
		g_bConfigured = FALSE;
	}

	// Get the POP port number.
	g_ushPOPPort = (u_short) GetPrivateProfileInt("POP", "Port", 0,
		g_szIniFile);
	if (g_ushPOPPort == 0) {
    	g_ushPOPPort = g_ushPOPDefaultPort;
		g_bConfigured = FALSE;
	}

	// Get the POP mailbox name.
	iStrLen = STR_BASE;
	g_pszMailbox = new char[1];
	do {
		delete[] g_pszMailbox;
		iStrLen += STR_GROW_QUANTUM;
		g_pszMailbox = new char[iStrLen];
		iReadLen = GetPrivateProfileString("POP", "Mailbox", "",
			g_pszMailbox, iStrLen, g_szIniFile);
	} while (iReadLen >= iStrLen - 1);
	if (lstrcmp(g_pszMailbox, "") == 0) {
		// If a null string got in the file or there was no entry,
		// then we still aren't configured.
		g_bConfigured = FALSE;
	}

	// Get the POP mailbox password.
	iStrLen = STR_BASE;
	g_pszPassword = new char[1];
	do {
		delete[] g_pszPassword;
		iStrLen += STR_GROW_QUANTUM;
		g_pszPassword = new char[iStrLen];
		iReadLen = GetPrivateProfileString("POP", "Password", "",
			g_pszPassword, iStrLen, g_szIniFile);
	} while (iReadLen >= iStrLen - 1);
	if (lstrcmp(g_pszPassword, "") == 0) {
		// If a null string got in the file or there was no entry,
		// then we still aren't configured.
		g_bConfigured = FALSE;
	}

	// Get the user name.
	iStrLen = STR_BASE;
	g_pszUserName = new char[1];
	do {
		delete[] g_pszUserName;
		iStrLen += STR_GROW_QUANTUM;
		g_pszUserName = new char[iStrLen];
		iReadLen = GetPrivateProfileString("UserInfo", "Name", "",
			g_pszUserName, iStrLen, g_szIniFile);
	} while (iReadLen >= iStrLen - 1);
	if (lstrcmp(g_pszUserName, "") == 0) {
		// If a null string got in the file or there was no entry,
		// then we still aren't configured.
		g_bConfigured = FALSE;
	}

	// Get the user address.
	iStrLen = STR_BASE;
	g_pszUserAddress = new char[1];
	do {
		delete[] g_pszUserAddress;
		iStrLen += STR_GROW_QUANTUM;
		g_pszUserAddress = new char[iStrLen];
		iReadLen = GetPrivateProfileString("UserInfo", "Address", "",
			g_pszUserAddress, iStrLen, g_szIniFile);
	} while (iReadLen >= iStrLen - 1);
	if (lstrcmp(g_pszUserAddress, "") == 0) {
		// If a null string got in the file or there was no entry,
		// then we still aren't configured.
		g_bConfigured = FALSE;
	}

    return TRUE;
}

/*********************************************************************
	WM_COMMAND HANDLERS
*********************************************************************/

/*
	Function: OnFileNewMessage
	Description:
		Create a new message editor window in which to compose
		a message.
*/
void OnFileNewMessage(HWND hwnd)
{
	MDICREATESTRUCT mdicreate;
	static int iNewIndex = 1;
	char szTitle[20];

	UNUSED_PARAMETER(hwnd);

	// Create the new window title.
	wsprintf(szTitle, "New Message %d", iNewIndex);
	iNewIndex++;

	mdicreate.szClass	= g_szEditorClass;
	mdicreate.szTitle	= szTitle;
	mdicreate.hOwner	= g_hInstance;
	mdicreate.x			= CW_USEDEFAULT;
	mdicreate.y			= CW_USEDEFAULT;
	mdicreate.cx		= CW_USEDEFAULT;
	mdicreate.cy		= CW_USEDEFAULT;
	mdicreate.style		= 0;
	mdicreate.lParam	= NULL;

	// Send a message to the client window to create the new child.
	SendMessage(g_hClient, WM_MDICREATE, 0,
		(LPARAM) (LPMDICREATESTRUCT) &mdicreate);
}

/*
	Function: OnFileClose
	Description:
		Close the currently active MDI child window.
*/
void OnFileClose(HWND hwnd)
{
	UNUSED_PARAMETER(hwnd);

	HWND hChild = GetMDIActiveChild(g_hClient);

	if (hChild == NULL)
    	return;

	if (SendMessage(hChild, WM_QUERYENDSESSION, 0, 0)) {
		SendMessage(g_hClient, WM_MDIDESTROY, (WPARAM) hChild, 0);
	}
}

/*
	Function: OnFileExit
	Description:
		Exits the application in response to a IDM_FILEEXIT message.
*/
void OnFileExit(HWND hwnd)
{
	SendMessage(hwnd, WM_CLOSE, 0, 0L);
}

/*
	Function: OnFileOpenMailbox
	Description:
		Create a new mailbox window.
*/
void OnFileOpenMailbox(HWND hwnd)
{
	MDICREATESTRUCT mdicreate;

	UNUSED_PARAMETER(hwnd);

	mdicreate.szClass	= g_szMailboxClass;
	mdicreate.szTitle	= "Mailbox";
	mdicreate.hOwner	= g_hInstance;
	mdicreate.x			= CW_USEDEFAULT;
	mdicreate.y			= CW_USEDEFAULT;
	mdicreate.cx		= CW_USEDEFAULT;
	mdicreate.cy		= CW_USEDEFAULT;
	mdicreate.style		= 0;
	mdicreate.lParam	= NULL;

	// Send a message to the client window to create the new child.
	SendMessage(g_hClient, WM_MDICREATE, 0,
		(LPARAM) (LPMDICREATESTRUCT) &mdicreate);
}

/*
	Function: OnMailConfigure
	Description:
		Pops up a dialog box to configure all the server name and port
		information.
*/
void OnMailConfigure(HWND hwnd)
{
	// Put up the dialog box.
	FARPROC lpfnConfigure =
		MakeProcInstance((FARPROC) ConfigureDlgProc, g_hInstance);
	BOOL bResult = DialogBox(g_hInstance, "CONFIGURE", hwnd,
		(DLGPROC) lpfnConfigure);
	FreeProcInstance(lpfnConfigure);

	// We're only configured if we were configured before or we just
	// got configured.  If we weren't before and the user hit cancel,
	// we still aren't.
	g_bConfigured = g_bConfigured || bResult;

	// Write out the .INI file.
	char szPort[20];
	if (bResult) { // Only do this if the user hit OK.
		WritePrivateProfileString("SMTP", "Server", g_pszSMTPServer,
			g_szIniFile);
		wsprintf(szPort, "%u", g_ushSMTPPort);
		WritePrivateProfileString("SMTP", "Port", szPort,
			g_szIniFile);
		WritePrivateProfileString("POP", "Server", g_pszPOPServer,
			g_szIniFile);
		wsprintf(szPort, "%u", g_ushPOPPort);
		WritePrivateProfileString("POP", "Port", szPort, g_szIniFile);
		WritePrivateProfileString("POP", "Mailbox", g_pszMailbox,
			g_szIniFile);
		WritePrivateProfileString("POP", "Password", g_pszPassword,
			g_szIniFile);
		WritePrivateProfileString("UserInfo", "Name",
			g_pszUserName, g_szIniFile);
		WritePrivateProfileString("UserInfo", "Address",
			g_pszUserAddress, g_szIniFile);
	}
}

/*
	Function: OnWindowCloseAll
	Description:
		Attempts to close all the child windows.
*/
void OnWindowCloseAll(HWND hwnd)
{
	UNUSED_PARAMETER(hwnd);

	// Simply enumerate all the child windows, calling CloseEnumProc
    // for each one.
	FARPROC lpfnCloseEnum = MakeProcInstance((FARPROC) CloseEnumProc,
		g_hInstance);
	EnumChildWindows(g_hClient, (WNDENUMPROC) lpfnCloseEnum, 0);
    FreeProcInstance(lpfnCloseEnum);
}

/*
	Function: CloseEnumProc
	Description:
		The enumeration function used in the OnWindowCloseAll
		function.
*/
BOOL CALLBACK CloseEnumProc(HWND hwnd, LPARAM lParam)
{
	UNUSED_PARAMETER(lParam);

	// EnumChildWindows picks up the title windows of iconized child
	// windows.  Skip over these icon titles by checking to see if
	// they are owned by anybody.  Icon titles are owned while child
	// windows aren't.
	if (GetWindow(hwnd, GW_OWNER) != NULL) {
		return TRUE;
	}
	// Restore the child window to normal size.
	SendMessage(GetParent(hwnd), WM_MDIRESTORE, (WPARAM) hwnd, 0);
	// Ask the child whether we can close it.
	if (!SendMessage(hwnd, WM_QUERYENDSESSION, 0, 0)) {
		// If not, move on.
		return TRUE;
	}
	// Close it!
	SendMessage(GetParent(hwnd), WM_MDIDESTROY, (WPARAM) hwnd, 0);
	// Go to the next child window.
	return TRUE;
}

/*
	Function: OnHelpAbout
	Description:
		Pops up the "About..." dialog box in response to IDM_HELPABOUT
		command messages.
*/
void OnHelpAbout(HWND hwnd)
{
	char szTitle[100];
	char szInfo[200];

	wsprintf(szTitle, "About %s", (LPSTR)g_szFullAppName);
	wsprintf(szInfo, "%s\r\nCopyright \251 1995\r\n"
		"David G. Roberts\r\n\r\nAn example program from the book\r\n"
		"'Developing for the Internet with Windows Sockets'",
		(LPSTR)g_szFullAppName);
	MessageBox(hwnd, szInfo, szTitle, MB_OK);
}

/*********************************************************************
	UTILITY FUNCTIONS
*********************************************************************/

/*
	Function: ForwardDefFrameProc
	Description:
		A short routine that provides some syntactic sugar to allow
		the windowsx.h FORWARD_WM_COMMAND macro to work with
		DefFrameProc.  Normally, FORWARD_WM_COMMAND packages its
		parameters and tries to call a function with the normal number
		and types of WndProc parameters (HWND, UINT, WPARAM, LPARAM).
		DefFrameProc requires the client HWND to also be present.
		Use FOWARD_WM_COMMAND to call this procedure which then adds
		the client HWND and calls DefFrameProc.
*/
void ForwardDefFrameProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	DefFrameProc(hwnd, g_hClient, msg, wParam, lParam);
}

/*
	Function: GetMDIActiveChild
	Description:
		This is one function that differs between Win16 and Win32.
		GetMDIActiveChild isolates those differences in this function
		rather than writing Win16/Win32 specific code everywhere.
*/
HWND GetMDIActiveChild(HWND hwndClient)
{
#ifdef WIN32
	return (HWND) SendMessage(hwndClient, WM_MDIGETACTIVE, 0, 0);
#else
	return (HWND) LOWORD(SendMessage(hwndClient, WM_MDIGETACTIVE,
		0, 0));
#endif
}

/*
	Function: VerifyDlgInt
	Description:
		Verifies that an edit field in a dialog box contains a valid
		integer.  The function works by calling GetDlgItemInt on the
		edit control and checking the bTranslated argument.  The
		caller may specify a signed or unsigned translation with the
		bSigned function argument.  If the item doesn't translate,
		the function highlights the text in the edit box, beeps, and
		returns FALSE.
*/
BOOL VerifyDlgInt(HWND hwnd, int iDlgItem, BOOL bSigned)
{
	BOOL bTranslated;

	// Try to retrieve the number.
	GetDlgItemInt(hwnd, iDlgItem, (LPBOOL) &bTranslated,
		bSigned);
	// If things didn't go well, set the focus to the
	// edit box, select all the text in it, and beep.
	if (!bTranslated) {
		SetEditErrorFocus(hwnd, iDlgItem);
		return FALSE;
	}
    return TRUE;
}

/*
	Function: VerifyDlgText
	Description:
		Verfies that an edit text control in a dialog box contains
		some text (i.e., is not null).  If the length of the text in
		the control is zero, the function sets the focus to the edit
		text item, beeps, and returns FALSE.
*/
BOOL VerifyDlgText(HWND hwnd, int iDlgItem)
{
	// Get the text length from the edit box.
	int nLen = Edit_GetTextLength(GetDlgItem(hwnd, iDlgItem));
	// If the user has given us a blank string, set the
	// focus back to the edit box and beep.
	if (nLen == 0) {
		SetEditErrorFocus(hwnd, iDlgItem);
		return FALSE;
	}
    return TRUE;
}

/*
	Function: SetEditErrorFocus
	Description:
		Sets the focus to a given dialog control, selects all its
		text, and beeps.  The dialog window handle is specified by
        hwnd and the id of the dialog item by iDlgItem.
*/
void SetEditErrorFocus(HWND hwnd, int iDlgItem)
{
	SetFocus(GetDlgItem(hwnd, iDlgItem));
	Edit_SetSel(GetDlgItem(hwnd, iDlgItem), 0, -1);
	MessageBeep(MB_ICONSTOP);
}

