/*
	File: NEWSART.CPP
	Description:
		Implements the article viewer window class for News Client.
	Copyright 1995, David G. Roberts
*/

#include <string.h>
#include <dos.h>
#include <fstream.h>
#include <windows.h>
#include <windowsx.h>
#include "newscli.rh"
#include "myassert.h"
#include "cnntp.h"

#define DECLARE
#include "newsclig.h"

/*
	CONSTANTS
*/
#define EDITCTLID		1
#define WM_NNTPCONNECTED	WM_USER
#define WM_NNTPGROUPDONE	(WM_USER + 1)
#define WM_NNTPARTICLEDONE	(WM_USER + 2)
#define WM_NNTPQUITDONE		(WM_USER + 3)

/*
	TYPES
*/

// ARTICLEDATA holds some data associated with the viewer window.
struct ARTICLEDATA {
	HWND		hwndEdit;
	char *		pszGroup;
	long		lArticleNum;
	CNNTP *		nntpobj;
    BOOL		bRetrieving;

	ARTICLEDATA::ARTICLEDATA()
		{ pszGroup = NULL; nntpobj = NULL; bRetrieving = FALSE; };
	ARTICLEDATA::~ARTICLEDATA()
		{ delete[] pszGroup; delete nntpobj; };
};

/*
	LOCAL FUNCTIONS
*/
static void OnClose(HWND);
static void OnCommand(HWND, int, HWND, UINT);
static BOOL OnCreate(HWND, CREATESTRUCT FAR*);
static void OnDestroy(HWND);
static void OnMDIActivate(HWND, BOOL, HWND, HWND);
static void OnNNTPConnect(HWND, UINT, WPARAM, LPARAM);
static void OnNNTPGroupDone(HWND, UINT, WPARAM, LPARAM);
static void OnNNTPArticleDone(HWND, UINT, WPARAM, LPARAM);
static void OnNNTPQuitDone(HWND, UINT, WPARAM, LPARAM);
static BOOL OnQueryEndSession(HWND hwnd);
static void OnSize(HWND, UINT, int, int);

/*
	UTILITY FUNCTIONS
*/
static void SetStatus(HWND, char *);
static BOOL PutFileIntoEditCtl(char *, HWND);

/*********************************************************************
	WINDOW AND DIALOG BOX PROCEDURES
*********************************************************************/

/*
	Function: ArticleWndProc
	Description:
		Window proc for the viewer window.
*/
LRESULT CALLBACK ArticleWndProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	switch (msg) {
		HANDLE_MSG(hwnd, WM_CLOSE, OnClose);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
		HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
		HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
		HANDLE_MSG(hwnd, WM_MDIACTIVATE, OnMDIActivate);
		HANDLE_MSG(hwnd, WM_QUERYENDSESSION, OnQueryEndSession);
		HANDLE_MSG(hwnd, WM_SIZE, OnSize);
		case WM_NNTPCONNECTED:
			OnNNTPConnect(hwnd, msg, wParam, lParam);
			break;
		case WM_NNTPGROUPDONE:
			OnNNTPGroupDone(hwnd, msg, wParam, lParam);
			break;
		case WM_NNTPARTICLEDONE:
			OnNNTPArticleDone(hwnd, msg, wParam, lParam);
			break;
		case WM_NNTPQUITDONE:
			OnNNTPQuitDone(hwnd, msg, wParam, lParam);
			break;
	}
	return DefMDIChildProc(hwnd, msg, wParam, lParam);
}

/*********************************************************************
	MESSAGE HANDLING FUNCTIONS
*********************************************************************/

/*
	Function: OnClose
	Description:
		Determine if the window can be closed, and if so, destroy it.
*/
void OnClose(HWND hwnd)
{
	if (SendMessage(hwnd, WM_QUERYENDSESSION, 0, 0)) {
		SendMessage(g_hClient, WM_MDIDESTROY, (WPARAM) hwnd, 0);
	}
}

/*
	Function: OnCommand
	Description:
		Handles WM_COMMAND messages.
*/
void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	ARTICLEDATA FAR *lpdata =
		(ARTICLEDATA FAR *)GetWindowLong(hwnd, 0);

	if (hwndCtl == NULL) {
		// It's a window message.
		switch (id) {
			case IDM_EDITCOPY:
				SendMessage(lpdata->hwndEdit, WM_COPY, 0, 0);
				return;
		}
	}
	else {
		ASSERT(id == EDITCTLID);
		// It's from our control.
		if (codeNotify == EN_ERRSPACE) {
			MessageBox(hwnd, "Editor out of space.", "Error",
				MB_OK | MB_ICONEXCLAMATION);
			return;
		}
	}
	// Make sure the default child proc sees things we don't
	// process ourselves!
	FORWARD_WM_COMMAND(hwnd, id, hwndCtl, codeNotify,
		DefMDIChildProc);
}

/*
	Function: OnCreate
	Description:
		Creates the editor window in which the messages are
		displayed, parses the message, and displays the first
		body part.
*/
BOOL OnCreate(HWND hwnd, CREATESTRUCT FAR* lpCreateStruct)
{
	// Allocate some private window data for ourselves.
	ARTICLEDATA FAR * lpdata = new ARTICLEDATA;
	SetWindowLong(hwnd, 0, (LONG) lpdata);

	// Create a multi-line edit control to do our work for us.
	lpdata->hwndEdit = CreateWindow(
		"EDIT",						// class name
		NULL,						// window name
		WS_CHILD | WS_VISIBLE |		// window styles
		WS_HSCROLL | WS_VSCROLL | ES_LEFT |
		ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL | ES_READONLY,
		0,0,0,0,					// location and size
		hwnd,						// parent
		(HMENU) EDITCTLID,			// child id, not really a menu
		g_hInstance,				// program instance
		NULL);						// creation data

	// Set the edit control's font to a fixed pitch.
	HFONT hfont = GetStockFont(SYSTEM_FIXED_FONT);
	SetWindowFont(lpdata->hwndEdit, hfont, FALSE);

	// Get the group name and article number from the creation
	// parameters lParam.
	GROUPANDARTICLE * pGroupAndArt = (GROUPANDARTICLE *)
		(((MDICREATESTRUCT FAR *)
		(lpCreateStruct->lpCreateParams))->lParam);
	lpdata->pszGroup = pGroupAndArt->pszGroupName;
	lpdata->lArticleNum = pGroupAndArt->lArticleNum;
    delete pGroupAndArt;

	// Create the CNNTP object and start the article retrieval
	// process.
	lpdata->nntpobj = new CNNTP;
	lpdata->nntpobj->Connect(hwnd, WM_NNTPCONNECTED, g_pszNNTPServer,
		g_ushNNTPPort);
	SetStatus(hwnd, " : connecting...");

	lpdata->bRetrieving = TRUE;

	return TRUE;
}

/*
	Function: OnDestroy
	Description:
		Handles a WM_DESTROY message.
*/
void OnDestroy(HWND hwnd)
{
	// Deallocate our private window data before we go away.
	ARTICLEDATA FAR *lpdata =
		(ARTICLEDATA FAR *)GetWindowLong(hwnd, 0);
	delete lpdata;
}

/*
	Function: OnMDIActivate
	Description:
		Takes care of putting the viewer menu into the menu
		bar when the viewer window becomes active.
*/
void OnMDIActivate(HWND hwnd, BOOL fActive, HWND hwndActivate,
	HWND hwndDeactivate)
{
	UNUSED_PARAMETER(hwndActivate);
	UNUSED_PARAMETER(hwndDeactivate);

	if (fActive) {
		// Put our menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hArticleMenu,
				g_hArticleWindowMenu));
		ARTICLEDATA FAR *lpdata =
			(ARTICLEDATA FAR*)GetWindowLong(hwnd, 0);
		// Make sure keystrokes go to the edit control.
		SetFocus(lpdata->hwndEdit);
	}
	else {
		// We're being deactivated, so put the main menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hFrameMenu,
				g_hFrameWindowMenu));
	}
	DrawMenuBar(g_hFrame);
}

/*
	Function: OnNNTPConnect
	Description:
		This function is invoked when the WM_NNTPCONNECTED
		message is received.  It checks for errors and then
		selects the NNTP group.
*/
void OnNNTPConnect(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	UNUSED_PARAMETER(msg);
	UNUSED_PARAMETER(lParam);

	ARTICLEDATA FAR * lpdata =
		(ARTICLEDATA FAR *) GetWindowLong(hwnd, 0);

	if (wParam != CNNTP::RESULT_SUCCESS) {
		MessageBox(hwnd, "Could not connect to server!", "Error",
			MB_OK | MB_ICONEXCLAMATION);
		lpdata->bRetrieving = FALSE;
		delete lpdata->nntpobj;
		PostMessage(hwnd, WM_CLOSE, 0, 0);
		return;
	}

	// Set the title to the group name + "retrieving article count..."
	SetStatus(hwnd, " : selecting group...");

	lpdata->nntpobj->Group(hwnd, WM_NNTPGROUPDONE,
		lpdata->pszGroup);
}

/*
	Function: OnNNTPGroupDone
	Description:
		Invoked when the window receives a WM_NNTPGROUPDONE message.
		The function checks for errors and then starts the
		article download process.
*/
void OnNNTPGroupDone(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	UNUSED_PARAMETER(msg);
	UNUSED_PARAMETER(lParam);

	ARTICLEDATA FAR * lpdata =
		(ARTICLEDATA FAR *) GetWindowLong(hwnd, 0);

	if (wParam != CNNTP::RESULT_SUCCESS) {
		MessageBox(hwnd, "Invalid group name!", "Error",
			MB_OK | MB_ICONEXCLAMATION);
		lpdata->bRetrieving = FALSE;
		delete lpdata->nntpobj;
		PostMessage(hwnd, WM_CLOSE, 0, 0);
		return;
	}

	// Get the article.
	SetStatus(hwnd, " : retrieving article...");
	lpdata->nntpobj->Article(hwnd, WM_NNTPARTICLEDONE,
		lpdata->lArticleNum);
}

/*
	Function: OnNNTPArticleDone
	Description:
		Invoked when the window receives a WM_NNTPARTICLEDONE
		message.  It deals with any errors, puts the message text
		into the edit control window, and then closes the NNTP
		connection.
*/
void OnNNTPArticleDone(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	UNUSED_PARAMETER(msg);

	CNNTP::ARTICLEANDFILENAME * pArtAndFile;

	ARTICLEDATA FAR * lpdata =
		(ARTICLEDATA FAR *) GetWindowLong(hwnd, 0);

	// Check return value from CNNTP object.
	switch (wParam) {
		case CNNTP::RESULT_SUCCESS:
			// Parse the article header.
			pArtAndFile = (CNNTP::ARTICLEANDFILENAME *) lParam;

			// Put the file into the viewer.
			PutFileIntoEditCtl(pArtAndFile->pszFilename,
				lpdata->hwndEdit);

			// Delete the file and the memory allocated for the
			// ARTICLEANDFILENAME structure.
			unlink(pArtAndFile->pszFilename);
			delete[] pArtAndFile->pszFilename;
            delete pArtAndFile;

			// Quit.
			SetStatus(hwnd, " : closing connection...");
			lpdata->nntpobj->Quit(hwnd, WM_NNTPQUITDONE);
			break;
		case CNNTP::RESULT_FAILURE:
        	// Something happened.  Bogus article number, perhaps?
			MessageBox(hwnd, "An error occurred while retrieving "
				"the article.", "Error!",
				MB_OK | MB_ICONEXCLAMATION);
			// Leave.
			SetStatus(hwnd, " : closing connection...");
			lpdata->nntpobj->Quit(hwnd, WM_NNTPQUITDONE);
			break;
		default:
			MessageBox(hwnd, "An error occurred while retrieving "
				"article headers.", "Error!",
				MB_OK | MB_ICONEXCLAMATION);
			lpdata->bRetrieving = FALSE;
			delete lpdata->nntpobj;
			PostMessage(hwnd, WM_CLOSE, 0, 0);
			break;
	}
	return;
}

/*
	Function: OnNNTPQuitDone
	Description:
		Invoked when the window receives a WM_NNTPQUITDONE message.
		The function destroys the CNNTP object and reports
		any error that may have resulted.
*/
void OnNNTPQuitDone(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	UNUSED_PARAMETER(msg);
	UNUSED_PARAMETER(lParam);

	ARTICLEDATA FAR * lpdata =
		(ARTICLEDATA FAR *) GetWindowLong(hwnd, 0);

	delete lpdata->nntpobj;
	lpdata->nntpobj = NULL;
	lpdata->bRetrieving = FALSE;

	if (wParam != CNNTP::RESULT_SUCCESS) {
		MessageBox(hwnd, "An error occurred while closing the "
			"connection.", "Error!",
			MB_OK | MB_ICONEXCLAMATION);
		lpdata->bRetrieving = FALSE;
		delete lpdata->nntpobj;
		PostMessage(hwnd, WM_CLOSE, 0, 0);
	}

	SetStatus(hwnd, "");
}

/*
	Function: OnQueryEndSession
	Description:
		Handles WM_QUERYENDSESSION messages. Returns TRUE if the
		window should be closed, FALSE otherwise.
*/
BOOL OnQueryEndSession(HWND hwnd)
{
	ARTICLEDATA FAR * lpdata =
		(ARTICLEDATA FAR *) GetWindowLong(hwnd, 0);
	if (lpdata->bRetrieving) {
		SendMessage(g_hClient, WM_MDIRESTORE, (WPARAM) hwnd, 0);
		int iResult = MessageBox(hwnd,
			"Retrieving article.  Close anyway?",
			"Close Anyway?", MB_YESNO | MB_ICONQUESTION);
		if (iResult == IDYES)
			return TRUE;
		else
			return FALSE;
	}
	else {
		return TRUE;
	}
}

/*
	Function: OnSize
	Description:
		Notifies the editor child window whenever the viewer window
		gets resized.
*/
void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
	UNUSED_PARAMETER(state);

	ARTICLEDATA FAR *lpdata =
		(ARTICLEDATA FAR *)GetWindowLong(hwnd, 0);
	MoveWindow(lpdata->hwndEdit, 0, 0, cx, cy, TRUE);
	// Must make sure we forward this to the default child proc
	// as well!!
	FORWARD_WM_SIZE(hwnd, state, cx, cy, DefMDIChildProc);
}

/*********************************************************************
	UTILITY FUNCTIONS
*********************************************************************/

/*
	Function: SetStatus
	Description:
		Changes the window title to include a status message.
		This is used to report the current connection status.
*/
void SetStatus(HWND hwnd, char * pszStatus)
{
	const char szArticle[] = "Article";
	int nLen = lstrlen(szArticle) + lstrlen(pszStatus);
	char * pszTemp = new char[nLen + 1];
	lstrcpy(pszTemp, szArticle);
	lstrcat(pszTemp, pszStatus);
	SetWindowText(hwnd, pszTemp);
	delete[] pszTemp;
}

/*
	Function: PutFileIntoEditCtl
	Description:
		Reads a file and stuffs the text into the specified
		edit control.  Short, simple, and to the point. :-)
*/
BOOL PutFileIntoEditCtl(char * pszFilename, HWND hwndEdit)
{
	// Open file.
	HFILE hFile = _lopen(pszFilename, OF_READ);
	if (hFile == HFILE_ERROR)
		return FALSE;

	// Determine file length by seeking to end of file.
	long lLength = _llseek(hFile, 0, 2);
	// Move back to beginning.
	_llseek(hFile, 0, 0); 

#ifndef WIN32
	// Win16 _lread can only handle 0xFFFE bytes at a time.
    // If we're compiling for Win16, clamp lLength to 0xFFFE.
	if (lLength > 0xFFFE)
		lLength = 0xFFFE;
#endif

	// Allocate the buffer.
	char * pBuffer = new char[lLength + 1];
	if (pBuffer == NULL) {
		_lclose(hFile);
		return FALSE;
	}

	// Read the file into the buffer.
	if ((HFILE)_lread(hFile, pBuffer, (UINT) lLength) == HFILE_ERROR) {
		_lclose(hFile);
		return FALSE;
	}
	_lclose(hFile);

	// Terminate the buffer.
    pBuffer[lLength] = '\0';

    // Stuff the text into the viewer edit control.
	SetWindowText(hwndEdit, pBuffer);
	delete[] pBuffer;

    return TRUE;
}


