/*
	File: NEWSGRP.CPP
	Description:
		Implements the group article selector for News Client.
	Copyright 1995, David G. Roberts
*/

#include <ctype.h>
#include <dos.h>
#include <io.h>
#include <stdlib.h>
#include <time.h>
#include <fstream.h>
#include <windows.h>
#include <windowsx.h>
#include "newscli.rh"
#include "myassert.h"
#include "cwinsock.h"
#include "clist.h"
#include "cnntp.h"
#include "cstrlist.h"
#include "rfc822.h"
#include "parsdate.h"
#include "utility.h"

#define DECLARE
#include "newsclig.h"

/*
	CONSTANTS
*/
#define LBCTLID		1
#define WM_NNTPCONNECTED	WM_USER
#define WM_NNTPGROUPDONE	(WM_USER + 1)
#define WM_NNTPHEADERDONE	(WM_USER + 2)
#define WM_NNTPQUITDONE		(WM_USER + 3)

/*
	TYPES
*/

// A little collection structure to hold summary data about
// a particular article.  Note that the collection deletes all the
// string members upon destruction.  Make copies beforehand!
struct SUMMARYINFO {
	char *	pszFrom;
	char *	pszSubject;
	time_t	time;
	long	lArticleNum;

	SUMMARYINFO() : pszFrom(NULL), pszSubject(NULL), time(0),
		lArticleNum(0) {};
	SUMMARYINFO(char * pszfr, char * pszsub, time_t d, int i) :
		pszFrom(pszfr), pszSubject(pszsub),
		time(d), lArticleNum(i) {};
	~SUMMARYINFO() {
		delete[] pszFrom;
		delete[] pszSubject;
	};
};

// A list class to manage SUMMARYINFO data.  Derived from CList.
class CSummaryList : public CList {
protected:
	virtual int Compare(void * pvThing1, void * pvThing2);
	virtual void DeleteData(void * pvThing);
public:
	CSummaryList() { };
	~CSummaryList();
	void Add(SUMMARYINFO * psi) { CList::Add((void *) psi); };
	void Insert(SUMMARYINFO * psi) { CList::Insert((void *) psi); };
	SUMMARYINFO * First(void)
		{ return (SUMMARYINFO *) CList::First(); };
	SUMMARYINFO * Next(void)
		{ return (SUMMARYINFO *) CList::Next(); };
};

CSummaryList::~CSummaryList()
{
	// Delete all the data in the list.
	SUMMARYINFO * psi = First();
	while (psi != NULL) {
		delete psi;
		psi = Next();
	}
}

int CSummaryList::Compare(void * pvThing1, void *pvThing2)
{
	SUMMARYINFO * psi1 = (SUMMARYINFO *) pvThing1;
	SUMMARYINFO * psi2 = (SUMMARYINFO *) pvThing2;

	if (psi1->lArticleNum < psi2->lArticleNum)
		return -1;
	else if (psi1->lArticleNum == psi2->lArticleNum)
		return 0;
	else
		return 1;
}

void CSummaryList::DeleteData(void * pvThing)
{
	delete (SUMMARYINFO *) pvThing;
}

// A little collection structure to hold some data that is associated
// with the group window.
struct GROUPDATA {
	HWND				hwndLB;
	BOOL				bRetrieving;
	CSummaryList *		pslSummaryList;
	char *				pszGroupName;
	CNNTP *				nntpobj;
    long				lTotal;
	long				lCurrent;

	GROUPDATA() : hwndLB(NULL), bRetrieving(FALSE),
		pslSummaryList(new CSummaryList), nntpobj(NULL) {};
	~GROUPDATA();
};

GROUPDATA::~GROUPDATA()
{
	if (bRetrieving) {
		bRetrieving = FALSE;
		nntpobj->Abort(FALSE, NULL, 0);
		delete nntpobj;
	}
	delete pslSummaryList;
	delete[] pszGroupName;
}

/*
	LOCAL FUNCTIONS
*/
static void OnClose(HWND);
static void OnCommand(HWND, int, HWND, UINT);
static BOOL OnCreate(HWND, CREATESTRUCT FAR*);
static void OnDestroy(HWND);
static void OnMDIActivate(HWND, BOOL, HWND, HWND);
static void OnNNTPConnect(HWND, UINT, WPARAM, LPARAM);
static void OnNNTPGroupDone(HWND, UINT, WPARAM, LPARAM);
static void OnNNTPHeaderDone(HWND, UINT, WPARAM, LPARAM);
static void OnNNTPQuitDone(HWND, UINT, WPARAM, LPARAM);
static BOOL OnQueryEndSession(HWND hwnd);
static void OnSize(HWND, UINT, int, int);

// MENU HANDLERS
static void OnNewsViewArticle(HWND);

// UTILITY FUNCTIONS
void SetRetrievalStatus(HWND);
void SetStatus(HWND, char *);
void CreateArticleList(HWND);
SUMMARYINFO * GetSummaryInfo(char *, long);
char * GetHeaderData(char *, char *);

/*********************************************************************
	WINDOW AND DIALOG BOX PROCEDURES
*********************************************************************/

/*
	Function: GroupWndProc
	Description:
		The window procedure for the group window class.  The
		function simply dispatches window messages to smaller
		"handler" functions.
*/
LRESULT CALLBACK GroupWndProc(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	switch (msg) {
    	HANDLE_MSG(hwnd, WM_CLOSE, OnClose);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
		HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
		HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
		HANDLE_MSG(hwnd, WM_MDIACTIVATE, OnMDIActivate);
		HANDLE_MSG(hwnd, WM_QUERYENDSESSION, OnQueryEndSession);
		HANDLE_MSG(hwnd, WM_SIZE, OnSize);
		case WM_NNTPCONNECTED:
			OnNNTPConnect(hwnd, msg, wParam, lParam);
			break;
		case WM_NNTPGROUPDONE:
			OnNNTPGroupDone(hwnd, msg, wParam, lParam);
			break;
		case WM_NNTPHEADERDONE:
			OnNNTPHeaderDone(hwnd, msg, wParam, lParam);
			break;
		case WM_NNTPQUITDONE:
			OnNNTPQuitDone(hwnd, msg, wParam, lParam);
			break;
	}
    return DefMDIChildProc(hwnd, msg, wParam, lParam);
}

/*********************************************************************
	MESSAGE HANDLING FUNCTIONS
*********************************************************************/

/*
	Function: OnClose
	Description:
		Handles WM_CLOSE messages.  Simply determines if this is
		a good time to destroy the window, and if so, does the deed.
*/
void OnClose(HWND hwnd)
{
	if (SendMessage(hwnd, WM_QUERYENDSESSION, 0, 0)) {
		SendMessage(g_hClient, WM_MDIDESTROY, (WPARAM) hwnd, 0);
	}
}

/*
	Function: OnCommand
	Description:
		Handles WM_COMMAND messages.
*/
void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	if (hwndCtl == NULL) {
		// It's a window message.
		switch (id) {
			case IDM_NEWSVIEWARTICLE:
				OnNewsViewArticle(hwnd);
				return;
		}
	}
	else {
		ASSERT(id == LBCTLID);
		// It's from our control.
		if (codeNotify == LBN_DBLCLK) {
			OnNewsViewArticle(hwnd);
			return;
		}
	}
	// Make sure the default child proc sees things we don't
	// process ourselves!
	FORWARD_WM_COMMAND(hwnd, id, hwndCtl, codeNotify,
		DefMDIChildProc);
}

/*
	Function: OnCreate
	Description:
		Handles WM_CREATE messages.  The function allocates
		private memory, creates a list box control to hold the
		summary listing, and then starts the header retrieval
		process.
*/
BOOL OnCreate(HWND hwnd, CREATESTRUCT FAR* lpCreateStruct)
{
	UNUSED_PARAMETER(lpCreateStruct);

	// Allocate some private window data for ourselves.
	GROUPDATA FAR * lpdata = new GROUPDATA;
	SetWindowLong(hwnd, 0, (LONG) lpdata);

	// Create a list box control to do our work for us.
	lpdata->hwndLB = CreateWindow(
		"LISTBOX",					// class name
		NULL,						// window name
		WS_CHILD | WS_VISIBLE |		// window styles
		LBS_NOTIFY | WS_VSCROLL,
		0,0,0,0,					// location and size
		hwnd,						// parent
		(HMENU) LBCTLID,			// child id, not really a menu
		g_hInstance,				// program instance
		NULL);						// creation data

	// Set the edit control's font to a fixed pitch.
	HFONT hfont = GetStockFont(SYSTEM_FIXED_FONT);
	SetWindowFont(lpdata->hwndLB, hfont, FALSE);

	// Get the name of the group from the window title.
	int nLen = GetWindowTextLength(hwnd);
	lpdata->pszGroupName = new char[nLen + 1];
	GetWindowText(hwnd, lpdata->pszGroupName, nLen + 1);

	// Set the title to the group name + "connecting..."
	SetStatus(hwnd, " : connecting...");

	// Create a CNNTP object and have it connect.
	lpdata->nntpobj = new CNNTP;
	lpdata->nntpobj->Connect(hwnd, WM_NNTPCONNECTED, g_pszNNTPServer,
		g_ushNNTPPort);

	lpdata->bRetrieving = TRUE;

	return TRUE;
}

/*
	Function: OnDestroy
	Description:
		Handles WM_DESTROY messages.  The function simply deletes
		the window's private data structure.
*/
void OnDestroy(HWND hwnd)
{
	// Deallocate our private window data before we go away.
	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);
	delete lpdata;
}

/*
	Function: OnMDIActivate
	Description:
		This function takes care of putting the appropriate menu
		into the menu bar when the window becomes active or inactive.
*/
void OnMDIActivate(HWND hwnd, BOOL fActive, HWND hwndActivate,
	HWND hwndDeactivate)
{
	UNUSED_PARAMETER(hwndActivate);
	UNUSED_PARAMETER(hwndDeactivate);

	if (fActive) {
		// Put our menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hGroupMenu,
				g_hGroupWindowMenu));
		GROUPDATA FAR * lpdata =
			(GROUPDATA FAR*) GetWindowLong(hwnd, 0);
		// Make sure keystrokes go to the edit control.
		SetFocus(lpdata->hwndLB);
	}
	else {
		// We're being deactivated, so put the main menu in place.
		SendMessage(g_hClient, WM_MDISETMENU,
			GET_WM_MDISETMENU_MPS(g_hFrameMenu, g_hFrameWindowMenu));
	}
	DrawMenuBar(g_hFrame);
}

/*
	Function: OnNNTPConnect
	Description:
		This function handles WM_NNTPCONNECTED messages.
		If no errors occurred while connecting, the function
		selects the appropriate news group.
*/
void OnNNTPConnect(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	UNUSED_PARAMETER(msg);
	UNUSED_PARAMETER(lParam);

	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);

	if (wParam != CNNTP::RESULT_SUCCESS) {
		MessageBox(hwnd, "Could not connect to server!", "Error",
			MB_OK | MB_ICONEXCLAMATION);
		lpdata->bRetrieving = FALSE;
		delete lpdata->nntpobj;
		PostMessage(hwnd, WM_CLOSE, 0, 0);
		return;
	}

	// Set the title to the group name + "retrieving article count..."
	SetStatus(hwnd, " : retrieving article count...");

	lpdata->nntpobj->Group(hwnd, WM_NNTPGROUPDONE,
		lpdata->pszGroupName);
}

/*
	Function: OnNNTPGroupDone
	Description:
		This function handles WM_NNTPGROUPDONE messages sent to the
		window.  If no error occurred while trying to select the
		right group, this function starts retrieving the first
		header.
*/
void OnNNTPGroupDone(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	UNUSED_PARAMETER(msg);

	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);

	if (wParam != CNNTP::RESULT_SUCCESS) {
		MessageBox(hwnd, "Invalid group name!", "Error",
			MB_OK | MB_ICONEXCLAMATION);
		lpdata->bRetrieving = FALSE;
		delete lpdata->nntpobj;
		PostMessage(hwnd, WM_CLOSE, 0, 0);
		return;
	}

	// Get the first and last article data.
	lpdata->lTotal = (long) lParam;
	lpdata->lCurrent = 0;

	// Get the first header.
	SetRetrievalStatus(hwnd);
	lpdata->nntpobj->Header(hwnd, WM_NNTPHEADERDONE);
}

/*
	Function: OnNNTPHeaderDone
	Description:
		This function is invoked when each header is retrieved.
		If no errors occurred, it parses the header and starts
		the retrieval process for the next.
*/
void OnNNTPHeaderDone(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	UNUSED_PARAMETER(msg);

	CNNTP::ARTICLEANDFILENAME * pArtAndFile;
	SUMMARYINFO * psi;
	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);

	// Check return value from CNNTP object.
	switch (wParam) {
		case CNNTP::RESULT_SUCCESS:
			// Goto the next header before parsing the current
			// header to speed things up.
			lpdata->lCurrent++;
			SetRetrievalStatus(hwnd);
			lpdata->nntpobj->NextHeader(hwnd, WM_NNTPHEADERDONE);

			// Parse the article header.
			pArtAndFile = (CNNTP::ARTICLEANDFILENAME *) lParam;
			psi = GetSummaryInfo(pArtAndFile->pszFilename,
				pArtAndFile->lArticleNum);
			lpdata->pslSummaryList->Add(psi);

			// Delete the file and the memory allocated for the
			// ARTICLEANDFILENAME structure.
			unlink(pArtAndFile->pszFilename);
			delete[] pArtAndFile->pszFilename;
            delete pArtAndFile;

			break;
		case CNNTP::RESULT_FAILURE:
			// There are no more messages.
			SetStatus(hwnd, " : closing connection...");
			lpdata->nntpobj->Quit(hwnd, WM_NNTPQUITDONE);
			break;
		default:
			MessageBox(hwnd, "An error occurred while retrieving "
				"article headers.", "Error!",
				MB_OK | MB_ICONEXCLAMATION);
			lpdata->bRetrieving = FALSE;
			delete lpdata->nntpobj;
			PostMessage(hwnd, WM_CLOSE, 0, 0);
			break;
	}
	return;
}

/*
	Function: OnNNTPQuitDone
	Description:
		This function is invoked when the connection to the NNTP
		server has been closed.  It destroys the CNNTP object and
		reports any errors that might have occurred.
*/
void OnNNTPQuitDone(HWND hwnd, UINT msg, WPARAM wParam,
	LPARAM lParam)
{
	UNUSED_PARAMETER(msg);
    UNUSED_PARAMETER(lParam);

	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);

	delete lpdata->nntpobj;
	lpdata->nntpobj = NULL;
	lpdata->bRetrieving = FALSE;

	if (wParam != CNNTP::RESULT_SUCCESS) {
		MessageBox(hwnd, "An error occurred while closing the "
			"connection.", "Error!",
			MB_OK | MB_ICONEXCLAMATION);
		lpdata->bRetrieving = FALSE;
		delete lpdata->nntpobj;
		PostMessage(hwnd, WM_CLOSE, 0, 0);
	}

	SetStatus(hwnd, "");

    CreateArticleList(hwnd);
}

/*
	Function: OnQueryEndSession
	Description:
		Handles WM_QUERYENDSESSION messages. Returns TRUE if the
		window should be closed, FALSE otherwise.
*/
BOOL OnQueryEndSession(HWND hwnd)
{
	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);
	if (lpdata->bRetrieving) {
		SendMessage(g_hClient, WM_MDIRESTORE, (WPARAM) hwnd, 0);
		int iResult = MessageBox(hwnd,
			"Retrieving article headers.  Close anyway?",
			"Close Anyway?", MB_YESNO | MB_ICONQUESTION);
		if (iResult == IDYES)
			return TRUE;
		else
			return FALSE;
	}
	else {
		return TRUE;
	}
}

/*
	Function: OnSize
	Description:
		Notifies the listbox child window whenever the viewer window
		gets resized.
*/
void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
	UNUSED_PARAMETER(state);

	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);
	MoveWindow(lpdata->hwndLB, 0, 0, cx, cy, TRUE);
	// Must make sure we forward this to the default child proc
	// as well!!
	FORWARD_WM_SIZE(hwnd, state, cx, cy, DefMDIChildProc);
}

/*********************************************************************
	MENU HANDLERS
*********************************************************************/

/*
	Function: OnNewsViewArticle
	Description:
		This function is invoked when the user chooses News/View
		Article or double-clicks on an article summary line in
		the group window.  The function creates a viewer MDI
		child window to retrieve and display the article.
*/
void OnNewsViewArticle(HWND hwnd)
{
	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);
	// Get the current selection index from the list box.
	int iCurSel = ListBox_GetCurSel(lpdata->hwndLB);
	if (iCurSel < 0) {
		MessageBox(hwnd, "You must make a selection first.",
			"Whoops!", MB_OK | MB_ICONHAND);
		return;
	}
	// Find the right summary info entry in the list that
	// corresponds to the selection index.
	SUMMARYINFO * psi = lpdata->pslSummaryList->First();
	for (int i = 0; i < iCurSel; i++,
		psi = lpdata->pslSummaryList->Next())
		; // Null loop body.

	// Create a little data structure to hold both the group
    // name and the article number.
	ASSERT(psi != NULL);
	GROUPANDARTICLE * pGroupAndArticle = new GROUPANDARTICLE;
	pGroupAndArticle->pszGroupName =
		DuplicateString(lpdata->pszGroupName);
	pGroupAndArticle->lArticleNum = psi->lArticleNum;

	// Create the viewer window.
	MDICREATESTRUCT mdicreate;
	mdicreate.szClass	= g_szArticleClass;
	mdicreate.szTitle	= "Article";
	mdicreate.hOwner	= g_hInstance;
	mdicreate.x			= CW_USEDEFAULT;
	mdicreate.y			= CW_USEDEFAULT;
	mdicreate.cx		= CW_USEDEFAULT;
	mdicreate.cy		= CW_USEDEFAULT;
	mdicreate.style		= 0;
	mdicreate.lParam	= (LPARAM)pGroupAndArticle;

	// Send a message to the client window to create the new child.
	SendMessage(g_hClient, WM_MDICREATE, 0,
		(LPARAM) (LPMDICREATESTRUCT) &mdicreate);

	return;
}

/*********************************************************************
	UTILITY FUNCITONS
*********************************************************************/

/*
	Function: SetRetrievalStatus
	Description:
		This function sets the window title to show what percentage
		of the group's headers have been retrieved.
*/
void SetRetrievalStatus(HWND hwnd)
{
	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);

	// Set the window title to indicate that we're retrieving
	// article headers.
	int iPercent = (int)((lpdata->lCurrent * 100) / lpdata->lTotal);
	const char pszStatusStr[] = " : retrieving headers (%d%%)";
	int nLen = lstrlen(pszStatusStr);
	char * pszTemp = new char[nLen + 10]; // 10 for percent and '\0'
	wsprintf(pszTemp, pszStatusStr, iPercent);
	SetStatus(hwnd, pszTemp);
	delete[] pszTemp;
}

/*
	Function: SetStatus
	Description:
		Changes the window title to include a status message.
		This is used to report the current connection status.
*/
void SetStatus(HWND hwnd, char * pszStatus)
{
	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);
	int nLen = lstrlen(lpdata->pszGroupName) +
		lstrlen(pszStatus);
	char * pszTemp = new char[nLen + 1];
	lstrcpy(pszTemp, lpdata->pszGroupName);
	lstrcat(pszTemp, pszStatus);
	SetWindowText(hwnd, pszTemp);
	delete[] pszTemp;
}

/*
	Function: CreateArticleList
	Description:
		This function creates a series of summary lines from the
		list of summary nodes.  Each summary line is the put into
		the listbox control.
*/
void CreateArticleList(HWND hwnd)
{
	const char *Months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	const int MAXFROM = 20;
	const int MAXSUBJECT = 50;

	// Clear out the list box.
	GROUPDATA FAR * lpdata =
		(GROUPDATA FAR *) GetWindowLong(hwnd, 0);
	HWND hwndLB = lpdata->hwndLB;
	SetWindowRedraw(hwndLB, FALSE);
	SendMessage(hwndLB, LB_RESETCONTENT, 0, 0);

	// Create a summary line for each item in the summary info list.
	SUMMARYINFO * psi = lpdata->pslSummaryList->First();
	while (psi != NULL) {
		// Format the summary line string using the information
        // in the summary record.
		char szSummaryLine[100];
		char szTruncFrom[MAXFROM + 1];
		char szTruncSubject[MAXSUBJECT + 1];
		struct tm * ptm = localtime(&(psi->time));
		lstrcpyn(szTruncFrom, psi->pszFrom, MAXFROM);
        lstrcpyn(szTruncSubject, psi->pszSubject, MAXSUBJECT);
		wsprintf(szSummaryLine, "%02d-%s-%02d  %-20s  %-50s",
			ptm->tm_mday, (LPSTR)Months[ptm->tm_mon], ptm->tm_year,
			(LPSTR)szTruncFrom, (LPSTR)szTruncSubject);
		ListBox_AddString(hwndLB, szSummaryLine);
		psi = lpdata->pslSummaryList->Next();
	}
	SetWindowRedraw(hwndLB, TRUE);
}

/*
	Function: GetSummaryInfo
	Description:
		This function creates a summary info node from the header
		information in a disk file.
*/
SUMMARYINFO * GetSummaryInfo(char * pszFileName, long lArticleNum)
{
	SUMMARYINFO * psiResult = new SUMMARYINFO;

	// Set the article number.
	psiResult->lArticleNum = lArticleNum;

	// Get the From: header.
	char * pszHeaderData = GetHeaderData("From:", pszFileName);
	if (pszHeaderData == NULL) {
		// If we couldn't find the header, create a zero-length
		// null terminated string.
		pszHeaderData = new char[1];
		*pszHeaderData = '\0';
	}
	psiResult->pszFrom = pszHeaderData;

	// Get the subject header.
	pszHeaderData = GetHeaderData("Subject:", pszFileName);
	if (pszHeaderData == NULL) {
		// If we couldn't find the header, create a zero-length
		// null terminated string.
		pszHeaderData = new char[1];
		*pszHeaderData = '\0';
	}
	psiResult->pszSubject = pszHeaderData;

	// Get the date header.
	pszHeaderData = GetHeaderData("Date:", pszFileName);
	if (pszHeaderData == NULL) {
		// If there was no date, pick a default time.
		psiResult->time = 0;
	}
	else {
    	// Else parse the date to return a time_t value.
		psiResult->time = ParseDateString(pszHeaderData);
        delete[] pszHeaderData;
	}

    // Return the SUMMARYINFO structure to the caller.
    return psiResult;
}

/*
	Function: GetHeaderData
	Description:
		This function reads a header field from a file containing
		a message.
*/
char * GetHeaderData(char * pszHeader, char * pszFileName)
{
	int i;
	ifstream ifsMessage(pszFileName);
	char szLineBuffer[1001];

	while (ifsMessage.peek() != EOF) {
		// Slurp up a line of data from the file.
		ifsMessage.getline(szLineBuffer, sizeof(szLineBuffer));
		// See if the line was simply a CRLF.
		if (lstrlen(szLineBuffer) == 0) {
			// If so, we reached the end of the headers before
			// finding the one we were looking for.  Exit the
			// while loop and return NULL.
			return NULL;
		}
		// See if this is the header we want.
		BOOL bMatch = TRUE;
		for (i = 0; i < lstrlen(pszHeader); i++) {
			if (tolower(pszHeader[i]) != tolower(szLineBuffer[i])) {
				// Nope, something didn't match.
				bMatch = FALSE;
				break;
			}
		}
		// If we matched, break out.
		if (bMatch)
			break;
	}
	// At this point, i==lstrlen(pszHeader).
	// Advance i past any whitespace that might have followed
	// the colon after the header name.
	while (IsLWSP(szLineBuffer[i]))
		i++;
	// Now, shrink the line down, eliminating the header name,
	// colon, and any following white space but retaining all
	// the info we really care about.
	char * pszResult = new char[lstrlen(szLineBuffer) - i + 1];
	int j = 0;
	while (szLineBuffer[i] != '\0') {
		pszResult[j++] = szLineBuffer[i++];
	}
    pszResult[j] = '\0';
	// Now, check to see if the following line is folded.  If
	// so, we should append it on to the current data.
	while (ifsMessage.peek() != EOF) {
		// We still have more file, so read in another line.
		ifsMessage.getline(szLineBuffer, sizeof(szLineBuffer));
		// See if it starts with linear white space.  If not,
		// then the line is not folded further and we're done.
		if (!IsLWSP(szLineBuffer[0]))
			break;
		// Okay, the line is folded.
		// See how long the combined line will be when we've
		// appended this line to the previous ones.
		int iLength = lstrlen(pszResult) + lstrlen(szLineBuffer);
		char * pszTemp = new char[iLength + 1];
		lstrcpy(pszTemp, pszResult);
		lstrcat(pszTemp, szLineBuffer);
		delete[] pszResult;
		pszResult = pszTemp;
	}

	return pszResult;
}

